import base64
import itertools
import json
import logging
import os
from time import time
from typing import Optional, Sequence

import boto3
import requests
import simplejson

from pscr_integration.test_lifecycle.active_session import is_smoke, is_k8s

logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("boto3").setLevel(logging.WARNING)
logging.getLogger("botocore").setLevel(logging.WARNING)

_logger = logging.getLogger(__name__)

request_id = 0


def get_container_ip() -> Sequence:
    if os.environ.get("CB_LC_POD_ADDRESS"):
        return [os.environ["CB_LC_POD_ADDRESS"]]
    response = requests.get("http://169.254.170.2/v2/metadata", timeout=30).json()
    _logger.debug(f"Metadata Response: {response}")
    ipv4addresses = []
    for container in response.get("Containers", []):
        if "Networks" in container:
            for network in container.get("Networks", []):
                if "IPv4Addresses" in network:
                    for ipv4address in network.get("IPv4Addresses", []):
                        ipv4addresses.append(ipv4address)
    _logger.info(f"Container IP addresses: {ipv4addresses}")
    return ipv4addresses


def create_user_context(org_id, user_id: int = None, user_name: str = None):
    if not user_id:
        user_id = 120214
    if not user_name:
        user_name = "test@example.com"

    # This stuff is mostly static and not needed in api calls but needs to be present,
    # need to update when that changes.
    return {
        "target_org": {
            "user_id": int(user_id),
            "user_name": user_name,
            "org_key": org_id,
            "org_id": 47979,
            "connector_id": "12121",
            "rds_url": "postgres.cbd.internal",
            "es_url": "master.default.es2.cbd.internal",
            "rbac_key": "R34VOUO8GR",
        },
        "login_org": {
            "user_id": int(user_id),
            "user_name": user_name,
            "org_key": org_id,
            "org_id": 47979,
            "connector_id": "12121",
            "rds_url": "postgres.cbd.internal",
            "es_url": "master.default.es2.cbd.internal",
            "rbac_key": "R34VOUO8GR",
        },
    }


def create_request_context():
    if is_smoke():
        th_gateway_url = f"http://{get_container_ip()[0]}:4000"
    else:
        th_gateway_url = "defense-dev01.cbdtest.io"
    return {
        "gateway_public_url": "defense-dev01.cbdtest.io",
        "gateway_ip": get_container_ip()[0] if is_k8s() else "10.95.3.12",
        "gateway_port": "4000",
        "th_gateway_url": th_gateway_url,
    }


class LambdaApiResponse:
    def __init__(self, response: dict, start: float):
        end = time()
        self.response = json.loads(response["Payload"].read())
        self.status_code = int(response.get("status_code", "500"))
        self.timestamp = start
        self.latency = int((end - start) * 1000)

    def __repr__(self):
        return f"status: {self.status_code} - latency: {self.latency}ms"

    def __str__(self):
        return str(self.response)


class LambdaApiRequest:
    def __init__(self, org_id: str, aws_lambda: str):
        self.org_id = org_id
        self.aws_lambda = aws_lambda
        self.lambda_client = boto3.client("lambda")

    def send_request(
        self,
        route: str,
        method: str,
        payload: dict,
        user_id: int = None,
        user_name: str = None,
    ) -> LambdaApiResponse:
        client_context = {
            "custom": {
                "X-User-Context": json.dumps(
                    create_user_context(org_id=self.org_id, user_id=user_id, user_name=user_name)
                ),
                "X-Original-Request-Url": route,
                "X-Original-Request-Method": method,
                "X-Auth-Token": "dummyauthtoken",
                "Cookie": "SESSION=abcd",
                "X-Request-Context": json.dumps(create_request_context()),
            }
        }
        p = bytes(json.dumps(payload), "utf-8")
        cc = base64.b64encode(bytearray(json.dumps(client_context), "utf-8")).decode("utf-8")
        start = time()
        res = self.lambda_client.invoke(
            FunctionName=self.aws_lambda,
            InvocationType="RequestResponse",
            LogType="None",
            ClientContext=cc,
            Payload=p,
        )
        return LambdaApiResponse(res, start)


class HTTPApiResponse:
    request_id = itertools.count()

    def __init__(self, response, start: float):
        end = time()
        self.status_code = response.status_code
        self.timestamp = start
        self.latency = int((end - start) * 1000)
        self._orig_response = response
        req = response.request
        if self.status_code < 200 or self.status_code > 299:
            _logger.error(
                f"HTTP RECV REQ_{next(HTTPApiResponse.request_id)} {self.status_code} {req.method} {req.url}:"
                f"{response.text}"
            )
        else:
            _logger.debug(f"HTTP RECV REQ_{next(HTTPApiResponse.request_id)} {self.status_code} {req.method} {req.url}")

        if self.status_code == 204:
            # https://tools.ietf.org/html/rfc7231#section-6.3.5 states:
            #
            #     A 204 response is terminated by the first empty line after
            #     the header fields because it cannot contain a message body.
            #
            # So we really shouldn't be trying to parse JSON out of these
            # responses.

            self.response = {}
            return

        try:
            self.response = response.json()
        except (json.decoder.JSONDecodeError, simplejson.errors.JSONDecodeError):
            _logger.error(f"Got non JSON response\n({req.method} {req.url}[{self.status_code}])\n{response.text}")
            self.response = {}

    def raise_for_status(self):
        self._orig_response.raise_for_status()

    def __repr__(self):
        return f"status: {self.status_code} - latency: {self.latency}ms"

    def __str__(self):
        return str(self.response)


class HTTPApiRequest:
    def __init__(self, org_id: str, url_prefix: str, headers=None):
        if headers is None:
            headers = {}
        self.org_id = org_id
        self.url_prefix = url_prefix
        self.headers = headers

    def send_request(
        self,
        route: str,
        method: str,
        payload: dict = None,
        params: dict = None,
        user_id: int = None,
        user_name: str = None,
        org_key: Optional[str] = None,
    ) -> HTTPApiResponse:
        # make sure to always use the /internal endpoints when contacting search-api to avoid rbac on smoke tests
        # this should be a temporary fix until search-api and feed-mgr are migrated to KC
        if (
            is_k8s()
            and is_smoke()
            and (self.url_prefix.endswith(":5003") or self.url_prefix.endswith(":5000"))
            and not route.startswith("/internal")
        ):
            route = "/internal" + route

        headers = {
            **{
                "X-User-Context": json.dumps(
                    create_user_context(
                        org_id=org_key or self.org_id,
                        user_id=user_id,
                        user_name=user_name,
                    )
                ),
                "X-Original-Request-Url": route,
                "X-Original-Request-Method": method,
                "X-Auth-Token": "dummyauthtoken",
                "Cookie": "SESSION=abcd",
                "X-Request-Context": json.dumps(create_request_context()),
            },
            **self.headers,
        }

        url = self.url_prefix + route

        start = time()
        if method == "GET":
            r = requests.get(url, headers=headers, json=payload, params=params, timeout=30)
        elif method == "POST":
            r = requests.post(url, headers=headers, json=payload, params=params, timeout=60)
        elif method == "PUT":
            r = requests.put(url, headers=headers, json=payload, params=params, timeout=60)
        elif method == "DELETE":
            r = requests.delete(url, headers=headers, json=payload, params=params, timeout=30)
        else:
            raise Exception(f"Invalid invoke method {method} for route {route}")

        return HTTPApiResponse(r, start)
