import os
import json
import logging
from enum import Enum
from functools import partial
from typing import Optional, Sequence, Dict, NamedTuple

from pscr_integration.test_lifecycle.active_session import is_k8s
from pscr_integration.api.api_request import (
    HTTPApiRequest,
    HTTPApiResponse,
    create_request_context,
    create_user_context,
)

_logger = logging.getLogger(__name__)


class MdrSuppressionClassificationBuckets(str, Enum):
    OFF = "OFF"
    LOW = "LOW"
    STANDARD = "STANDARD"
    HIGH = "HIGH"


class WatchlistEvaluationJobProperties(NamedTuple):
    threat_hunt_name: Optional[str] = None
    threat_hunt_id: Optional[str] = None
    mdr_suppression_rules_enabled: Optional[bool] = False
    mdr_suppress_above_classification_bucket: Optional[MdrSuppressionClassificationBuckets] = None

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)


class FederatedSearchAPI:
    TYPE_PROCESS = "processes"
    TYPE_ENRICHED_EVENTS = "enriched_events"

    def __init__(
        self,
        privatelink_hostname: str,
    ):
        privatelink_hostname = os.environ.get("FEDERATED_SEARCH_API_PRIVATELINK_HOSTNAME", privatelink_hostname)
        if is_k8s() and not privatelink_hostname:
            privatelink_hostname = "psc-lc-federated-search"
        url_prefix = "http://" + privatelink_hostname + ":5005"
        request_context = create_request_context()
        if not request_context["th_gateway_url"].startswith("http"):
            request_context.update(th_gateway_url="https://" + request_context["th_gateway_url"])
        user_context = create_user_context(org_id="federated")
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "X-Request-Context": json.dumps(request_context),
            "X-User-Context": json.dumps(user_context),
        }
        self._make_request = partial(HTTPApiRequest, org_id="federated", url_prefix=url_prefix, headers=headers)

    def _send_request(
        self,
        route: str,
        params: Optional[dict] = None,
        payload: Optional[dict] = None,
        method: str = "GET",
    ) -> HTTPApiResponse:
        request = self._make_request()
        return request.send_request(route=route, method=method, payload=payload, params=params)

    def api_docs(self) -> HTTPApiResponse:
        """
        Get Application health status.
        :return: an ApiResponse with either a valid response or error information
        """
        params = {"group": 1}
        return self._send_request("/internal/federated_search/api_docs", params=params)

    def health_check(self) -> HTTPApiResponse:
        """
        Get Application health status.
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request("/internal/federated_search/actuator/healthcheck")

    def create_search(
        self,
        search_type: str,
        query: str,
        fields: Optional[Sequence[str]] = None,
        rows: Optional[int] = None,
        org_keys: Optional[Sequence[str]] = None,
        customer_groups: Optional[Sequence[str]] = None,
        sort: Optional[Sequence[Dict[str, str]]] = None,
        time_range: Optional[Dict[str, str]] = None,
    ) -> HTTPApiResponse:
        """
        Creates a federated search job. The results for the job may be requested using the job ID returned.
        :param search_type: "process" or "enriched_events" type
        :param query: a lucene query like "process_name:svchost.exe"
        :param fields: list of fields in the format ["process_username"]
        :param rows: integer
        :param org_keys: list of org keys in the format ["org1","org2"]
        :param customer_groups: list of org group names in the format ["group1","greoup2"]
        :param sort: in the format [{"field": "device_timestamp", "order": "asc"}]
        :param time_range: in the format {"start": "2023-01-18T18:34:04Z", "end": "2023-01-18T18:34:04Z"}
        :return: an ApiResponse with either a valid response or error information
        """
        _logger.debug(f"Query as submitted: {query}")
        params = {"query": query}

        if fields is not None:
            params.update(fields=fields)
        if rows is not None:
            params.update(rows=rows)
        if org_keys is not None:
            params.update(org_keys=org_keys)
        if customer_groups is not None:
            params.update(customer_groups=customer_groups)
        if rows is not None:
            params.update(rows=rows)
        if sort is not None:
            params.update(sort=sort)
        if time_range is not None:
            params.update(time_range=time_range)

        return self._send_request(
            f"/internal/federated_search/v1/{search_type}/search_jobs",
            method="POST",
            payload=params,
        )

    def get_search_status(self, search_type: str, job_id: str) -> HTTPApiResponse:
        """
        Get federated search status.
        :param search_type: "process" or "enriched_events" type
        :param job_id: federated search job ID "ab6f2371-7a88-4b8f-a868-8fbf19a66bcd"
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request(f"/internal/federated_search/v1/{search_type}/search_jobs/{job_id}")

    def get_search_stats(
        self, search_type: str, job_id: str, org_keys: Optional[Sequence[str]] = None
    ) -> HTTPApiResponse:
        """
        Get federated search status with org stats.
        :param search_type: "process" or "enriched_events" type
        :param job_id: federated search job ID "ab6f2371-7a88-4b8f-a868-8fbf19a66bcd"
        :param org_keys: list of org keys in the format ["org1","org2"]
        :return: an ApiResponse with either a valid response or error information
        """
        params = {}

        if org_keys is not None:
            params.update(org_keys=org_keys)

        return self._send_request(
            f"/internal/federated_search/v1/{search_type}/search_jobs/{job_id}/stats",
            params=params,
        )

    def get_search_results(
        self, search_type: str, job_id: str, org_keys: Optional[Sequence[str]] = None
    ) -> HTTPApiResponse:
        """
        Get federated search results with status and org stats.
        :param search_type: "process" or "enriched_events" type
        :param job_id: federated search job ID "ab6f2371-7a88-4b8f-a868-8fbf19a66bcd"
        :param org_keys: list of org keys in the format ["org1","org2"]
        :return: an ApiResponse with either a valid response or error information
        """
        params = {}

        if org_keys is not None:
            params.update(org_keys=org_keys)

        return self._send_request(
            f"/internal/federated_search/v1/{search_type}/search_jobs/{job_id}/results",
            params=params,
        )

    def change_status_search(self, search_type: str, job_id: str, status: str = "cancel") -> HTTPApiResponse:
        """
        Change status (cancel) of a federated search
        :param search_type: "process" or "enriched_events" type
        :param job_id: federated search job ID "ab6f2371-7a88-4b8f-a868-8fbf19a66bcd"
        :param status: change status (e.g. "cancel")
        :return: an ApiResponse with either a valid response or error information
        """

        return self._send_request(
            f"/internal/federated_search/v1/{search_type}/search_jobs/{job_id}/status",
            params={status: status},
            method="PUT",
        )

    def get_watchlist_evaluation_status(self, job_id: str) -> HTTPApiResponse:
        """
        Get federated search status.
        :param job_id: federated search job ID "ab6f2371-7a88-4b8f-a868-8fbf19a66bcd"
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request(f"/internal/federated_search/v1/processes/watchlist_evaluation/{job_id}")

    def get_watchlist_evaluation_stats(self, job_id: str, org_keys: Optional[Sequence[str]] = None) -> HTTPApiResponse:
        """
        Get federated search status with org stats.
        :param job_id: federated search job ID "ab6f2371-7a88-4b8f-a868-8fbf19a66bcd"
        :param org_keys: list of org keys in the format ["org1","org2"]
        :return: an ApiResponse with either a valid response or error information
        """
        params = {}

        if org_keys is not None:
            params.update(org_keys=org_keys)

        return self._send_request(
            f"/internal/federated_search/v1/processes/watchlist_evaluation/{job_id}/stats",
            params=params,
        )

    def create_watchlist_evaluation(
        self,
        report_id: str,
        feed_id: Optional[str] = None,
        time_range: Optional[Dict[str, str]] = None,
        watchlist_evaluation_job_properties: Optional[WatchlistEvaluationJobProperties] = None,
        org_keys: Optional[Sequence[str]] = None,
        customer_groups: Optional[Sequence[str]] = None,
    ) -> HTTPApiResponse:
        """
        Creates a federated search job. The results for the job may be requested using the job ID returned.
        :param report_id: the feed id to evaluate like "f7YrnvFITpG3aFruxqj23A-TH_Hunting_report_1"
        :param feed_id: the feed id to evaluate like "m32s8vagRxOr5u3PScSWA"
        :param org_keys: list of org keys in the format ["org1","org2"]
        :param customer_groups: list of org group names in the format ["group1","greoup2"]
        :param watchlist_evaluation_job_properties: properties of watchlist evaluation
        :param time_range: in the format {"start": "2023-01-18T18:34:04Z", "end": "2023-01-18T18:34:04Z"}
        :return: an ApiResponse with either a valid response or error information
        """
        params = {"report_id": report_id}

        if feed_id is not None:
            params.update(feed_id=feed_id)
        if org_keys is not None:
            params.update(org_keys=org_keys)
        if customer_groups is not None:
            params.update(customer_groups=customer_groups)
        if watchlist_evaluation_job_properties is not None:
            params.update(watchlist_evaluation_job_properties=watchlist_evaluation_job_properties)
        if time_range is not None:
            params.update(time_range=time_range)

        return self._send_request(
            "/internal/federated_search/v1/processes/watchlist_evaluation",
            method="POST",
            payload=params,
        )

    def change_status_watchlist_evaluation(self, job_id: str, status: str = "cancel") -> HTTPApiResponse:
        """
        Change status (cancel) of a federated search
        :param job_id: federated search job ID "ab6f2371-7a88-4b8f-a868-8fbf19a66bcd"
        :param status: change status (e.g. "cancel")
        :return: an ApiResponse with either a valid response or error information
        """

        return self._send_request(
            f"/internal/federated_search/v1/processes/watchlist_evaluation/{job_id}/status",
            params={status: status},
            method="PUT",
        )

    def get_csr(self, search_type: str, job_id: str, orgs: Optional[Sequence[str]] = None) -> HTTPApiResponse:
        """
        Get federated search results with status and org stats.
        :param search_type: "process" or "enriched_events" type
        :param job_id: federated search job ID "ab6f2371-7a88-4b8f-a868-8fbf19a66bcd"
        :param orgs: list of org keys in the format {"org1": "http..","org2": "http..."}
        :return: an ApiResponse with either a valid response or error information
        """
        params = {}

        if orgs is not None:
            params.update(orgs=orgs)

        return self._send_request(
            f"/internal/federated_search/v1/{search_type}/search_jobs/{job_id}/csr",
            params=params,
        )

    def upsert_customer_group(
        self,
        name: Optional[str] = None,
        org_keys: Optional[Sequence[str]] = None,
    ) -> HTTPApiResponse:
        """
        Update or create group of customer orgs. Name will be used as identifier
        :param name: customer group name like "group_key"
        :param org_keys: list of org keys in the format ["org1","org2"]
        :return: an ApiResponse with either a valid response or error information
        """

        params = {}

        if name is not None:
            params.update(name=name)
        if org_keys is not None:
            params.update(org_keys=org_keys)

        return self._send_request(
            "/internal/federated_search/v1/customer_groups",
            method="PUT",
            payload=params,
        )

    def get_customer_groups(
        self,
    ) -> HTTPApiResponse:
        """
        Get a list of customer groups.
        :return: an ApiResponse with either a valid response or error information
        """

        return self._send_request(
            "/internal/federated_search/v1/customer_groups",
        )

    def get_customer_group(self, name: str) -> HTTPApiResponse:
        """
        Get a customer groups details.
        :param name: customer group name like "group_key"
        :return: an ApiResponse with either a valid response or error information
        """

        return self._send_request(
            f"/internal/federated_search/v1/customer_groups/{name}",
        )

    def delete_customer_group(self, name: str) -> HTTPApiResponse:
        """
        Delete a customer groups.
        :param name: customer group name like "group_key"
        :return: an ApiResponse with either a valid response or error information
        """

        return self._send_request(
            f"/internal/federated_search/v1/customer_groups/{name}",
            method="DELETE",
        )
