import os
from functools import partial
from typing import List, Optional

from pscr_integration.test_lifecycle.active_session import is_k8s
from pscr_integration.api.api_request import HTTPApiRequest, HTTPApiResponse


class FeedApi:
    def __init__(self, privatelink_hostname: str, org_id: str):
        self.org_id = org_id
        privatelink_hostname = os.environ.get("FEED_MANAGER_API_PRIVATELINK_HOSTNAME", privatelink_hostname)
        if is_k8s() and not privatelink_hostname:
            privatelink_hostname = "psc-lc-feed-mgr"
        url_prefix = "http://" + privatelink_hostname + ":5000"
        self._make_request = partial(HTTPApiRequest, org_id=org_id, url_prefix=url_prefix)

    def _send_request(self, route: str, method: str = "GET", params: Optional[dict] = None) -> HTTPApiResponse:
        request = self._make_request()
        return request.send_request(route, method, params)

    def create_feed(self, feedinfo: dict, reports: Optional[List[dict]] = None) -> HTTPApiResponse:
        params = {"feedinfo": feedinfo}
        if reports is not None:
            params.update(reports=reports)
        return self._send_request("/feedmgr/v1/feed", method="POST", params=params)

    def delete_feed(self, feed_id: str) -> HTTPApiResponse:
        return self._send_request(f"/feedmgr/v1/feed/{feed_id}", method="DELETE")

    def get_feeds(self, include_public: Optional[bool] = None) -> HTTPApiResponse:
        return self._send_request(
            "/feedmgr/v1/feed",
            params={"include_public": include_public} if include_public else None,
        )

    def get_feed(self, feed_id: str) -> HTTPApiResponse:
        return self._send_request(f"/feedmgr/v1/feed/{feed_id}")

    def get_feedinfo(self, feed_id: str) -> HTTPApiResponse:
        return self._send_request(f"/feedmgr/v1/feed/{feed_id}/feedinfo")

    def update_feedinfo(self, feed_id: str, feedinfo: dict) -> HTTPApiResponse:
        return self._send_request(f"/feedmgr/v1/feed/{feed_id}/feedinfo", method="PUT", params=feedinfo)

    def get_reports(self, feed_id: str) -> HTTPApiResponse:
        return self._send_request(f"/feedmgr/v1/feed/{feed_id}/report")

    def get_report(self, feed_id: str, report_id: str) -> HTTPApiResponse:
        return self._send_request(f"/feedmgr/v1/feed/{feed_id}/report/{report_id}")

    def update_reports(self, feed_id: str, params: Optional[dict] = None) -> HTTPApiResponse:
        return self._send_request(f"/feedmgr/v1/feed/{feed_id}/report", method="POST", params=params)

    def update_report(self, feed_id: str, report_id: str, report: dict) -> HTTPApiResponse:
        return self._send_request(
            f"/feedmgr/v1/feed/{feed_id}/report/{report_id}",
            method="PUT",
            params=report,
        )
