import os
import logging
from functools import partial
from typing import Optional, Sequence

from pscr_integration.test_lifecycle.active_session import is_k8s
from pscr_integration.api.api_request import HTTPApiRequest, HTTPApiResponse
from pscr_integration.search.searcher_type import FacetParams

_logger = logging.getLogger(__name__)


class SearchAPI:
    def __init__(
        self,
        privatelink_hostname: str,
        org_id: str,
        device_id_start: Optional[int] = None,
        device_id_end: Optional[int] = None,
    ):
        self.org_id = org_id
        self.device_id_start = device_id_start
        self.device_id_end = device_id_end
        privatelink_hostname = os.environ.get("SEARCH_API_PRIVATELINK_HOSTNAME", privatelink_hostname)
        if is_k8s() and not privatelink_hostname:
            privatelink_hostname = "psc-lc-search-api"
        self._make_request = partial(
            HTTPApiRequest,
            org_id=org_id,
            url_prefix="http://" + privatelink_hostname + ":5003",
        )

    def _send_request(self, route: str, method: str = "GET", params: Optional[dict] = None) -> HTTPApiResponse:
        request = self._make_request()
        return request.send_request(route, method, params)

    def create_process_result(
        self,
        q: str,
        fl: Optional[Sequence[str]] = None,
        min_device_ts: Optional[str] = None,
        max_device_ts: Optional[str] = None,
        min_backend_ts: Optional[str] = None,
        max_backend_ts: Optional[str] = None,
        sort: Optional[str] = None,
        start: Optional[str] = None,
        rows: Optional[str] = None,
        fq: Optional[str] = None,
        cache: Optional[bool] = None,
        time_allowed: Optional[str] = None,
        facet: Optional[FacetParams] = None,
        suggest: Optional[str] = None,
        hl: Optional[str] = None,
        terms: Optional[str] = None,
        group: Optional[str] = None,
        all_fields: Optional[bool] = False,
        comprehensive_search: Optional[bool] = False,
    ) -> HTTPApiResponse:
        """
        This endpoint starts a search with the given query parameters.
        :param q: Solr query to run
        :param fl: list of fields in the format ["process_username"]
        :param min_device_ts: Start time for which cores to search based on device timestamp
        :param max_device_ts: End time for which cores to search based on device timestamp
        :param min_backend_ts: Start time for which cores to search based on backend timestamp
        :param max_backend_ts: End time for which cores to search based on backend timestamp
        :param sort: Solr sort
        :param start: Solr start for pagination
        :param rows: Solr rows for pagination
        :param fq: Solr filter query
        :param cache: Solr cache
        :param time_allowed: Solr timeAllowed
        :param facet: Solr facets
        :param suggest: Solr suggesters
        :param hl: Solr highlighters
        :param terms: Solr terms
        :param group: Solr grouping
        :param cb.all_fields: search all fields
        :param comprehensive_search comprehensive search internal parameter
        :return: an ApiResponse with either a valid response or error information
        """
        query = (
            f"device_id:[{self.device_id_start} TO {self.device_id_end}] AND ({q})"
            if all((self.device_id_end, self.device_id_start))
            else q
        )
        _logger.debug(f"Query as submitted: {query}")
        search_params = {
            "q": query,
            "cb.comprehensive_search": bool(comprehensive_search),
        }
        params = {"search_params": search_params}
        if min_device_ts is not None:
            search_params.update({"cb.min_device_timestamp": min_device_ts})
        if max_device_ts is not None:
            search_params.update({"cb.max_device_timestamp": max_device_ts})
        if min_backend_ts is not None:
            search_params.update({"cb.min_backend_timestamp": min_backend_ts})
        if max_backend_ts is not None:
            search_params.update({"cb.max_backend_timestamp": max_backend_ts})
        if sort is not None:
            search_params.update(sort=sort)
        if start is not None:
            search_params.update(start=start)
        if rows is not None:
            search_params.update(rows=rows)
        if fq is not None:
            search_params.update(fq=fq)
        if cache is not None:
            search_params.update(cache=cache)
        if time_allowed is not None:
            search_params.update(timeAllowed=time_allowed)
        if facet is not None:
            search_params.update(
                {
                    "facet": True,
                    "facet.field": facet.field,
                    "facet.mincount": facet.mincount,
                }
            )
        if suggest is not None:
            search_params.update(suggest=suggest)
        if hl is not None:
            search_params.update(hl=hl)
        if terms is not None:
            search_params.update(terms=terms)
        if group is not None:
            search_params.update(group=group)
        if fl is not None:
            params.update(fl=fl)
        if all_fields is not None:
            search_params.update({"cb.all_fields": all_fields})

        return self._send_request(
            f"/search/v1/orgs/{self.org_id}/processes/search_jobs",
            method="POST",
            params=params,
        )

    def get_process_result(
        self,
        query_id: str,
        start_row: Optional[str] = None,
        row_count: Optional[str] = None,
        format: Optional[str] = None,
    ) -> HTTPApiResponse:
        """
        This endpoint retrieves results from the given query_id.
        :param query_id: Query ID to find results for
        :param start_row: The start row for pagination over just this result set
        :param row_count: The row count for pagination over just this result set
        :param format: the format for the results, either json or csv
        :return: an ApiResponse with either a valid response or error information
        """
        url = (
            f"/search/v1/orgs/{self.org_id}/processes/search_jobs/{query_id}/results"
            f"?start={start_row if start_row is not None else 0}"
        )
        if row_count is not None:
            url += f"&rows={row_count}"
        if format is not None:
            url += f"&format={format}"

        return self._send_request(url)

    def get_process_results(self) -> HTTPApiResponse:
        """
        This endpoint retrieves a list of all query IDs, both for queries that are in progress and that have completed.
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request(f"/search/v1/orgs/{self.org_id}/processes/search_jobs")

    def delete_process_result(self, query_id: str) -> HTTPApiResponse:
        """
        This endpoint cancels a search with the given query_id, if that search has not already finished.
        :param query_id: Query ID to cancel
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request(
            f"/search/v1/orgs/{self.org_id}/processes/search_jobs/{query_id}",
            method="DELETE",
        )

    def evaluate_processes(self, watchlist_id: str, report_id: str, ioc_id: Optional[str]) -> HTTPApiResponse:
        """
        This endpoint evaluates historical feeds, tags documents and sends alerts as needed
        :param watchlist_id: the id for the watchlist to evaluate
        :param report_id: the id for the report to evaluate
        :param ioc_id: the id for the IOC to evaluate
        :return: an ApiResponse with either a valid response or error information
        """
        params = {"watchlist_id": watchlist_id, "report_id": report_id}
        if ioc_id is not None:
            params.update(ioc_id=ioc_id)

        return self._send_request(
            f"/search/v1/orgs/{self.org_id}/processes/watchlist_evaluation",
            method="POST",
            params=params,
        )

    def get_process_limits(self) -> HTTPApiResponse:
        """
        This endpoint returns the upper and lower limits of what online searchers cover.
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request(f"/search/v1/orgs/{self.org_id}/processes/limits")

    def get_process_result_status(self, query_id: str) -> HTTPApiResponse:
        """
        This endpoint retrieves the status of a search for the given query_id.
        :param query_id: Query ID to find status for
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request(f"/search/v1/orgs/{self.org_id}/processes/search_jobs/{query_id}")

    def get_process_suggestions(self, query: str, count: int = 10) -> HTTPApiResponse:
        """
        This endpoint gets suggestions for a process query
        :param query: the query to get suggestions for
        :param count: the number of suggestions to return
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request(
            f"/search/v1/orgs/{self.org_id}/processes/search_suggestions?suggest.q={query}&suggest.count={count}"
        )

    def get_process_validation(
        self,
        query: str,
        min_backend_timestamp: Optional[int] = None,
        max_backend_timestamp: Optional[int] = None,
    ) -> HTTPApiResponse:
        """
        This endpoint validates the given query with an indexer.
        :param query: Solr query to validate
        :param min_backend_timestamp: start time for query
        :param max_backend_timestamp: end time for query
        :return: an ApiResponse with either a valid response or error information
        """
        url = f"/search/v1/orgs/{self.org_id}/processes/search_validation?q={query}"
        if min_backend_timestamp is not None:
            url += f"&cb.min_backend_timestamp={min_backend_timestamp}"
        if max_backend_timestamp is not None:
            url += f"&cb.max_backend_timestamp={max_backend_timestamp}"

        return self._send_request(url)

    def get_events(
        self,
        q: str,
        cb_process_guid: str,
        min_backend_ts: Optional[str] = None,
        max_backend_ts: Optional[str] = None,
        sort: Optional[str] = None,
        start: Optional[str] = None,
        rows: Optional[str] = None,
        fq: Optional[str] = None,
        cache: Optional[bool] = None,
        time_allowed: Optional[str] = None,
        facet: Optional[FacetParams] = None,
        suggest: Optional[str] = None,
        hl: Optional[str] = None,
        terms: Optional[str] = None,
    ) -> HTTPApiResponse:
        """
        This endpoint searched for events with the given query parameters.
        :param q: Solr query to run (requires a process_guid term)
        :param cb_process_guid: the process guid to be used for internal query work
        :param min_backend_ts: Start time for which cores to search based on backend timestamp
        :param max_backend_ts: End time for which cores to search based on backend timestamp
        :param sort: Solr sort
        :param start: Solr start for pagination
        :param rows: Solr rows for pagination
        :param fq: Solr filter query
        :param cache: Solr cache
        :param time_allowed: Solr timeAllowed
        :param facet: Solr facets
        :param suggest: Solr suggesters
        :param hl: Solr highlighters
        :param terms: Solr terms
        :return: an ApiResponse with either a valid response or error information
        """
        search_params = {"cb.process_guid": cb_process_guid, "q": q}
        params = {"search_params": search_params}
        if min_backend_ts is not None:
            search_params.update({"cb.min_backend_timestamp": min_backend_ts})
        if max_backend_ts is not None:
            search_params.update({"cb.max_backend_timestamp": max_backend_ts})
        if sort is not None:
            search_params.update(sort=sort)
        if start is not None:
            search_params.update(start=start)
        if rows is not None:
            search_params.update(rows=rows)
        if fq is not None:
            search_params.update(fq=fq)
        if cache is not None:
            search_params.update(cache=cache)
        if time_allowed is not None:
            search_params.update(timeAllowed=time_allowed)
        if facet is not None:
            search_params.update(
                {
                    "facet": True,
                    "facet.field": facet.field,
                    "facet.mincount": facet.mincount,
                }
            )
        if suggest is not None:
            search_params.update(suggest=suggest)
        if hl is not None:
            search_params.update(hl=hl)
        if terms is not None:
            search_params.update(terms=terms)

        return self._send_request(
            f"/search/v1/orgs/{self.org_id}/events/_search",
            method="POST",
            params=params,
        )

    def get_event_suggestions(self, query: str, count: int = 10) -> HTTPApiResponse:
        """
        This endpoint gets suggestions for a events query
        :param query: the query to get suggestions for
        :param count: the number of suggestions to return
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request(
            f"/search/v1/orgs/{self.org_id}/events/search_suggestions?suggest.q={query}&suggest.count={count}"
        )

    def get_event_validation(
        self,
        query: str,
        min_backend_timestamp: Optional[int],
        max_backend_timestamp: Optional[int],
    ) -> HTTPApiResponse:
        url = f"/search/v1/orgs/{self.org_id}/events/search_validation?q={query}"
        if min_backend_timestamp is not None:
            url += f"&cb.min_backend_timestamp={min_backend_timestamp}"
        if max_backend_timestamp is not None:
            url += f"&cb.max_backend_timestamp={max_backend_timestamp}"

        return self._send_request(f"/search/v1/orgs/{self.org_id}/events_validation")

    def get_report_hits(self, process_guid: str) -> HTTPApiResponse:
        """
        This endpoint returns report hits information
        :param process_guid: the process guid to look for hits on
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request(f"/search/v1/orgs/{self.org_id}/report_hits&process_guid={process_guid}")

    def get_process_tree(self, process_guid: str, parent_guid: Optional[str]) -> HTTPApiResponse:
        """
        This endpoint returns a tree for a specific process.
        :param process_guid: the guid of the process
        :param parent_guid: (optional) the guid of the parent process
        :return: an ApiResponse with either a valid response or error information
        """
        url = f"/search/v1/orgs/{self.org_id}/processes/tree?process_guid={process_guid}"
        if parent_guid:
            url += f"&parent_guid={parent_guid}"

        return self._send_request(url)

    def get_process_summary(self, process_guid: str, parent_guid: Optional[str]) -> HTTPApiResponse:
        """
        This endpoint returns a summary for a specific process.
        :param process_guid: the guid of the process
        :param parent_guid: (optional) the guid of the parent process
        :return: an ApiResponse with either a valid response or error information
        """
        url = f"/search/v1/orgs/{self.org_id}/processes/summary?process_guid={process_guid}"
        if parent_guid is not None:
            url += f"&parent_guid={parent_guid}"

        return self._send_request(url)

    def get_process_query_ids(self) -> HTTPApiResponse:
        return self._send_request(f"/search/v1/orgs/{self.org_id}/processes/search_jobs")

    def create_enriched_events_result(
        self,
        q: str,
        fl: Optional[Sequence[str]] = None,
        min_device_ts: Optional[str] = None,
        max_device_ts: Optional[str] = None,
        min_backend_ts: Optional[str] = None,
        max_backend_ts: Optional[str] = None,
        sort: Optional[str] = None,
        start: Optional[str] = None,
        rows: Optional[str] = None,
        all_fields: Optional[bool] = False,
        fq: Optional[str] = None,
    ) -> HTTPApiResponse:
        """
        Create an enriched events search job. The results for the search job may be requested using the job ID returned.
        This route will not request facets.
        :param q: a lucene query like "process_name:svchost.exe"
        :param fl: list of fields in the format ["process_username"]
        :param rows: integer
        :param sort: in the format [{"field": "device_timestamp", "order": "asc"}]
        :param start: integer
        :param fq: Solr filter query
        :param cb.all_fields: search all fields
        :param min_device_ts: Start time for which cores to search based on device timestamp
        :param max_device_ts: End time for which cores to search based on device timestamp
        :param min_backend_ts: Start time for which cores to search based on backend timestamp
        :param max_backend_ts: End time for which cores to search based on backend timestamp
        :return: an ApiResponse with either a valid response or error information
        """
        query = (
            f"device_id:[{self.device_id_start} TO {self.device_id_end}] AND ({q})"
            if all((self.device_id_end, self.device_id_start))
            else q
        )
        _logger.debug(f"Query as submitted: {query}")
        search_params = {"q": query}
        params = {"search_params": search_params}

        if min_device_ts is not None:
            search_params.update({"cb.min_device_timestamp": min_device_ts})
        if max_device_ts is not None:
            search_params.update({"cb.max_device_timestamp": max_device_ts})
        if min_backend_ts is not None:
            search_params.update({"cb.min_backend_timestamp": min_backend_ts})
        if max_backend_ts is not None:
            search_params.update({"cb.max_backend_timestamp": max_backend_ts})
        if sort is not None:
            search_params.update(sort=sort)
        if start is not None:
            search_params.update(start=start)
        if rows is not None:
            search_params.update(rows=rows)
        if fq is not None:
            search_params.update(fq=fq)
        if fl is not None:
            search_params.update(fl=fl)
        if all_fields is not None:
            search_params.update({"cb.all_fields": all_fields})

        return self._send_request(
            f"/search/v1/orgs/{self.org_id}/enriched_events/search_jobs",
            method="POST",
            params=params,
        )

    def get_enriched_events_result(
        self,
        job_id: str,
        start_row: Optional[str] = None,
        row_count: Optional[str] = None,
    ) -> HTTPApiResponse:
        """
        This endpoint retrieves enriched results from the given query_id.
        :param job_id: Query ID to find results for
        :param start_row: The start row for pagination over just this result set
        :param row_count: The row count for pagination over just this result set
        :return: an ApiResponse with either a valid response or error information
        """
        url = (
            f"/search/v1/orgs/{self.org_id}/enriched_events/search_jobs/{job_id}/results"
            f"?start={start_row if start_row is not None else 0}"
        )
        if row_count is not None:
            url += f"&rows={row_count}"
        return self._send_request(url)
