import logging
from typing import Optional, Sequence, Dict

from pscr_integration.api.api_request import HTTPApiResponse
from pscr_integration.api.search_api import SearchAPI
from pscr_integration.util import get_sqs_job_id

_logger = logging.getLogger(__name__)


class SearchAPIV2(SearchAPI):
    def create_process_result_v2(
        self,
        query: str,
        criteria: Optional[Dict[str, Sequence[str]]] = None,
        exclusions: Optional[Dict[str, Sequence[str]]] = None,
        fields: Optional[Sequence[str]] = None,
        rows: Optional[int] = None,
        sort: Optional[Sequence[Dict[str, str]]] = None,
        start: Optional[int] = None,
        time_range: Optional[Dict[str, str]] = None,
        use_internal: bool = False,
    ) -> HTTPApiResponse:
        """
        Creates a process search job. The results for the search job may be requested using the job ID returned.
        This route will not request facets.
        :param query: a lucene query like "process_name:svchost.exe"
        :param criteria: in the format {"device_name": ["Win7x64"]}
        :param exclusions: in the format {"device_name": ["Win7x64"]}
        :param fields: list of fields in the format ["process_username"]
        :param rows: integer
        :param sort: in the format [{"field": "device_timestamp", "order": "asc"}]
        :param start: integer
        :param time_range: in the format {"start": "2020-01-18T18:34:04Z", "end": "2020-01-18T18:34:04Z"}
        :param use_internal: use internal to avoid auth middleware
        :return: an ApiResponse with either a valid response or error information
        """
        query = (
            f"device_id:[{self.device_id_start} TO {self.device_id_end}] AND ({query})"
            if all((self.device_id_end, self.device_id_start))
            else query
        )
        _logger.debug(f"Query as submitted: {query}")
        params = {"query": query}

        if criteria is not None:
            params.update(criteria=criteria)
        if exclusions is not None:
            params.update(exclusions=exclusions)
        if fields is not None:
            params.update(fields=fields)
        if rows is not None:
            params.update(rows=rows)
        if sort is not None:
            params.update(sort=sort)
        if start is not None:
            params.update(start=start)
        if time_range is not None:
            params.update(time_range=time_range)

        return self._send_request(
            f'{"/internal" if use_internal else ""}' f"/search/v2/orgs/{self.org_id}/processes/search_jobs",
            method="POST",
            params=params,
        )

    def get_process_result_v2(
        self,
        job_id: str,
        start_row: Optional[str] = None,
        row_count: Optional[str] = None,
        use_internal: bool = False,
    ) -> HTTPApiResponse:
        """
        This endpoint retrieves results from the given query_id.
        :param job_id: Query ID to find results for
        :param start_row: The start row for pagination over just this result set
        :param row_count: The row count for pagination over just this result set
        :param use_internal: use internal to avoid auth middleware
        :return: an ApiResponse with either a valid response or error information
        """
        job_id = get_sqs_job_id(job_id)

        url = (
            f'{"/internal" if use_internal else ""}'
            f"/search/v2/orgs/{self.org_id}/processes/search_jobs/{job_id}/results"
            f"?start={start_row if start_row is not None else 0}"
        )
        if row_count is not None:
            url += f"&rows={row_count}"
        return self._send_request(url)

    # pylint: disable=arguments-differ
    # we should've created a v2 for this
    def create_enriched_events_result(
        self,
        query: str,
        criteria: Optional[Dict[str, Sequence[str]]] = None,
        exclusions: Optional[Dict[str, Sequence[str]]] = None,
        fields: Optional[Sequence[str]] = None,
        rows: Optional[int] = None,
        sort: Optional[Sequence[Dict[str, str]]] = None,
        start: Optional[int] = None,
        time_range: Optional[Dict[str, str]] = None,
        use_internal: bool = False,
    ):
        """
        Create an enriched events search job. The results for the search job may be requested using the job ID returned.
        This route will not request facets.
        :param query: a lucene query like "process_name:svchost.exe"
        :param criteria: in the format {"device_name": ["Win7x64"]}
        :param exclusions: in the format {"device_name": ["Win7x64"]}
        :param fields: list of fields in the format ["process_username"]
        :param rows: integer
        :param sort: in the format [{"field": "device_timestamp", "order": "asc"}]
        :param start: integer
        :param time_range: in the format {"start": "2020-01-18T18:34:04Z", "end": "2020-01-18T18:34:04Z"}
        :param use_internal: use internal to avoid auth middleware
        :return: an ApiResponse with either a valid response or error information
        """
        params = {"query": query}
        if criteria is not None:
            params.update(criteria=criteria)
        if exclusions is not None:
            params.update(exclusions=exclusions)
        if fields is not None:
            params.update(fields=fields)
        if rows is not None:
            params.update(rows=rows)
        if sort is not None:
            params.update(sort=sort)
        if start is not None:
            params.update(start=start)
        if time_range is not None:
            params.update(time_range=time_range)

        return self._send_request(
            f'{"/internal" if use_internal else ""}' f"/search/v2/orgs/{self.org_id}/enriched_events/search_jobs",
            method="POST",
            params=params,
        )

    # pylint: disable=arguments-differ
    # we should've created a v2 for this
    def get_enriched_events_result(
        self,
        job_id: str,
        start_row: Optional[str] = None,
        row_count: Optional[str] = None,
        use_internal: bool = False,
    ) -> HTTPApiResponse:
        """
        This endpoint retrieves enriched results from the given query_id.
        :param job_id: Query ID to find results for
        :param start_row: The start row for pagination over just this result set
        :param row_count: The row count for pagination over just this result set
        :param use_internal: use internal to avoid auth middleware
        :return: an ApiResponse with either a valid response or error information
        """
        job_id = get_sqs_job_id(job_id)

        url = (
            f'{"/internal" if use_internal else ""}'
            f"/search/v2/orgs/{self.org_id}/enriched_events/search_jobs/{job_id}/results"
            f"?start={start_row if start_row is not None else 0}"
        )
        if row_count is not None:
            url += f"&rows={row_count}"
        return self._send_request(url)

    def create_process_detail_v2(self, process_guids):
        """
        Creates a process search job. The results for the search job may be requested using the job ID returned.
        This route will not request facets.
        :param process_guids: a list of process_guid. Maximum number is 100 process_guid.
        :return: an ApiResponse with either a valid response or error information
        """

        params = {"process_guids": process_guids}

        return self._send_request(
            f"/search/v2/orgs/{self.org_id}/processes/detail_jobs",
            method="POST",
            params=params,
        )

    def get_process_detail_v2(
        self,
        job_id: str,
        start_row: Optional[str] = None,
        row_count: Optional[str] = None,
        use_internal: bool = False,
    ) -> HTTPApiResponse:
        """
        This endpoint retrieves results from the given query_id.
        :param job_id: Query ID to find results for
        :param start_row: The start row for pagination over just this result set
        :param row_count: The row count for pagination over just this result set
        :param use_internal: use internal to avoid auth middleware
        :return: an ApiResponse with either a valid response or error information
        """
        job_id = get_sqs_job_id(job_id)
        url = (
            f'{"/internal" if use_internal else ""}'
            f"/search/v2/orgs/{self.org_id}/processes/detail_jobs/{job_id}/results"
            f"?start={start_row if start_row is not None else 0}"
        )
        if row_count is not None:
            url += f"&rows={row_count}"
        return self._send_request(url)

    def create_events_detail_result(self, event_ids):
        """
        Create an enriched events search job. The results for the search job may be requested using the job ID returned.
        This route will not request facets.
        :param event_ids: a list of event_id. Maximum number is 100 event_id.
        :return: an ApiResponse with either a valid response or error information
        """
        params = {"event_ids": event_ids}

        return self._send_request(
            f"/search/v2/orgs/{self.org_id}/enriched_events/detail_jobs",
            method="POST",
            params=params,
        )

    def get_events_detail_result(
        self,
        job_id: str,
        start_row: Optional[str] = None,
        row_count: Optional[str] = None,
        use_internal: bool = False,
    ) -> HTTPApiResponse:
        """
        This endpoint retrieves enriched results from the given query_id.
        :param job_id: Query ID to find results for
        :param start_row: The start row for pagination over just this result set
        :param row_count: The row count for pagination over just this result set
        :param use_internal: use internal to avoid auth middleware
        :return: an ApiResponse with either a valid response or error information
        """
        url = (
            f'{"/internal" if use_internal else ""}'
            f"/search/v2/orgs/{self.org_id}/enriched_events/detail_jobs/{job_id}/results"
            f"?start={start_row if start_row is not None else 0}"
        )
        if row_count is not None:
            url += f"&rows={row_count}"
        return self._send_request(url)

    def create_process_summary_v2(self, process_guid: str, parent_guid: Optional[str], use_internal: bool = False):
        """
        Creates a process summary job for a given process_guid (and parent_guid optionally).
        :param process_guid: the guid of the process
        :param parent_guid: (optional) the guid of the parent process
        :param use_internal: use internal to avoid auth middleware
        :return: an ApiResponse with either a valid response or error information
        """
        params = {"process_guid": process_guid}
        if parent_guid is not None:
            params.update(parent_guid=parent_guid)

        return self._send_request(
            f'{"/internal" if use_internal else ""}' f"/search/v2/orgs/{self.org_id}/processes/summary_jobs",
            method="POST",
            params=params,
        )

    def get_process_summary_v2(self, job_id: str, tree: bool = False, use_internal: bool = False) -> HTTPApiResponse:
        """
        This endpoint returns a summary for a specific process.
        :param job_id: Query ID to find results for
        :param tree: False - summary format (default), True - tree format
        :param use_internal: use internal to avoid auth middleware
        :return: an ApiResponse with either a valid response or error information
        """
        url = (
            f'{"/internal" if use_internal else ""}'
            f"/search/v2/orgs/{self.org_id}/processes/summary_jobs/{job_id}/results"
        )
        if tree:
            url += "?format=tree"
        return self._send_request(url)

    def create_complete_events_result(
        self,
        query: str,
        process_guid: str,
        use_internal: bool = False,
    ):
        """
        Create a complete events search job. The results for the search job may be requested using the job ID returned.
        This route will not request facets.
        :param query: a lucene query like "process_name:svchost.exe"
        :param process_guid: the guid of the process
        :return: an ApiResponse with either a valid response or error information
        """
        params = {"query": query, "process_guid": process_guid}

        return self._send_request(
            f'{"/internal" if use_internal else ""}' f"/search/v3/orgs/{self.org_id}/complete_events/search_jobs",
            method="POST",
            params=params,
        )

    def get_complete_events_result(
        self,
        job_id: str,
        start_row: Optional[str] = None,
        row_count: Optional[str] = None,
        use_internal: bool = False,
    ) -> HTTPApiResponse:
        """
        This endpoint retrieves complete results from the given query_id.
        :param job_id: Query ID to find results for
        :param start_row: The start row for pagination over just this result set
        :param row_count: The row count for pagination over just this result set
        :param use_internal: use internal to avoid auth middleware
        :return: an ApiResponse with either a valid response or error information
        """
        job_id = get_sqs_job_id(job_id)

        url = (
            f'{"/internal" if use_internal else ""}'
            f"/search/v3/orgs/{self.org_id}/complete_events/search_jobs/{job_id}/results"
            f"?start={start_row if start_row is not None else 0}"
        )
        if row_count is not None:
            url += f"&rows={row_count}"
        return self._send_request(url)

    def create_observations_result(
        self,
        query: str,
        criteria: Optional[Dict[str, Sequence[str]]] = None,
        exclusions: Optional[Dict[str, Sequence[str]]] = None,
        fields: Optional[Sequence[str]] = None,
        rows: Optional[int] = None,
        sort: Optional[Sequence[Dict[str, str]]] = None,
        start: Optional[int] = None,
        time_range: Optional[Dict[str, str]] = None,
        use_internal: bool = False,
    ) -> HTTPApiResponse:
        """
        Creates an observations search job. The results for the search job may be requested using the job ID returned.
        This route will not request facets.
        :param query: a lucene query like "process_name:svchost.exe"
        :param criteria: in the format {"device_name": ["Win7x64"]}
        :param exclusions: in the format {"device_name": ["Win7x64"]}
        :param fields: list of fields in the format ["process_username"]
        :param rows: integer
        :param sort: in the format [{"field": "device_timestamp", "order": "asc"}]
        :param start: integer
        :param time_range: in the format {"start": "2020-01-18T18:34:04Z", "end": "2020-01-18T18:34:04Z"}
        :param use_internal: use internal to avoid auth middleware
        :return: an ApiResponse with either a valid response or error information
        """
        query = (
            f"device_id:[{self.device_id_start} TO {self.device_id_end}] AND ({query})"
            if all((self.device_id_end, self.device_id_start))
            else query
        )
        _logger.debug(f"Query as submitted: {query}")
        params = {"query": query}

        if criteria is not None:
            params.update(criteria=criteria)
        if exclusions is not None:
            params.update(exclusions=exclusions)
        if fields is not None:
            params.update(fields=fields)
        if rows is not None:
            params.update(rows=rows)
        if sort is not None:
            params.update(sort=sort)
        if start is not None:
            params.update(start=start)
        if time_range is not None:
            params.update(time_range=time_range)

        return self._send_request(
            f'{"/internal" if use_internal else ""}' f"/search/v2/orgs/{self.org_id}/observations/search_jobs",
            method="POST",
            params=params,
        )

    def get_observations_result(
        self,
        job_id: str,
        start_row: Optional[str] = None,
        row_count: Optional[str] = None,
        use_internal: bool = False,
    ) -> HTTPApiResponse:
        """
        This endpoint retrieves results from the given query_id.
        :param job_id: Query ID to find results for
        :param start_row: The start row for pagination over just this result set
        :param row_count: The row count for pagination over just this result set
        :param use_internal: use internal to avoid auth middleware
        :return: an ApiResponse with either a valid response or error information
        """
        job_id = get_sqs_job_id(job_id)

        url = (
            f'{"/internal" if use_internal else ""}'
            f"/search/v2/orgs/{self.org_id}/observations/search_jobs/{job_id}/results"
            f"?start={start_row if start_row is not None else 0}"
        )
        if row_count is not None:
            url += f"&rows={row_count}"
        return self._send_request(url)

    def create_observations_detail_result(self, observation_ids):
        """
        Create an observations details job. The results for the search job may be requested using the job ID returned.
        This route will not request facets.
        :param observation_ids: a list of observation_id. Maximum number is 100 observation_ids.
        :return: an ApiResponse with either a valid response or error information
        """
        params = {"observation_ids": observation_ids}

        return self._send_request(
            f"/search/v2/orgs/{self.org_id}/observations/detail_jobs",
            method="POST",
            params=params,
        )

    def get_observations_detail_result(
        self,
        job_id: str,
        start_row: Optional[str] = None,
        row_count: Optional[str] = None,
        use_internal: bool = False,
    ) -> HTTPApiResponse:
        """
        This endpoint retrieves enriched results from the given query_id.
        :param job_id: Query ID to find results for
        :param start_row: The start row for pagination over just this result set
        :param row_count: The row count for pagination over just this result set
        :param use_internal: use internal to avoid auth middleware
        :return: an ApiResponse with either a valid response or error information
        """
        url = (
            f'{"/internal" if use_internal else ""}'
            f"/search/v2/orgs/{self.org_id}/observations/detail_jobs/{job_id}/results"
            f"?start={start_row if start_row is not None else 0}"
        )
        if row_count is not None:
            url += f"&rows={row_count}"
        return self._send_request(url)
