import os
from functools import partial
from typing import Optional

from pscr_integration.api.api_request import HTTPApiRequest, HTTPApiResponse
from pscr_integration.test_lifecycle.active_session import is_k8s


class SearchHistoryAPI:
    def __init__(self, privatelink_hostname: str, org_key: str):
        self.org_key = org_key
        privatelink_hostname = os.environ.get("SEARCH_API_PRIVATELINK_HOSTNAME", privatelink_hostname)
        if is_k8s() and not privatelink_hostname:
            privatelink_hostname = "psc-lc-search-api"
        url_prefix = "http://" + privatelink_hostname + ":5003"
        self._make_request = partial(HTTPApiRequest, org_id=org_key, url_prefix=url_prefix)

    def _send_request(
        self,
        route: str,
        method: str = "GET",
        params: Optional[dict] = None,
        user_id: Optional[int] = None,
        user_name: Optional[str] = None,
        org_key: Optional[str] = None,
    ) -> HTTPApiResponse:
        request = self._make_request()
        return request.send_request(
            route,
            method,
            params,
            user_id=user_id,
            user_name=user_name,
            org_key=org_key,
        )

    def get_searches(
        self,
        query: str,
        start: Optional[int] = None,
        rows: Optional[int] = None,
        user_id: Optional[int] = None,
        user_name: Optional[str] = None,
        type: Optional[str] = None,
    ) -> HTTPApiResponse:
        """
        Get information for past searches
        :param query: query to run
        :param start: start count used for pagination
        :param rows: max number of results to return, used for pagination
        :param query_type: "text" searching or a lucene "query_string"
        :param user_id: the user to get the favorited searches for
        :param user_name: the user name to get the favorited searches for
        :return: an ApiResponse with either a valid response or error information
        """

        url = (
            f"/search/history/v2/orgs/{self.org_key}/searches?"
            f"query={ query }&type={ type }&start={ start }&rows={ rows }"
        )

        return self._send_request(
            url,
            method="GET",
            user_id=user_id,
            user_name=user_name,
        )

    def get_favorited_searches(
        self,
        query: str,
        type: str,
        start: Optional[int] = None,
        rows: Optional[int] = None,
        user_id: Optional[int] = None,
        user_name: Optional[str] = None,
    ) -> HTTPApiResponse:
        """
        Get information for favorited searches
        :param query: query to run
        :param start: start count used for pagination
        :param rows: max number of results to return, used for pagination
        :param query_type: "text" searching or a lucene "query_string"
        :param user_id: the user to get the favorited searches for
        :param user_name: the user name to get the favorited searches for
        :return: an ApiResponse with either a valid response or error information
        """

        url = (
            f"/search/history/v2/orgs/{self.org_key}/favorites?"
            f"query={ query }&type={ type }&start={ start }&rows={ rows }"
        )
        return self._send_request(
            url,
            method="GET",
            user_id=user_id,
            user_name=user_name,
        )

    def save_search(
        self,
        search_id: str,
        name: str,
        favorited: Optional[bool] = False,
        search_time: Optional[str] = None,
        search_type: Optional[str] = None,
        user_id: Optional[int] = None,
        user_name: Optional[str] = None,
        search_params: Optional[dict] = None,
        is_public: Optional[bool] = False,
    ) -> HTTPApiResponse:
        """
        Save a search in search history
        :param search_id: the id for the search
        :param favorited: whether the search has been favorited
        :param search_time: the time the search was initiated
        :param search_type: what type of saved search it is, used to specify where the search was saved from
        :param user_id: the id for the user saving the search
        :param user_name: the name for the user saving the search
        :param search_params: the search parameters for the search
        :return: an ApiResponse with either a valid response or error information
        """
        params = {"org_key": self.org_key, "id": search_id, "name": name}

        if favorited:
            params.update({"favorited": favorited})
        if search_time is not None:
            params.update({"search_time": search_time})
        if search_type is not None:
            params.update({"type": search_type})
        if is_public:
            params.update({"is_public": is_public})

        if search_params is not None:
            params.update({"search_params": search_params})

        return self._send_request(
            f"/search/history/v2/orgs/{self.org_key}/favorites",
            method="POST",
            params=params,
            user_id=user_id,
            user_name=user_name,
        )

    def update_search(
        self,
        search_id: str,
        favorited: Optional[bool] = False,
        search_time: Optional[str] = None,
        search_type: Optional[str] = None,
        user_id: Optional[int] = None,
        user_name: Optional[str] = None,
        search_params: Optional[dict] = None,
        name: Optional[str] = None,
    ) -> HTTPApiResponse:
        """
        Update a search in search history
        :param search_id: the id for the search
        :param favorited: whether the search has been favorited
        :param search_time: the time the search was initiated
        :param search_type: what type of saved search it is, used to specify where the search was saved from
        :param user_id: the id for the user saving the search
        :param user_name: the name for the user saving the search
        :param search_params: the search parameters for the search
        :param name: the name of the search
        :return: an ApiResponse with either a valid response or error information
        """
        params = {"org_key": self.org_key, "id": search_id}

        if favorited:
            params.update({"favorited": favorited})
        if search_time is not None:
            params.update({"search_time": search_time})
        if search_type is not None:
            params.update({"type": search_type})
        if search_params is not None:
            params.update({"search_params": search_params})
        if name is not None:
            params.update({"name": name})

        return self._send_request(
            f"/search/history/v2/orgs/{self.org_key}/favorites/{search_id}",
            method="PUT",
            params=params,
            user_id=user_id,
            user_name=user_name,
        )

    def delete_search(
        self,
        search_id: str,
        user_id: Optional[int] = None,
        user_name: Optional[str] = None,
        org_key: Optional[str] = None,
    ):
        """
        Delete a past search
        :param search_id: id for the search to delete
        :return: an ApiResponse with either a valid response or error information
        :param user_id: the user id to use foer the request
        :param user_name: the user name to use for the request
        """
        return self._send_request(
            f"/search/history/v2/orgs/{org_key or self.org_key}/favorites/{search_id}",
            method="DELETE",
            user_id=user_id,
            user_name=user_name,
            org_key=org_key,
        )

    def delete_favorite(
        self,
        search_id: str,
        user_id: Optional[int] = None,
        user_name: Optional[str] = None,
        org_key: Optional[str] = None,
    ):
        """
        Delete a past search
        :param search_id: id for the search to delete
        :return: an ApiResponse with either a valid response or error information
        :param user_id: the user id to use foer the request
        :param user_name: the user name to use for the request
        """
        return self._send_request(
            f"/search/history/v2/orgs/{org_key or self.org_key}/favorites/{search_id}",
            method="DELETE",
            user_id=user_id,
            user_name=user_name,
            org_key=org_key,
        )
