import logging
import uuid
from contextlib import contextmanager
from copy import deepcopy
from functools import partial
from http import HTTPStatus
from typing import Dict

import pytest

from pscr_integration.api.federated_search_api import FederatedSearchAPI
from pscr_integration.api.feed_api import FeedApi
from pscr_integration.api.feed_api_v2 import FeedApiV2
from pscr_integration.api.feed_search_api import FeedSearchAPI
from pscr_integration.api.watchlist_api import WatchlistApi
from pscr_integration.api.watchlist_api_v2 import WatchlistApiV2
from pscr_integration.api.watchlist_api_v3 import WatchlistApiV3
from pscr_integration.search.searcher import Searcher
from pscr_integration.test_lifecycle.feed_data import feed_data, feed_data_4
from pscr_integration.util import _report_iocsv2
from pscr_integration.util import wait_until
from pscr_integration.lc_services import LcServices

_logger = logging.getLogger(__name__)


@pytest.mark.skipif(
    LcServices.target_service() not in LcServices.search_services(),
    reason=f"Test is irrelevant for f{LcServices.target_service()}.",
)
@pytest.mark.usefixtures("query_api")
@pytest.mark.smoke_test
class TestSearchAPI:
    def test_start_list_status_result_tree(self, query_api: Searcher):
        query_id = query_api.submit_query(query_string="*:*").response["query_id"]
        status_res = query_api.get_status(query_id)
        assert int(status_res.status_code) == HTTPStatus.OK
        assert int(status_res.response["contacted"]) > 0

        assert wait_until(
            lambda: f"{query_api.org_id}/{query_id}" in query_api.get_query_ids().response["query_ids"],
            timeout=15,
        )

        def _valid_results():
            try:
                resp = query_api.get_results(query_id).response
                assert resp["query_id"] == query_id
                assert resp["data"]
                return resp
            except AssertionError:
                return False

        res = wait_until(_valid_results)
        # Look for a "small" event
        # NOTE THAT THIS IS WRONG BECAUSE *_count do not return the true values for our test data.
        # TODO: Properly test /tree route
        el = next(e for e in res["data"] if "parent_guid" in e and int(e.get("childproc_count", 0)) == 0)
        _res = query_api.get_process_tree(process_guid=el["process_guid"], parent_guid=el["parent_guid"]).response

    def test_limits(self, query_api: Searcher):
        limit_res = query_api.get_process_limits()
        assert int(limit_res.status_code) == HTTPStatus.OK
        assert limit_res.response["time_bounds"]["lower"] and limit_res.response["time_bounds"]["upper"]

    def test_validate(self, query_api: Searcher):
        res = query_api.validate_query(query="*:*")
        assert int(res.status_code) == HTTPStatus.OK
        assert bool(res.response["valid"])


@pytest.mark.skipif(
    LcServices.target_service() not in LcServices.feed_services(),
    reason=f"Test is irrelevant for f{LcServices.target_service()}.",
)
@pytest.mark.usefixtures("query_api")
@pytest.mark.parametrize("feed_api", [FeedApi, FeedApiV2])
@pytest.mark.smoke_test
class TestFeedAPI:
    def test_get_feeds(self, config, load_org_1, feed_api):
        f_api, feed_id = setup_feed_api(config, feed_api, load_org_1, feed_data)
        rsp = f_api.get_feeds()
        assert len(rsp.response["results"]) == 2
        feed = next(f for f in rsp.response["results"] if f["id"] == feed_id)
        assert feed["name"] == feed_data["feedinfo"]["name"]
        assert feed["provider_url"] == feed_data["feedinfo"]["provider_url"]
        rsp = f_api.get_feed(feed_id)
        assert len(rsp.response["reports"]) == len(feed_data["reports"])
        assert rsp.response["feedinfo"]["category"] == feed_data["feedinfo"]["category"]
        rsp = f_api.get_feedinfo(feed_id)
        assert rsp.response["access"] == feed_data["feedinfo"]["access"]

    def test_update_feedinfo(self, config, load_org_1, feed_api):
        f_api, feed_id = setup_feed_api(config, feed_api, load_org_1, feed_data)
        update_feedinfo = {
            "name": "TempIntegrationTestFeed",
            "provider_url": "https://test.feeder.com",
            "summary": "Integration Test",
            "category": "CAT",
            "access": "private",
        }
        rsp = f_api.update_feedinfo(feed_id, update_feedinfo)
        assert rsp.status_code == 200
        rsp = f_api.get_feeds()
        feed = next(f for f in rsp.response["results"] if f["id"] == feed_id)
        assert feed["name"] == "TempIntegrationTestFeed"
        assert feed["provider_url"] == "https://test.feeder.com"

    def test_get_reports(self, config, load_org_1, feed_api):
        f_api, feed_id = setup_feed_api(config, feed_api, load_org_1, feed_data)
        # backend translates all iocs -> iocs_v2.  We need to match that
        reports_iocsv2 = [_report_iocsv2(report) for report in feed_data["reports"]]

        rsp = f_api.get_reports(feed_id)
        assert len(rsp.response["results"]) == len(feed_data["reports"])
        for report in rsp.response["results"]:
            assert report in reports_iocsv2
        for report in reports_iocsv2:
            assert report in rsp.response["results"]
        rsp = f_api.get_report(feed_id, "report2")
        report2 = {}
        for report in reports_iocsv2:
            if report["id"] == "report2":
                report2 = report
        assert report2 == rsp.response["report"]

    def test_update_reports(self, config, load_org_1, feed_api):
        f_api, feed_id = setup_feed_api(config, feed_api, load_org_1, feed_data)
        report1 = {}
        update_reports = deepcopy(feed_data["reports"])
        for report in update_reports:
            if report["id"] == "report1":
                report["description"] = "test_changed"
                report["timestamp"] = report["timestamp"] + 1
                report1 = report
            if report["id"] == "report2":
                report["severity"] = 2
                report["timestamp"] = 1
            if report["id"] == "report3":
                report["title"] = "changed_test"
                report["timestamp"] = 1
        rsp = f_api.update_reports(feed_id, {"reports": update_reports})
        assert rsp.status_code == 200
        rsp = f_api.get_reports(feed_id)
        for report in rsp.response["results"]:
            if report["id"] == "report1":
                assert report["description"] == "test_changed"
                report1["timestamp"] = report["timestamp"] + 1
            if report["id"] == "report2":
                assert report["severity"] == 2
            if report["id"] == "report3":
                assert report["title"] == "changed_test"
        report1["iocs"] = {"ipv4": ["2.1.168.192", "1.2.2.127"]}
        rsp = f_api.update_report(feed_id, "report1", report1)
        assert rsp.status_code == 200
        rsp = f_api.get_report(feed_id, "report1")
        assert _report_iocsv2(report1) == rsp.response["report"]

    def test_cannot_update_reports_with_invalid_queries(self, config, load_org_1, feed_api):
        f_api, feed_id = setup_feed_api(config, feed_api, load_org_1, feed_data)
        reports = [
            {
                "id": "report_with_invalid_query_ioc",
                "timestamp": 0,
                "link": "https://test.feed.com/report2",
                "title": "Report 2 Title",
                "description": "Report 2 Description",
                "severity": 7,
                "iocs": None,
                "iocs_v2": [
                    {
                        "id": "test_ioc_invalid",
                        "match_type": "query",
                        "field": None,
                        "values": ["bogus:bogus"],  # should fail due to invalid field name
                        "link": "https://test.feed.com/report_with_invalid_query_ioc/test_ioc_invalid",
                    }
                ],
            }
        ]
        rsp = f_api.update_reports(feed_id, {"reports": reports})
        assert rsp.status_code == 400

    def test_cannot_update_reports_with_value_search_queries(self, config, load_org_1, feed_api):
        f_api, feed_id = setup_feed_api(config, feed_api, load_org_1, feed_data)
        reports = [
            {
                "id": "report_with_value_query_ioc",
                "timestamp": 0,
                "link": "https://test.feed.com/report2",
                "title": "Report 2 Title",
                "description": "Report 2 Description",
                "severity": 7,
                "iocs": None,
                "iocs_v2": [
                    {
                        "id": "test_ioc_value_search",
                        "match_type": "query",
                        "field": None,
                        "values": ["something"],  # should fail because we don't allow values without fields in IOCs
                        "link": "https://test.feed.com/report_with_value_query_ioc/test_ioc_value_search",
                    }
                ],
            }
        ]
        rsp = f_api.update_reports(feed_id, {"reports": reports})
        assert rsp.status_code == 400


@pytest.mark.skipif(
    LcServices.target_service() not in LcServices.feed_services(),
    reason=f"Test is irrelevant for f{LcServices.target_service()}.",
)
@pytest.mark.parametrize("feed_api", [FeedApi, FeedApiV2])
@pytest.mark.smoke_test
class TestFeedSearchAPI:
    def test_health_check(self, config, feed_api, load_org_1):
        feed_search_api, feed_api, _ = setup_feed_search_api(config, FeedApiV2, load_org_1, feed_data)
        res = feed_search_api.health_check()
        assert res.status_code == 200

    def test_basic_feed_search(self, config, feed_api, load_org_1):
        feed_search_api, feed_api, feed_id = setup_feed_search_api(config, feed_api, load_org_1, feed_data)
        all_reports = feed_search_api.search(org_key=load_org_1["loader"].org_id, query=f"(feed_id:{feed_id})")
        assert all_reports.response["hits"]["total"] == len(feed_data["reports"])
        sorted_feed_data = sorted(feed_data["reports"], key=lambda report: report["id"])
        sorted_response = sorted(all_reports.response["hits"]["hits"], key=get_feed_id_from_search)
        for feed_data_report, feed_search_report in zip(sorted_feed_data, sorted_response):
            assert feeds_equal(feed_data_report, feed_search_report)

    def test_delete_feed_recognized(self, config, feed_api, load_org_1):
        feed_search_api, feed_api, _ = setup_feed_search_api(config, feed_api, load_org_1, feed_data)
        with make_test_feed(feed_api, feed_data_4, feed_search_api, load_org_1["loader"].org_id) as id_to_delete:
            feed_search_partial = partial(
                feed_search_api.search,
                org_key=load_org_1["loader"].org_id,
                query=f"(feed_id:{id_to_delete})",
            )

            assert wait_until(partial(feed_present_in_search, feed_search_partial, id_to_delete, 1))
            delete_res = feed_api.delete_feed(id_to_delete)
            assert delete_res.status_code == 204
            assert wait_until(partial(feed_not_present_in_search, feed_search_partial, id_to_delete))

    def test_modify_feed_recognized(self, config, feed_api, load_org_1):
        feed_search_api, feed_api, _ = setup_feed_search_api(config, feed_api, load_org_1, feed_data)
        with make_test_feed(feed_api, feed_data_4, feed_search_api, load_org_1["loader"].org_id) as id_to_modify:
            feed_search_partial = partial(
                feed_search_api.search,
                org_key=load_org_1["loader"].org_id,
                query=f"(feed_id:{id_to_modify})",
            )
            initial_res = wait_until(partial(feed_present_in_search, feed_search_partial, id_to_modify, 1))
            assert initial_res
            updated_feed_info = deepcopy(feed_data_4["feedinfo"])
            updated_feed_info["name"] = "ModifiedIntegrationTest"
            feed_api.update_feedinfo(id_to_modify, updated_feed_info)

            def updated_feed_recognized_in_search():
                res = feed_search_api.search(
                    org_key=load_org_1["loader"].org_id,
                    query=f"(feed_id:{id_to_modify})",
                )
                found_feeds = res.response["hits"]["hits"]
                relevant_feeds = [
                    f["_source"]["feed"] for f in found_feeds if f["_source"]["feed"]["feed_id"] == id_to_modify
                ]
                if res.status_code == 200:
                    for feed in relevant_feeds:
                        if feed["feed_name"] == updated_feed_info["name"]:
                            return res
                return None

            res = wait_until(updated_feed_recognized_in_search)
            found_feeds = res.response["hits"]["hits"]
            feed = next(f["_source"]["feed"] for f in found_feeds if f["_source"]["feed"]["feed_id"] == id_to_modify)
            assert feed["feed_name"] == updated_feed_info["name"]


@pytest.mark.skipif(
    LcServices.target_service() not in LcServices.federated_search_services(),
    reason=f"Test is irrelevant for f{LcServices.target_service()}.",
)
@pytest.mark.smoke_test
class TestFederatedSearchAPI:
    def test_health_check(self, config):
        res = FederatedSearchAPI(
            privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
        ).health_check()
        assert res.status_code == 200
        assert res.response["status"] == "UP"


@pytest.mark.skipif(
    LcServices.target_service() not in LcServices.feed_services(),
    reason=f"Test is irrelevant for f{LcServices.target_service()}.",
)
@pytest.mark.usefixtures("query_api")
@pytest.mark.parametrize("feed_api", [FeedApi, FeedApiV2])
@pytest.mark.parametrize("watchlist_api", [WatchlistApi, WatchlistApiV2, WatchlistApiV3])
@pytest.mark.smoke_test
class TestWatchlistAPI:
    def test_watchlists(self, config, load_org_1, feed_api, watchlist_api):
        feed_id = setup_feed_api(config, feed_api, load_org_1, feed_data)[1]
        watchlist2_name = feed_id + str(uuid.uuid4())[:8]
        w_api, watchlist_id = setup_watchlist_api(watchlist_api, config, load_org_1)

        # Create a new watchlist so we don't mess up the original watchlist used by other tests
        watchlist2 = w_api.create_feed_watchlist(feed_id, watchlist2_name)
        assert watchlist2.status_code == 200
        watchlist_id2 = w_api.classifier_watchlist_to_watchlist(watchlist2.response).id
        assert watchlist_id != watchlist_id2

        rsp = w_api.get_watchlists()
        results = [w_api.classifier_watchlist_to_watchlist(result) for result in rsp.response["results"]]
        expected_watchlists = {
            watchlist_id: {
                "name": f"{feed_id} Watchlist load_org_1",
                "description": f"Watchlist for feed {feed_id}",
                "watchlist_id": watchlist_id,
                "classifier_key": "feed_id",
                "classifier_value": feed_id,
                "tags_enabled": True,
                "alerts_enabled": True,
            },
            watchlist_id2: {
                "name": f"{feed_id} Watchlist {watchlist2_name}",
                "description": f"Watchlist for feed {feed_id}",
                "watchlist_id": watchlist_id2,
                "classifier_key": "feed_id",
                "classifier_value": feed_id,
                "tags_enabled": True,
                "alerts_enabled": False,
            },
        }

        assert len(results) >= len(expected_watchlists.values())

        for result in results:
            if result.id in expected_watchlists:
                expected = expected_watchlists[result.id]
                assert expected["name"] == result.name
                assert expected["description"] == result.description
                assert expected["watchlist_id"] == result.id
                assert expected["classifier_key"] == result.classifier_key
                assert expected["classifier_value"] == result.classifier_value
                assert expected["tags_enabled"] == result.tags_enabled
                assert expected["alerts_enabled"] == result.alerts_enabled

        watchlist_rsp = w_api.get_watchlist(watchlist_id2)
        watchlist = w_api.classifier_watchlist_to_watchlist(watchlist_rsp.response)
        assert watchlist.classifier_value == feed_id

        rsp = w_api.delete_watchlist(watchlist_id2)
        assert rsp.status_code == 204

        rsp = w_api.get_watchlist(watchlist_id2)
        assert rsp.status_code == 404

        watchlist_id2 = w_api.classifier_watchlist_to_watchlist(w_api.create_feed_watchlist(feed_id).response).id
        assert watchlist_id != watchlist_id2

        watchlist = w_api.classifier_watchlist_to_watchlist(w_api.get_watchlist(watchlist_id2).response)
        assert watchlist.classifier_value == feed_id
        assert watchlist.id == watchlist_id2

        rsp = w_api.delete_watchlist(watchlist_id2)
        assert rsp.status_code == 204

        clean_up_watchlist_api(watchlist_api, config, load_org_1)

    def test_alerts(self, config, load_org_1, watchlist_api, feed_api):
        w_api, watchlist_id = setup_watchlist_api(watchlist_api, config, load_org_1)

        rsp = w_api.get_watchlist_alert_status(watchlist_id)
        assert rsp.response["alert"] is True
        rsp = w_api.remove_watchlist_alert_status(watchlist_id)
        assert rsp.status_code == 204
        rsp = w_api.get_watchlist_alert_status(watchlist_id)
        assert rsp.response["alert"] is False
        rsp = w_api.add_watchlist_alert(watchlist_id)
        assert rsp.response["alert"] is True

        clean_up_watchlist_api(watchlist_api, config, load_org_1)

    def test_report_ignore(self, config, load_org_1, feed_api, watchlist_api):
        feed_id = setup_feed_api(config, feed_api, load_org_1, feed_data)[1]
        w_api, _ = setup_watchlist_api(watchlist_api, config, load_org_1)

        report1_id = f"{feed_id}-report1"
        report2_id = f"{feed_id}-report2"

        rsp = w_api.get_report_ignore(report1_id)
        assert rsp.response["ignored"] is False
        rsp = w_api.get_report_ioc_ignore(report2_id, "test_ioc")
        assert rsp.response["ignored"] is False
        rsp = w_api.add_report_ignore(report1_id)
        assert rsp.response["ignored"] is True
        rsp = w_api.add_report_ioc_ignore(report2_id, "test_ioc")
        assert rsp.response["ignored"] is True
        rsp = w_api.get_report_ignore(report1_id)
        assert rsp.response["ignored"] is True
        rsp = w_api.get_report_ioc_ignore(report2_id, "test_ioc")
        assert rsp.response["ignored"] is True
        rsp = w_api.remove_report_ignore(report1_id)
        assert rsp.status_code == 204
        rsp = w_api.remove_report_ioc_ignore(report2_id, "test_ioc")
        assert rsp.status_code == 204

        clean_up_watchlist_api(watchlist_api, config, load_org_1)

    def test_feed_severity_overrides(self, config, load_org_1, feed_api, watchlist_api):
        feed_id = setup_feed_api(config, feed_api, load_org_1, feed_data)[1]
        w_api, _ = setup_watchlist_api(watchlist_api, config, load_org_1)

        report1_id = f"{feed_id}-report1"
        report2_id = f"{feed_id}-report2"

        rsp = w_api.get_severity_overrides()
        assert rsp.response["results"] == []
        rsp = w_api.get_report_severity_override(report1_id)
        assert rsp.status_code == 404
        rsp = w_api.set_report_severity_override(report2_id, 9)
        assert rsp.response["severity"] == 9
        rsp = w_api.get_severity_overrides()
        assert rsp.response["results"] == [{"report_id": report2_id, "severity": 9}]
        rsp = w_api.get_report_severity_override(report2_id)
        assert rsp.response["severity"] == 9
        rsp = w_api.remove_report_severity_override(report2_id)
        assert rsp.status_code == 204

        clean_up_watchlist_api(watchlist_api, config, load_org_1)

    def test_watchlist_telemetry(self, config, load_org_1, feed_api, watchlist_api):
        feed_id = setup_feed_api(config, feed_api, load_org_1, feed_data)[1]
        w_api, watchlist_id = setup_watchlist_api(watchlist_api, config, load_org_1)

        # Create a new watchlist so we don't mess up the original watchlist used by other tests
        watchlist2_rsp = w_api.create_feed_watchlist(feed_id, "telemetry_test" + str(uuid.uuid4().bytes))
        assert watchlist2_rsp.status_code == 200
        watchlist2 = w_api.classifier_watchlist_to_watchlist(watchlist2_rsp.response)

        telemetries_rsp = w_api.get_watchlist_telemetry(
            watchlist_ids=[watchlist_id, watchlist2.id], intervals=[1440, 10080, 43200]
        )
        assert telemetries_rsp.status_code == 200
        assert len(telemetries_rsp.response["telemetry"]) == 6

        telemetries_rsp = w_api.get_watchlist_telemetry(watchlist_ids=[watchlist_id])
        assert telemetries_rsp.status_code == 200
        assert len(telemetries_rsp.response["telemetry"]) == 1

        assert w_api.delete_watchlist(watchlist2.id).status_code == 204

        clean_up_watchlist_api(watchlist_api, config, load_org_1)


@pytest.mark.skipif(
    LcServices.target_service() not in LcServices.org_retention_services(),
    reason=f"Test is irrelevant for f{LcServices.target_service()}.",
)
@pytest.mark.usefixtures("org_retention_api")
class TestRetentionAPI:
    def test_org_retentions(self, org_retention_api: Dict):
        api = org_retention_api["api"]

        for org, retentions in org_retention_api["retentions"].items():
            for retention_class, retention_days in retentions.items():
                assert retention_days == api.get_org_retention(org, retention_class)

    def test_internal_get_all_retentions(self, org_retention_api: Dict):
        all_retentions = org_retention_api["api"].get_internal_all_retentions()

        for retention in all_retentions:
            org = retention["org"]
            retention_class = retention["retention_class"]
            retention_days = retention["retention_days"]

            #  Only check orgs created by this test suite, there may be others if re-run in same environment
            #  Re-evaluate if we add a delete route for org retentions,
            #  when we do that the fixture should clean them up.
            if org in org_retention_api["retentions"].keys():
                assert retention_days == org_retention_api["retentions"][org][retention_class]

    def test_internal_get_retention_defaults(self, org_retention_api: Dict):
        defaults = org_retention_api["api"].get_internal_retention_defaults()

        assert "default" in defaults
        assert "longterm" in defaults


@pytest.mark.skipif(
    LcServices.target_service() not in LcServices.search_history_services(),
    reason=f"Test is irrelevant for f{LcServices.target_service()}.",
)
@pytest.mark.usefixtures("search_history_api")
class TestSearchHistory:
    def test_get_searches_for_org(self, load_org_1: Dict, search_history_api: Dict):
        api = search_history_api["api"]

        searches = api.get_searches(
            query="%2A%3A%2A",
            rows=100,
            start=0,
            user_id=120214,
            user_name="test@example.com",
            type="PROCESSES",
        )

        assert searches.status_code == 200
        searches = searches.response
        assert len(searches) == 2
        for saved_search in load_org_1["saved_searches"]:
            assert saved_search["favorited"] or saved_search["id"] in map(lambda x: x["id"], searches)

    def test_get_favorited_searches_for_org(self, load_org_1: Dict, search_history_api: Dict):
        api = search_history_api["api"]

        expected_search = {}
        for search in load_org_1["saved_searches"]:
            if search["favorited"]:
                expected_search = search
                break

        searches = api.get_favorited_searches(
            query="%2A%3A%2A",
            type=expected_search["type"],
            rows=100,
            start=0,
            user_id=120214,
            user_name="test@example.com",
        )
        assert searches.status_code == 200
        searches = searches.response
        assert len(searches) == 2
        assert searches[0]["favorited"] or searches[0]["id"] == expected_search["id"]
        assert searches[0]["favorited"] == expected_search["favorited"]

    def test_update_search(self, load_org_1: Dict, search_history_api: Dict):
        api = search_history_api["api"]

        search_id = str(uuid.uuid4())
        search = {
            "id": search_id,
            "org_key": load_org_1["loader"].org_id,
            "search_time": "2008-09-15T17:53:00Z",
            "type": "ENRICHED_EVENTS",
            "user_id": 120214,
            "user_name": "test@example.com",
            "search_params": {
                "query": "*:*",
                "criteria": {},
                "exclusions": {},
                "start": 0,
                "time_range": {},
                "fields": None,
            },
            "name": "name",
        }

        saved_search = api.save_search(
            search_id=search["id"],
            name=search["name"],
            search_time=search["search_time"],
            search_type=search["type"],
            user_id=search["user_id"],
            user_name=search["user_name"],
            search_params=search["search_params"],
        )

        assert saved_search.status_code == 200
        saved_search = saved_search.response

        saved_search["name"] = "newname"
        actual_saved_search = api.update_search(
            search_id=saved_search["id"],
            name=saved_search["name"],
            search_time=saved_search["search_time"],
            search_type=saved_search["type"],
            user_id=saved_search["user_id"],
            user_name=saved_search["user_name"],
            search_params=saved_search["search_params"],
        )

        search_id = actual_saved_search.response["id"]
        assert actual_saved_search.status_code == 200
        actual_saved_search = actual_saved_search.response
        assert saved_search["name"] == actual_saved_search["name"]

        assert api.delete_search(search_id).status_code == 204

    def test_delete_public_search_as_owner(self, load_org_1: Dict, search_history_api: Dict):
        api = search_history_api["api"]

        search_id = str(uuid.uuid4())
        search = {
            "id": search_id,
            "favorited": True,
            "is_public": True,
            "org_key": load_org_1["loader"].org_id,
            "search_time": "2008-09-15T17:53:00Z",
            "type": "ENRICHED_EVENTS",
            "user_id": 120214,
            "user_name": "test@example.com",
            "search_params": {
                "query": "*:*",
                "criteria": {},
                "exclusions": {},
                "start": 0,
                "time_range": {},
                "fields": None,
            },
            "name": "name",
        }

        saved_search = api.save_search(
            search_id=search["id"],
            name=search["name"],
            search_time=search["search_time"],
            search_type=search["type"],
            user_id=search["user_id"],
            user_name=search["user_name"],
            search_params=search["search_params"],
            is_public=search["is_public"],
        )
        assert saved_search.status_code == 200

        search_id = saved_search.response["id"]
        delete_response = api.delete_search(search_id)

        assert delete_response.status_code == 204

    def test_delete_private_search_as_owner(self, load_org_1: Dict, search_history_api: Dict):
        api = search_history_api["api"]

        search_id = str(uuid.uuid4())
        search = {
            "id": search_id,
            "favorited": True,
            "is_public": False,
            "org_key": load_org_1["loader"].org_id,
            "search_time": "2008-09-15T17:53:00Z",
            "type": "ENRICHED_EVENTS",
            "user_id": 120214,
            "user_name": "test@example.com",
            "search_params": {
                "query": "*:*",
                "criteria": {},
                "exclusions": {},
                "start": 0,
                "time_range": {},
                "fields": None,
            },
            "name": "name",
        }

        saved_search = api.save_search(
            search_id=search["id"],
            name=search["name"],
            favorited=search["favorited"],
            search_time=search["search_time"],
            search_type=search["type"],
            user_id=search["user_id"],
            user_name=search["user_name"],
            search_params=search["search_params"],
            is_public=search["is_public"],
        )
        assert saved_search.status_code == 200
        search_id = saved_search.response["id"]
        user_id = saved_search.response["user_id"]

        delete_response = api.delete_search(search_id, user_id)

        assert delete_response.status_code == 204

    def test_delete_public_search_as_different_user_in_same_org(self, load_org_1: Dict, search_history_api: Dict):
        api = search_history_api["api"]

        search_id = str(uuid.uuid4())
        search = {
            "id": search_id,
            "favorited": True,
            "is_public": True,
            "org_key": load_org_1["loader"].org_id,
            "search_time": "2008-09-15T17:53:00Z",
            "type": "ENRICHED_EVENTS",
            "user_id": 120214,
            "user_name": "test@example.com",
            "search_params": {
                "query": "*:*",
                "criteria": {},
                "exclusions": {},
                "start": 0,
                "time_range": {},
                "fields": None,
            },
            "name": "name",
        }

        saved_search = api.save_search(
            search_id=search["id"],
            name=search["name"],
            favorited=search["favorited"],
            search_time=search["search_time"],
            search_type=search["type"],
            user_id=search["user_id"],
            user_name=search["user_name"],
            search_params=search["search_params"],
            is_public=search["is_public"],
        )
        assert saved_search.status_code == 200

        search_id = saved_search.response["id"]
        delete_response = api.delete_search(search_id, user_id=121212, org_key=search["org_key"])
        assert delete_response.status_code == 204

    def test_delete_public_search_in_different_org(self, load_org_1: Dict, search_history_api: Dict):
        api = search_history_api["api"]

        search_id = str(uuid.uuid4())
        search = {
            "id": search_id,
            "favorited": True,
            "is_public": True,
            "org_key": load_org_1["loader"].org_id,
            "search_time": "2008-09-15T17:53:00Z",
            "type": "ENRICHED_EVENTS",
            "user_id": 120214,
            "user_name": "test@example.com",
            "search_params": {
                "query": "*:*",
                "criteria": {},
                "exclusions": {},
                "start": 0,
                "time_range": {},
                "fields": None,
            },
            "name": "name",
        }

        saved_search = api.save_search(
            search_id=search["id"],
            name=search["name"],
            favorited=search["favorited"],
            search_time=search["search_time"],
            search_type=search["type"],
            user_id=search["user_id"],
            user_name=search["user_name"],
            search_params=search["search_params"],
            is_public=search["is_public"],
        )
        assert saved_search.status_code == 200

        search_id = saved_search.response["id"]

        delete_response = api.delete_search(search_id, org_key="R76KWHRN")
        assert delete_response.status_code == 404

    def test_delete_private_search_as_different_user_in_same_org(self, load_org_1: Dict, search_history_api: Dict):
        api = search_history_api["api"]

        search_id = str(uuid.uuid4())
        search = {
            "id": search_id,
            "favorited": True,
            "is_public": False,
            "org_key": load_org_1["loader"].org_id,
            "search_time": "2008-09-15T17:53:00Z",
            "type": "ENRICHED_EVENTS",
            "user_id": 120214,
            "user_name": "test@example.com",
            "search_params": {
                "query": "*:*",
                "criteria": {},
                "exclusions": {},
                "start": 0,
                "time_range": {},
                "fields": None,
            },
            "name": "name",
        }

        saved_search = api.save_search(
            search_id=search["id"],
            name=search["name"],
            favorited=search["favorited"],
            search_time=search["search_time"],
            search_type=search["type"],
            user_id=search["user_id"],
            user_name=search["user_name"],
            search_params=search["search_params"],
            is_public=search["is_public"],
        )
        assert saved_search.status_code == 200

        search_id = saved_search.response["id"]

        delete_response = api.delete_search(search_id, user_id=111111, org_key=saved_search.response["org_key"])

        assert delete_response.status_code == 404


def setup_feed_api(config, feed_api, load_org_1, feeds):
    f_api = feed_api(
        privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
        org_id=load_org_1["loader"].org_id,
    )
    feed_id = load_org_1["feed_id"]

    f_api.update_feedinfo(feed_id, feeds["feedinfo"])
    f_api.update_reports(feed_id, {"reports": feeds["reports"]})

    return f_api, feed_id


def setup_feed_search_api(config, feed_api, load_org_1, feeds):
    f_api, f_id = setup_feed_api(config, feed_api, load_org_1, feeds)
    f_search_api = FeedSearchAPI(
        privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
        org_id=load_org_1["loader"].org_id,
    )
    return f_search_api, f_api, f_id


def setup_watchlist_api(watchlist_api, config, load_org_1):
    w_api = watchlist_api(
        privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
        org_id=load_org_1["loader"].org_id,
    )
    watchlist_id = load_org_1["watchlist_id"]

    return w_api, watchlist_id


def clean_up_watchlist_api(watchlist_api, config, load_org_1):
    w_api, watchlist_id = setup_watchlist_api(watchlist_api, config, load_org_1)
    feed_id = load_org_1["feed_id"]

    w_api.add_watchlist_alert(watchlist_id)
    w_api.remove_report_ignore(f"{feed_id}-report1")
    w_api.remove_report_ioc_ignore(f"{feed_id}-report2", "test_ioc")
    w_api.remove_report_severity_override(f"{feed_id}-report2")


@contextmanager
def make_test_feed(feed_api, feeds, feed_search_api, org_key):
    feed = feed_api.create_feed(feedinfo=feeds["feedinfo"], reports=feeds["reports"])
    feed_search_partial = partial(
        feed_search_api.search,
        org_key=org_key,
        query=f"(feed_id:{feed.response['id']})",
    )
    tmp = wait_until(
        partial(
            feed_present_in_search,
            feed_search_partial,
            feed.response["id"],
            len(feeds["reports"]),
        ),
        timeout=30,
    )
    if not tmp:
        raise Exception("Create feed time out")
    try:
        yield feed.response["id"]
    except Exception as e:
        raise Exception from e
    finally:
        feed_api.delete_feed(feed.response["id"])
        tmp = wait_until(
            partial(feed_not_present_in_search, feed_search_partial, feed.response["id"]),
            timeout=30,
        )


def len_change(feed_search_api, initial_len, load_org_1):
    res = feed_search_api.search(org_key=load_org_1["loader"].org_id, query="(*:*)")
    if len(res.response["hits"]["hits"]) != res.response["hits"]["total"]:
        raise Exception(
            f"Response {len(res.response['hits']['hits'])} items long, but response specifies "
            f"{res.response['hits']['total']} items"
        )
    if res.response["hits"]["total"] != initial_len:
        return res
    if res.status_code != 200:
        raise Exception(f"BAD STATUS CODE {res.status_code}")
    return None


def len_change_feed(feed_api, initial_len):
    all_reports = get_all_reports(feed_api)
    if len(all_reports) != initial_len:
        return all_reports
    return None


def feed_present_in_search(feed_search_api_partial, feed_id_to_check, expected_number_of_reports):
    res = feed_search_api_partial()
    found_feed_reports = res.response["hits"]["hits"]
    found_reports_num = 0
    for f in found_feed_reports:
        if f["_source"]["feed"]["feed_id"] == feed_id_to_check:
            found_reports_num += 1
        if found_reports_num == expected_number_of_reports:
            return found_feed_reports
    return None


def feed_not_present_in_search(feed_search_api_partial, feed_id_to_check):
    res = feed_search_api_partial()
    found_feed_reports = res.response["hits"]["hits"]
    for f in found_feed_reports:
        if f["_source"]["feed"]["feed_id"] == feed_id_to_check:
            return False
    return True


def get_all_reports(feed_api):
    accum = []
    for feed in feed_api.get_feeds().response["results"]:
        a = feed_api.get_reports(feed["id"]).response["results"]
        accum += a
    return accum


def gen_report_from_number(num):
    return {
        "id": f"report{num}",
        "timestamp": 0,
        "link": f"https://test.feed.com/report{num}",
        "title": f"Report {num} Title",
        "description": f"Report {num} Description",
        "severity": 8,
        "iocs": None,
        "iocs_v2": [
            {
                "id": "test_ioc_2",
                "match_type": "query",
                "field": "123",
                "values": ["process_name:explorer*"],
                "link": "https://test.feed.com/report5/test_ioc",
            }
        ],
        "iocs_total_count": 1,
        "visibility": None,
        "tags": None,
    }


def gen_unique_chars():
    return str(uuid.uuid4())[:8]


def feeds_equal(from_feed_data, from_feed_search):
    src = from_feed_search["_source"]
    return src["id"].split("-")[1] == from_feed_data["id"]


def get_feed_id_from_search(feed_search_report):
    return feed_search_report["_id"].split("-")[1]
