import logging
from functools import partial
from typing import Iterable, Tuple, List

import pytest

from pscr_integration.search.searcher import Searcher
from pscr_integration.util import wait_until
from pscr_integration.lc_services import LcServices

_logger = logging.getLogger(__name__)


@pytest.mark.skipif(
    LcServices.target_service() not in LcServices.search_services(),
    reason=f"Test is irrelevant for f{LcServices.target_service()}.",
)
@pytest.mark.usefixtures("query_api")
@pytest.mark.smoke_test
class TestQueries:
    def test_child_parent(self, query_api: Searcher):
        query = partial(_query, query_api=query_api, expect_data=True)
        all_results = query()
        parent_proc, child_proc = _get_child_parent_process(all_results)
        assert child_proc
        assert parent_proc

        # Basic query
        results_pid = query(
            q=f'process_guid:{child_proc["process_guid"]} AND ' f'parent_guid:{parent_proc["process_guid"]}'
        )
        assert results_pid and child_proc in results_pid

        # Subquery
        subquery = f'process_guid:{child_proc["process_guid"]} OR process_guid:{parent_proc["process_guid"]}'
        result_subquery = query(q=subquery)
        assert result_subquery
        result_sub_child = query(q=f'({subquery}) AND parent_guid:{child_proc["parent_guid"]}')
        assert result_sub_child and child_proc in result_sub_child
        assert all(result in result_subquery for result in result_sub_child)

    def test_lucene_features(self, query_api: Searcher):
        # Wildcard
        query = partial(_query, query_api=query_api, expect_data=True)
        proc = next(r for r in query() if r.get("device_name"))
        result_wildcard = query(q=f'process_guid:{proc["process_guid"][:-4]}*')
        assert result_wildcard and proc in result_wildcard
        # TODO: Range never return for any fields
        # Range
        # ts = parse(proc["device_timestamp"])
        # one_sec = timedelta(seconds=1)
        # result_range = query(q=f'device_timestamp:[{"".join([str(ts - one_sec)[:-9], "Z"])} '
        #                        f'TO {"".join([str(ts + one_sec)[:-9], "Z"])}]')
        # assert len(result_range) and proc in result_range


def _query(query_api: Searcher, q: str = "*:*", expect_data=True) -> List[dict]:
    query_id = query_api.submit_query(query_string=q).response["query_id"]

    def search_complete():
        rsp = query_api.get_results(query_id=query_id, rows=0).response
        meta = rsp["response_header"]["searchers_meta"]
        return meta["completed"] == meta["contacted"]

    assert wait_until(search_complete), f"Search did not complete, response: {query_api.get_results(query_id=query_id)}"
    data = query_api.get_results(query_id=query_id, rows=100).response["data"]
    if expect_data:
        assert data, f"No data. Full response content {query_api.get_results(query_id=query_id)}"
    return data


def _get_child_parent_process(results: Iterable) -> Tuple[dict, dict]:
    """
    Returns a single instance, not every instance.
    :param results: corpus to go over
    :return: (parent, child)
    """
    child, parent = None, None
    children_itr = (result for result in results if result.get("parent_guid"))
    for child_instance in children_itr:
        parent_itr = (
            result
            for result in results
            if result.get("process_guid") and child_instance["parent_guid"] == result["process_guid"]
        )
        try:
            parent = next(parent_itr)
            child = child_instance
            break
        except StopIteration:
            # No parent in data, find another child
            continue
    else:
        pytest.fail("No viable child/parent pair")

    return parent, child
