import os
from functools import partial
from typing import Optional, List, Dict, Any
from uuid import uuid4

import pytest

from pscr_integration.test_lifecycle.active_session import is_k8s
from pscr_integration.api.api_request import HTTPApiRequest, HTTPApiResponse
from pscr_integration.watchlists.watchlist import Watchlist


@pytest.mark.smoke_test
class WatchlistApi:
    def __init__(self, privatelink_hostname: str, org_id: str):
        self.org_id = org_id
        privatelink_hostname = os.environ.get("FEED_MANAGER_API_PRIVATELINK_HOSTNAME", privatelink_hostname)
        if is_k8s() and not privatelink_hostname:
            privatelink_hostname = "psc-lc-feed-mgr"
        url_prefix = "http://" + privatelink_hostname + ":5000"
        self._make_request = partial(HTTPApiRequest, org_id=org_id, url_prefix=url_prefix)

    def _send_request(
        self,
        route: str,
        method: str = "GET",
        data: Dict[Any, Any] = None,
        params: Dict[Any, Any] = None,
    ) -> HTTPApiResponse:
        request = self._make_request()
        return request.send_request(route, method, payload=data, params=params)

    def get_watchlists(self) -> HTTPApiResponse:
        return self._send_request("/watchlistmgr/v1/watchlist")

    def get_watchlist(self, watchlist_id: str) -> HTTPApiResponse:
        return self._send_request(f"/watchlistmgr/v1/watchlist/{watchlist_id}")

    def create_feed_watchlist(self, feed_id: str, name: Optional[str] = None) -> HTTPApiResponse:
        if name is None:
            name = str(uuid4())[:8]

        return self._send_request(
            "/watchlistmgr/v1/watchlist",
            method="POST",
            data={
                "classifier": {
                    "name": f"{feed_id} Watchlist {name}",
                    "description": f"Watchlist for feed {feed_id}",
                    "classifier_key": "feed_id",
                    "classifier_value": feed_id,
                }
            },
        )

    def delete_watchlist(self, watchlist_id: str) -> HTTPApiResponse:
        return self._send_request(f"/watchlistmgr/v1/watchlist/{watchlist_id}", method="DELETE")

    def get_watchlist_alert_status(self, watchlist_id: str) -> HTTPApiResponse:
        return self._send_request(f"/watchlistmgr/v1/watchlist/{watchlist_id}/alert")

    def add_watchlist_alert(self, watchlist_id: str) -> HTTPApiResponse:
        return self._send_request(f"/watchlistmgr/v1/watchlist/{watchlist_id}/alert", method="PUT")

    def remove_watchlist_alert_status(self, watchlist_id: str) -> HTTPApiResponse:
        return self._send_request(f"/watchlistmgr/v1/watchlist/{watchlist_id}/alert", method="DELETE")

    def get_report_ignore(self, report_id: str) -> HTTPApiResponse:
        return self._send_request(f"/watchlistmgr/v1/report/{report_id}/ignore")

    def get_report_ioc_ignore(self, report_id: str, ioc_id: str) -> HTTPApiResponse:
        return self._send_request(f"/watchlistmgr/v1/report/{report_id}/ioc/{ioc_id}/ignore")

    def add_report_ignore(self, report_id: str) -> HTTPApiResponse:
        return self._send_request(f"/watchlistmgr/v1/report/{report_id}/ignore", method="PUT")

    def add_report_ioc_ignore(self, report_id: str, ioc_id: str) -> HTTPApiResponse:
        return self._send_request(f"/watchlistmgr/v1/report/{report_id}/ioc/{ioc_id}/ignore", method="PUT")

    def remove_report_ignore(self, report_id: str) -> HTTPApiResponse:
        return self._send_request(f"/watchlistmgr/v1/report/{report_id}/ignore", method="DELETE")

    def remove_report_ioc_ignore(self, report_id: str, ioc_id: str) -> HTTPApiResponse:
        return self._send_request(f"/watchlistmgr/v1/report/{report_id}/ioc/{ioc_id}/ignore", method="DELETE")

    def get_severity_overrides(self) -> HTTPApiResponse:
        return self._send_request("/watchlistmgr/v1/severity")

    def get_report_severity_override(self, report_id: str) -> HTTPApiResponse:
        return self._send_request(f"/watchlistmgr/v1/severity/report/{report_id}")

    def set_report_severity_override(self, report_id: str, severity: int) -> HTTPApiResponse:
        return self._send_request(
            f"/watchlistmgr/v1/severity/report/{report_id}",
            method="PUT",
            data={"severity": severity},
        )

    def remove_report_severity_override(self, report_id: str) -> HTTPApiResponse:
        return self._send_request(f"/watchlistmgr/v1/severity/report/{report_id}", method="DELETE")

    def get_watchlist_telemetry(self, watchlist_ids: List[str], intervals: List[int] = None) -> HTTPApiResponse:
        return self._send_request(
            f'/watchlistmgr/v3/orgs/{self.org_id}/watchlists/{",".join(watchlist_ids)}/telemetry',
            method="GET",
            params={"intervals": ",".join(str(interval) for interval in intervals)} if intervals else None,
        )

    @staticmethod
    def classifier_watchlist_to_watchlist(watchlist: Dict) -> Watchlist:
        return Watchlist(
            id=watchlist["classifier"]["watchlist_id"],
            name=watchlist["classifier"]["name"],
            description=watchlist["classifier"]["description"],
            classifier_key=watchlist["classifier"]["classifier_key"],
            classifier_value=watchlist["classifier"]["classifier_value"],
            tags_enabled=watchlist["classifier"]["tags_enabled"],
            alerts_enabled=watchlist["classifier"]["alerts_enabled"],
        )
