import os
import json
from functools import partial

from pscr_integration.test_lifecycle.active_session import is_k8s
from pscr_integration.api.api_request import (
    HTTPApiRequest,
    HTTPApiResponse,
    create_request_context,
)


class XdreControllerApi:
    def __init__(self, privatelink_hostname: str):
        privatelink_hostname = os.environ.get("XDRE_API_PRIVATELINK_HOSTNAME", privatelink_hostname)
        if is_k8s() and not privatelink_hostname:
            privatelink_hostname = "psc-lc-xdre-controller"
        url_prefix = "http://" + privatelink_hostname + ":5007"
        request_context = create_request_context()
        if not request_context["th_gateway_url"].startswith("http"):
            request_context.update(th_gateway_url="https://" + request_context["th_gateway_url"])
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "X-Request-Context": json.dumps(request_context),
        }
        self._make_request = partial(HTTPApiRequest, org_id="xdre", url_prefix=url_prefix, headers=headers)

    def _send_request(self, route: str, method: str) -> HTTPApiResponse:
        request = self._make_request()
        return request.send_request(route=route, method=method)

    def health_check(self) -> HTTPApiResponse:
        """
        Get application health status.
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request("/xdre/actuator/health", method="GET")

    def create_replay(self, org_key: str) -> HTTPApiResponse:
        """
        Creates a replay.
        :param org_key: org key "_08joRsq"
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request(f"/xdre/v1beta/orgs/{org_key}/replays", method="POST")

    def get_replay(self, org_key: str, replay_id: str) -> HTTPApiResponse:
        """
        Gets a replay.
        :param org_key: org key "_08joRsq"
        :param replay_id: external id "bee7a0b7-3797-4b9c-984a-fbde5616e2bd"
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request(f"/xdre/v1beta/orgs/{org_key}/replays/{replay_id}", method="GET")

    def get_replays(self, org_key: str) -> HTTPApiResponse:
        """
        Gets replays by org.
        :param org_key: org key "_08joRsq"
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request(f"/xdre/v1beta/orgs/{org_key}/replays", method="GET")

    def get_details(self, org_key: str) -> HTTPApiResponse:
        """
        Gets details of org.
        :param org_key: org key "_08joRsq"
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request(f"/xdre/v1beta/orgs/{org_key}/details", method="GET")

    def start_livestreaming(self, org_key: str) -> HTTPApiResponse:
        """
        Starts livestreaming for org.
        :param org_key: org key "_08joRsq"
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request(f"/xdre/v1beta/orgs/{org_key}/livestreaming/start", method="POST")

    def stop_livestreaming(self, org_key: str) -> HTTPApiResponse:
        """
        Stops livestreaming for org.
        :param org_key: org key "_08joRsq"
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request(f"/xdre/v1beta/orgs/{org_key}/livestreaming/stop", method="POST")

    def start_coldstorage(self, org_key: str) -> HTTPApiResponse:
        """
        Starts coldstorage for org.
        :param org_key: org key "_08joRsq"
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request(f"/xdre/v1beta/orgs/{org_key}/coldstorage/start", method="POST")

    def stop_coldstorage(self, org_key: str) -> HTTPApiResponse:
        """
        Stops coldstorage for org.
        :param org_key: org key "_08joRsq"
        :return: an ApiResponse with either a valid response or error information
        """
        return self._send_request(f"/xdre/v1beta/orgs/{org_key}/coldstorage/stop", method="POST")
