import asyncio
import itertools
import logging
import os
import time
from datetime import datetime

import boto3
import pytest
import yaml
from pscr_common.metadata import MetadataProvider
from pscr_common.metrics import get_reporter
from pscr_common.stream import stream
from py.xml import html
from pyformance import MetricsRegistry

from pscr_integration.api.feed_api import FeedApi
from pscr_integration.api.watchlist_api import WatchlistApi
from pscr_integration.data_loader import Loader
from pscr_integration.lc_services import LcServices
from pscr_integration.logs import cloudwatch_reader
from pscr_integration.search.searcher_service import SearcherService
from pscr_integration.test_lifecycle.active_session import (
    set_execution_mode,
    is_ecs,
    is_k8s,
    is_smoke,
    has_clm,
    TestMode,
)
from pscr_integration.test_lifecycle.deprovision import deprovision, validate_deprovisioning
from pscr_integration.test_lifecycle.feed_data import feed_data, feed_data_2
from pscr_integration.test_lifecycle.rbac_server import rbac_run
from pscr_integration.test_lifecycle.saved_search_data import get_saved_search_data

_logger = logging.getLogger(__name__)
_tc_id = itertools.count()
test_reports = ["report.html", "junit.xml"]


def pytest_configure(config):
    _logger.info("In pytest configure %s", os.getcwd())
    if config.option.markexpr == "smoke_test":
        set_execution_mode(TestMode.smoke)

    # clean up test reports before test started
    for report in test_reports:
        if os.path.exists(report):
            os.remove(report)


def pytest_unconfigure(config):
    _logger.warning("In pytest unconfigure")
    bucket_name = cloudwatch_reader.CloudwatchLogRead.S3_BUCKET
    s3_client = boto3.client("s3")
    # copy report to s3 bucket
    for report in test_reports:
        if not os.path.exists(report):
            _logger.warning("report '%s' does not exists!", report)
            continue
        file_path = os.path.join("pscr", os.environ.get("ENV_NAME"), report)
        _logger.warning("uploading %s to s3://%s/%s", report, bucket_name, file_path)
        try:
            s3_client.upload_file(report, bucket_name, file_path)
        except Exception as ex:
            _logger.error(
                "Failed to upload %s to s3://%s/%s: %s",
                report,
                bucket_name,
                file_path,
                ex,
            )


def add_env_vars_to_output(env_vars):
    output_path = "/tmp/output.yml"
    with open(output_path, "w+") as f:
        output = yaml.safe_load(f)
    if output is None:
        output = {}
    output["env_vars"] = env_vars
    with open(output_path, "w") as f:
        yaml.dump(output, f)


@pytest.fixture(scope="session", autouse=True)
def config():
    config_file = os.path.dirname(os.path.abspath(__file__)) + "/configuration_values.yml"
    with open(config_file, mode="r", encoding="utf-8") as config_stream:
        config_data = yaml.load(config_stream, Loader=yaml.FullLoader)
        env_name = os.environ.get("ENV_NAME")
    data = {
        "config_file": config_data,
        "env_vars": {
            "ENV_NAME": env_name,
            "INDEX_REGISTRY_LAMBDA": os.environ.get("INDEX_REGISTRY_LAMBDA"),
            "API_PRIVATELINK_HOSTNAME": os.environ.get("API_PRIVATELINK_HOSTNAME"),
            "ECS_CLUSTER_NAME": os.environ.get("ECS_CLUSTER_NAME"),
            "ECS_CLUSTER_SECURITY_GROUP": os.environ.get("ECS_CLUSTER_SECURITY_GROUP"),
            "PRIVATE_SUBNET": os.environ.get("PRIVATE_SUBNET"),
            "TEST_ORG1_TASK_DEFINITION": os.environ.get("TEST_ORG1_TASK_DEFINITION"),
            "DOCUMENTS_BUCKET": os.environ.get("DOCUMENTS_BUCKET"),
            "EVENTS_STREAM": os.environ.get("EVENTS_STREAM"),
            "DEPROVISIONING_SQS_QUEUE": os.environ.get("DEPROVISIONING_SQS_QUEUE"),
        },
        "log_groups": [
            f"/pscr/{env_name}/document_builder",
            f"/pscr/{env_name}/ingress_detector",
            f"/pscr/{env_name}/housekeeper_{env_name}__pscr_indexer",
            f"/pscr/{env_name}/housekeeper_{env_name}__pscr_query_detector",
            f"/pscr/{env_name}/indexer",
            f"/pscr/{env_name}/query_detector",
            f"/pscr/{env_name}/search_service",
            f"/pscr/{env_name}/federated_search",
            f"/pscr/{env_name}/searcher",
            f"/pscr/{env_name}/feed_manager",
            f"/pscr/{env_name}/org_retention",
            f"/pscr/{env_name}/{env_name}__pscr_indexer/fluent_bit_logzio",
            f"/pscr/{env_name}/{env_name}__pscr_indexer_wavefront_cadvisor",
            f"/pscr/{env_name}/{env_name}__pscr_main/fluent_bit_logzio",
            f"/pscr/{env_name}/{env_name}__pscr_main_wavefront_cadvisor",
            f"/pscr/{env_name}/{env_name}__pscr_query_detector/fluent_bit_logzio",
            f"/pscr/{env_name}/{env_name}__pscr_query_detector_wavefront_cadvisor",
            f"/pscr/{env_name}/{env_name}__pscr_worker/fluent_bit_logzio",
            f"/pscr/{env_name}/{env_name}__pscr_worker_wavefront_cadvisor",
            f"/pscr/{env_name}/index_registry_dbtool",
            f"/pscr/{env_name}/intelligent_feed_search",
            f"/pscr/{env_name}/org_mapper",
            f"/pscr/{env_name}/raw_event_replay",
            f"/pscr/{env_name}/search_result_mem_agg_java",
            f"/pscr/{env_name}/telegraf-signalfx",
            f"/pscr/{env_name}/telegraf_metric_proxy",
            f"/pscr/{env_name}/vpc_flow",
            f"/pscr/{env_name}/wavefront_proxy",
            f"/aws/lambda/{env_name}__index_registry",
            f"/aws/lambda/{env_name}__index_registry_s3_trigger",
            f"/aws/lambda/{env_name}__index_registry_sql_purge",
            f"/aws/lambda/{env_name}__index_registry_metrics",
            f"/aws/lambda/{env_name}__ecs_indexer_lifecycle_hook_lambda",
            f"/aws/lambda/{env_name}__ecs_lifecycle_hook_lambda",
            f"/aws/lambda/{env_name}__ecs_query_detector_lifecycle_hook_lambda",
            f"/aws/lambda/{env_name}__feed_telemetry_aggregator",
            f"/aws/lambda/{env_name}__indexer_ec2_ecs_capacity_sync",
            f"/aws/lambda/{env_name}__query_detector_ec2_ecs_capacity_sync",
            f"/aws/lambda/{env_name}__reputation_cache_updater",
        ],
        "tests_start_time": datetime.now(),
    }
    _logger.info(f"Config Data: {data}")
    add_env_vars_to_output(data["env_vars"])
    return data


# @TODO Remove this when we have search history
@pytest.fixture(scope="session", autouse=True)
def clean_results(config):
    if is_smoke() and not is_k8s():
        bucket = boto3.resource("s3").Bucket(f'{config["env_vars"]["ENV_NAME"]}-solr-results-bucket')
        # Doc claims that we can only delete 1000 at a time
        while list(bucket.objects.filter(Prefix="search/test_").limit(1000)):
            bucket.objects.filter(Prefix="search/test_").delete()
    yield


@pytest.fixture(scope="session", autouse=True)
def rbac_server():
    if is_smoke():
        rbac_run()


@pytest.fixture(scope="session", autouse=True)
def loader_1(config):
    device_count = config["config_file"]["test_constants"]["org1_device_count"]

    _loader = Loader(
        task_definition=config["env_vars"]["TEST_ORG1_TASK_DEFINITION"],
        ecs_cluster_name=config["env_vars"]["ECS_CLUSTER_NAME"],
        security_group=config["env_vars"]["ECS_CLUSTER_SECURITY_GROUP"],
        subnet=config["env_vars"]["PRIVATE_SUBNET"],
        documents_bucket=config["env_vars"]["DOCUMENTS_BUCKET"],
        events_stream=config["env_vars"]["EVENTS_STREAM"],
        device_count=device_count,
        org_id_suffix="1",
    )

    return _loader


@pytest.fixture(scope="session", autouse=True)
def loader_2(config):
    device_count = config["config_file"]["test_constants"]["org1_device_count"]

    _loader = Loader(
        task_definition=config["env_vars"]["TEST_ORG1_TASK_DEFINITION"],
        ecs_cluster_name=config["env_vars"]["ECS_CLUSTER_NAME"],
        security_group=config["env_vars"]["ECS_CLUSTER_SECURITY_GROUP"],
        subnet=config["env_vars"]["PRIVATE_SUBNET"],
        documents_bucket=config["env_vars"]["DOCUMENTS_BUCKET"],
        events_stream=config["env_vars"]["EVENTS_STREAM"],
        device_count=device_count,
        org_id_suffix="2",
    )

    return _loader


@pytest.fixture(scope="session")
def load_org_1(config, loader_1, loader_2):
    if LcServices.target_service() in LcServices.nta_services():
        loader_1.load_nta()

    # use loader 2 to load in data so query IOCs can be properly validated
    loader_2.load_data()

    query = SearcherService(
        config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
        loader_2.org_id,
        device_id_start=loader_2.device_id_start,
        device_id_end=loader_2.device_id_end,
    )

    query.count_query(2000, "*:*", 1)

    _logger.info("Loading feeds")
    feed_api = FeedApi(config["env_vars"]["API_PRIVATELINK_HOSTNAME"], loader_1.org_id)
    watchlist_api = WatchlistApi(config["env_vars"]["API_PRIVATELINK_HOSTNAME"], loader_1.org_id)
    # Get rid of any existing feeds that may have been left around from previous smoke tests
    for feed in feed_api.get_feeds().response["results"]:
        feed_api.delete_feed(feed["id"])

    feedinfo = feed_data["feedinfo"]
    feedinfo["owner"] = loader_1.org_id
    feed = feed_api.create_feed(feedinfo, feed_data["reports"])
    if feed.status_code != 200:
        raise RuntimeError(f"Unable to create feed: {feed.response}")

    feed_id = feed.response["id"]
    watchlist_id = watchlist_api.create_feed_watchlist(feed_id, "load_org_1").response["classifier"]["watchlist_id"]

    alert = watchlist_api.add_watchlist_alert(watchlist_id)
    if alert.status_code != 200:
        raise RuntimeError(f"Unable to add watchlist alert: {alert.response}")

    feedinfo2 = feed_data_2["feedinfo"]
    feedinfo2["owner"] = loader_1.org_id
    feed2 = feed_api.create_feed(feedinfo2, feed_data_2["reports"])
    if feed2.status_code != 200:
        raise RuntimeError(f"Unable to create feed: {feed2.response}")

    feed_id2 = feed2.response["id"]
    # Sleep some to let feed take effect
    time.sleep(120)
    _logger.info("Loading data now")
    loader_1.load_data()
    _logger.info("Finished data now")

    _logger.info(f"Test 1 org id: {loader_1.org_id}")

    _logger.info("Waiting for data to be indexed")
    query = SearcherService(
        config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
        loader_1.org_id,
        device_id_start=loader_1.device_id_start,
        device_id_end=loader_1.device_id_end,
    )
    query.count_query(
        2000,
        "-report_id:*",
        config["config_file"]["test_constants"]["org1_process_count"],
        0,
    )
    _logger.info("Data is indexed into stage 0")

    yield {
        "loader": loader_1,
        "feed_id": feed_id,
        "feed_id_2": feed_id2,
        "watchlist_id": watchlist_id,
        "saved_searches": get_saved_search_data(loader_1.org_id),
    }

    if is_ecs():
        feed_api.delete_feed(feed_id)
        feed_api.delete_feed(feed_id2)


# noinspection SpellCheckingInspection
@pytest.fixture(scope="session", autouse=True)
def deprovisioning_validation(config, load_org_1, loader_2):
    if is_ecs() or not has_clm():
        yield
        return

    loader_1 = load_org_1["loader"]
    org_keys = (loader_1.org_id, loader_2.org_id)
    clm_listener_sqs_queue = config["env_vars"]["DEPROVISIONING_SQS_QUEUE"]

    yield
    for org_key in org_keys:
        deprovision(clm_listener_sqs_queue, org_key, org_key)
    for org_key in org_keys:
        validate_deprovisioning(config["env_vars"]["API_PRIVATELINK_HOSTNAME"], org_key)


@pytest.fixture(scope="session", autouse=True)
def scan_logs(config):
    ts = int(time.time())

    yield

    if not is_smoke():
        log_scanner = cloudwatch_reader.CloudwatchLogRead(
            log_groups=config["log_groups"], lower_bound=ts, write_to_s3=True
        )
        log_summaries = log_scanner.get_log_summary()
        for log_line in log_summaries.values():
            assert "WARN" not in log_line
            assert "ERROR" not in log_line


def pytest_collection_modifyitems(session, config, items):
    """
    This is a hook that only runs debug marked tests if any are present
    """
    debug_tests = stream(items).filter(lambda t: list(t.iter_markers(name="debug"))).to_list()
    if debug_tests:
        items[:] = debug_tests


# noinspection PyProtectedMember
@pytest.fixture(scope="session", autouse=True)
def configure_html_report_env(request):
    bucket_name = cloudwatch_reader.CloudwatchLogRead.S3_BUCKET
    env_name = os.environ.get("ENV_NAME")
    test_log_path = f"s3://{bucket_name}/pscr/{env_name}/"
    lambda_log_path = f"s3://{bucket_name}/aws/lambda/{env_name}/"
    # pylint: disable=protected-access
    request.config._metadata.update({"_Env name": env_name})
    request.config._metadata.update({"_Logs pscr test path": f"aws s3 ls {test_log_path}"})
    request.config._metadata.update({"_Logs aws lambda path": f"aws s3 ls {lambda_log_path}"})
    # cloudwatch logs
    run_outfile = "run_output.txt"
    if os.path.exists(run_outfile):
        with open(run_outfile, mode="r", encoding="utf-8") as f:
            for line in f.readlines():
                try:
                    key, value = line.strip().split(":=")
                    request.config._metadata.update({"_" + key: value})
                except ValueError:
                    pass


@pytest.mark.optionalhook
def pytest_html_results_table_header(cells):
    cells.insert(0, html.th("TC_ID", class_="sortable numeric initial-sort"))
    cells.pop()


@pytest.mark.optionalhook
def pytest_html_results_table_row(report, cells):
    if not report:
        return
    cells.insert(0, html.td(next(_tc_id)))
    cells.pop()


def report_metrics(metric_name, status, test_name=None):
    metrics_registry = MetricsRegistry()
    env = os.environ.get("ENV_NAME", "unknown")
    env_meta = MetadataProvider().get_metadata()
    metrics_reporter = get_reporter(
        metrics_registry,
        env,
        env_meta.task_id,
        "integration_tests",
        env_meta.image.rsplit(":", maxsplit=1)[-1],
        os.getpid(),
    )
    metrics_reporter.with_tag("status", status)
    if test_name:
        metrics_reporter.with_tag("tc_name", test_name)

    if metrics_reporter is None:
        _logger.error("Something went wrong when creating the metrics reporter")
        raise EnvironmentError("Couldn't find the env vars for creating the metrics reporter")

    metrics_registry.counter(f"smoke_tests.{metric_name}").inc()
    asyncio.run(metrics_reporter.report_now())


@pytest.hookimpl
def pytest_sessionstart(session):
    session.results = {}


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    outcome = yield
    result = outcome.get_result()

    if result.when == "call":
        item.session.results[item] = result


@pytest.hookimpl
def pytest_sessionfinish(session, exitstatus):
    # Enable reporting metrics only for smoke tests,
    # because integration tests are running on fargate launched containers
    if not is_smoke():
        return

    for result in session.results:
        report_metrics("test_case_outcome", session.results[result].outcome, session.results[result].nodeid)

    status = {0: "passed", 1: "failed", 2: "interrupted", 3: "internal error", 4: "usage error", 5: "no tests"}
    report_metrics("test_run_outcome", status.get(exitstatus, "unknown"))
