import itertools
import json
import logging
import os
import random
import string
import struct
import time
from datetime import datetime, timedelta
import boto3
import pscr_common.metadata
import yaml
from botocore.config import Config
from proto_proto_pscr.cb_psc_event_pb2 import CbEventMsg, CbEventBatchMsg

from pscr_integration.test_lifecycle.active_session import is_smoke, is_gov
from pscr_integration.util import generate_zip, little_end

_logger = logging.getLogger(__name__)


def _get_random_alphanumeric_string(length):
    letters_and_digits = string.ascii_letters + string.digits
    result_str = "".join((random.choice(letters_and_digits) for i in range(length)))
    return result_str


def append_org_to_output(org_key):
    output_path = "/tmp/output.yml"
    with open(output_path, "r") as f:
        output = yaml.safe_load(f)
    if output is None:
        output = {}
    if output.get("org_keys") is None:
        output["org_keys"] = []
    output["org_keys"].append(org_key)
    with open(output_path, "w") as f:
        yaml.dump(output, f)


class Loader:
    def __init__(
        self,
        ecs_cluster_name,
        subnet,
        security_group,
        task_definition,
        documents_bucket,
        events_stream,
        org_id_suffix,
        device_count=1,
        org_id=None,
        nta_device_id=None,
    ):
        config = Config(retries={"max_attempts": 10})
        metadata = pscr_common.metadata.MetadataProvider().get_metadata()
        self.ecs_client = boto3.client("ecs", config=config)
        self.s3_client = boto3.client("s3")
        self.ecs_cluster_name = ecs_cluster_name
        alt_name = f"_{_get_random_alphanumeric_string(7)}"
        if metadata.runtime == "k8s":
            clean_name = (
                metadata.container_name.replace("psc-lc-", "").replace("--test", "").replace("-", "").replace("_", "")
            )
            alt_name = f"_{clean_name[:8]}{org_id_suffix}{metadata.container_id[:4]}"
        self.org_id = org_id if org_id else alt_name
        append_org_to_output(self.org_id)
        self.device_id_start = int(time.time())
        self.device_count = device_count
        # Subtract 1 for inclusive range
        self.device_id_end = self.device_id_start + device_count - 1
        self.nta_device_id = nta_device_id if nta_device_id else str(random.randint(10**4, 10**7))
        self.subnet = subnet
        self.security_group = security_group
        self.task_definition = task_definition
        self.documents_bucket = documents_bucket
        self.events_stream = events_stream
        self.data_loaded = False
        _logger.info(f"Created loader for org {self.org_id}")

    def delete_data(self):
        self._s3_folder_delete(self.documents_bucket, f"documents/org_id={self.org_id}")
        # TODO  Need to add raw, optimized, and results buckets

    def _read_events(self, read_nta=False):
        res_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources")
        data_contents = []
        for data in sorted(os.listdir(res_dir)):
            with open(os.path.join(res_dir, data), "rb") as f:
                if data.startswith("nta-smoke"):
                    if read_nta:
                        data_contents.append(f.read())
                else:
                    if not read_nta:
                        data_contents.append(f.read())

        return data_contents

    def _load_batch(self, batch, org_id, device_id, os_meta, kinesis_client):
        metadata = json.dumps(
            {
                "version": 1,
                "orgId": org_id,
                "deviceId": device_id,
                "deviceGroup": f"GROUP_{device_id}",
                "deviceName": f"HOST_{device_id}",
                "selfReportedDeviceIp": "1.1.1.1",
                "backendReportedDeviceIp": "1.2.2.2",
                "orgSize": 100,
                "ingressTime": time.time(),
                "osType": os_meta,
            }
        ).encode("utf-8")
        metadata_length = struct.pack(">i", len(metadata))
        kinesis_package = metadata_length + metadata + batch
        kinesis_client.put_record(
            StreamName=self.events_stream,
            PartitionKey=device_id,
            Data=kinesis_package,
        )

    def load_data(self):
        if self.data_loaded:
            _logger.warning(f"Data for org {self.org_id} already loaded. Skipping loading of data.")
            return
        _logger.info(f"Loading data for org {self.org_id}")
        data_contents = self._read_events()
        kinesis = boto3.client("kinesis")
        for i, os_meta in zip(
            range(self.device_id_start, self.device_id_end + 1),
            itertools.cycle(("WINDOWS", "OSX", "LINUX")),
        ):
            for data in data_contents:
                device_id = str(i)
                self._load_batch(data, self.org_id, device_id, os_meta, kinesis)
        self.data_loaded = True

    def learning_period_generator(
        self,
        event: CbEventMsg,
        learning_period_days: int = 10,
        events_per_day: int = 3,
    ):
        """Generate past events so the llanta can "learn".
        After a baseline has been created, generate an anomalous event to trigger detectors.

        Updates the event inplace.

        Args:
            event (CbAugmentedEventBatchMsg): the event on which to base the learning period events. Updated inplace.
            learning_period_days (int, optional): for how many days back to generate events. Defaults to 10.
            events_per_day (int, optional): how many events per day to generate. Defaults to 3.
        """

        for timestamp_offset_days in range(learning_period_days, -1, -1):
            for event_idx in range(events_per_day, -1, -1):
                time_ago = datetime.now() - timedelta(days=timestamp_offset_days, minutes=event_idx)
                last_p_timestamp = int(time_ago.timestamp())
                first_p_timestamp = int((time_ago - timedelta(milliseconds=2)).timestamp())

                event.netconn_op.target.create_time.seconds = first_p_timestamp
                event.netconn_op.flow_info.first_packet_timestamp.seconds = first_p_timestamp
                event.netconn_op.flow_info.last_packet_timestamp.seconds = last_p_timestamp

                # change 3 random chars of the community_id
                community_id = list(event.netconn_op.target.community_id)
                for i in random.sample(range(len(event.netconn_op.target.community_id)), 3):
                    community_id[i] = random.choice(string.digits)
                event.netconn_op.target.community_id = "".join(community_id)

                if timestamp_offset_days == 0:
                    # generate an anomalous event for the current day to trigger the port profiler
                    event.netconn_op.target.remote_port = little_end(random.randint(1, 65535))

                yield event

    def load_nta(self):
        """Load data for the NTA tests"""
        if is_smoke() and (not is_gov()):
            kinesis = boto3.client("kinesis")
            for data_contents in self._read_events(read_nta=True):
                eventBatch = CbEventBatchMsg()
                eventBatch.ParseFromString(data_contents)

                for _ in self.learning_period_generator(eventBatch.events[0]):
                    # eventBatch is getting modified inplace
                    data = generate_zip(("event.proto", eventBatch.SerializeToString()))
                    self._load_batch(data, self.org_id, self.nta_device_id, "WINDOWS", kinesis)

    def _s3_folder_delete(self, bucket, key_prefix):
        bucket = self.s3_client.get_bucket(bucket)
        for key in bucket.list(prefix=key_prefix):
            key.delete()
