import base64
import logging
import time
from datetime import datetime, timedelta

import boto3
import snappy

from pscr_integration.kinesis.aws_kinesis_agg import deaggregator

_logger = logging.getLogger(__name__)

logging.getLogger("boto3").setLevel(logging.WARNING)
logging.getLogger("botocore").setLevel(logging.WARNING)


def get_records(stream: str, timeout: int = 1800, decompress: bool = True) -> list:
    #  Implemented as a generator, will yield each record gets it.
    kinesis = boto3.client("kinesis")
    timeout_time = datetime.now() + timedelta(seconds=timeout)
    shards = kinesis.list_shards(StreamName=stream)
    if len(shards["Shards"]) == 0:
        _logger.error(f"No shards in stream {stream}")
        return []

    shard_it = kinesis.get_shard_iterator(
        StreamName=stream,
        ShardId=shards["Shards"][0]["ShardId"],
        ShardIteratorType="TRIM_HORIZON",
    )["ShardIterator"]
    retry = 0
    while True:
        out = kinesis.get_records(ShardIterator=shard_it)
        if not out:
            _logger.error("Empty get records response returned from Kinesis stream")
            break

        # repackage the records to the format aws uses when lambdas are triggered from kinesis streams
        # the de-aggregator library expects that.
        records = []
        for record in out["Records"]:
            records.append(
                {
                    "kinesis": {
                        "data": base64.b64encode(record["Data"]),
                        "kinesisSchemaVersion": 1,
                        "sequenceNumber": record["SequenceNumber"],
                    }
                }
            )
        user_records = deaggregator.deaggregate_records(records)
        for user_record in user_records:
            try:
                data = base64.b64decode(user_record["kinesis"]["data"])
                if decompress:
                    data = snappy.decompress(data)
            except snappy.UncompressError as e:
                _logger.error(e)
                raise
            yield data
            timeout_time = datetime.now() + timedelta(seconds=timeout)
        shard_it = out["NextShardIterator"]

        if datetime.now() >= timeout_time:
            _logger.warning("Timed out waiting for Kinesis records")
            break
        retry += 1
        _logger.info(f"Still waiting for all the records, loop: {retry}")
        time.sleep(10)
