import os
from enum import Enum


class LcServices:
    """Class used to determine which service we're running smoke/poke tests for - the target service.
    We also define groups of services involved in various functionalities of LC.
    With this we can skip test cases that are irrelevant to the current target service.

    (The default target service is _LcService.SMOKE,
        which is part of all service groups, since we want to run all test cases on it)
    """

    class _LcService(Enum):
        """Services for which we run smoke tests"""

        DOCBUILDER = 1
        FEDERATED_SEARCH = 2
        FEED_MANAGER = 3
        FEED_SEARCH = 4
        INDEX_REGISTRY = 5
        INDEXER = 6
        INGRESS_DETECTOR = 7
        NTA_BATCHING_TASK = 8
        NTA_STREAMING_TASK = 9
        ORG_RETENTION = 10
        QUERY_DETECTOR = 11
        SEARCH_AGGREGATOR = 12
        SEARCH_API = 13
        SEARCHER = 14
        SMOKE = 15

    # mapping between helm chart name and service
    _service_name_map = {
        "docbuild": _LcService.DOCBUILDER,
        "psc-lc-federated-search": _LcService.FEDERATED_SEARCH,
        "psc-lc-feed-mgr": _LcService.FEED_MANAGER,
        "psc-lc-feed-search": _LcService.FEED_SEARCH,
        "psc-lc-idx-reg-app": _LcService.INDEX_REGISTRY,
        "indexer": _LcService.INDEXER,
        "ingdet": _LcService.INGRESS_DETECTOR,
        "psc-lc-nta-batching-task": _LcService.NTA_BATCHING_TASK,
        "psc-lc-nta-streaming-task": _LcService.NTA_STREAMING_TASK,
        "psc-lc-org-ret": _LcService.ORG_RETENTION,
        "qrydetect": _LcService.QUERY_DETECTOR,
        "psc-lc-search-agg-app": _LcService.SEARCH_AGGREGATOR,
        "psc-lc-search-api": _LcService.SEARCH_API,
        "searcher": _LcService.SEARCHER,
        "psc-lc-smoke": _LcService.SMOKE,
    }

    @classmethod
    def target_service(cls):
        """Get the service we're currently targeting to test"""
        target_service_name = os.getenv("TARGET_SERVICE_NAME", "psc-lc-smoke")
        if target_service_name in cls._service_name_map:
            return cls._service_name_map[target_service_name]
        return cls._LcService.SMOKE

    @classmethod
    def search_services(cls):
        """Get the services involved in searches"""
        return {
            cls._LcService.DOCBUILDER,
            cls._LcService.SEARCH_AGGREGATOR,
            cls._LcService.SEARCH_API,
            cls._LcService.SEARCHER,
            cls._LcService.INDEXER,
            cls._LcService.INDEX_REGISTRY,
            cls._LcService.SMOKE,
        }

    @classmethod
    def federated_search_services(cls):
        """Get the services involved in federated searches"""
        return cls.search_services().union({cls._LcService.FEDERATED_SEARCH})

    @classmethod
    def feed_services(cls):
        """Get the services involved in feed management"""
        return {
            cls._LcService.FEED_MANAGER,
            cls._LcService.FEED_SEARCH,
            cls._LcService.SEARCH_API,
            cls._LcService.SMOKE,
        }

    @classmethod
    def feed_hit_services(cls):
        """Get the services involved in getting feed hits"""
        return cls.search_services().union(
            {cls._LcService.INGRESS_DETECTOR, cls._LcService.QUERY_DETECTOR, cls._LcService.FEED_MANAGER}
        )

    @classmethod
    def org_retention_services(cls):
        """Get the services involved in org retention"""
        return {cls._LcService.ORG_RETENTION, cls._LcService.SMOKE}

    @classmethod
    def search_history_services(cls):
        """Get the services involved in search history"""
        return {cls._LcService.SEARCH_API, cls._LcService.SMOKE}

    @classmethod
    def nta_services(cls):
        """Get the services involved in network detection"""
        return cls.search_services().union({cls._LcService.NTA_STREAMING_TASK, cls._LcService.NTA_BATCHING_TASK})
