import os
import json
import logging
from datetime import datetime, timedelta, timezone
from typing import Iterable, Dict, AnyStr, Any, Sequence

import boto3
import smart_open

_logger = logging.getLogger(__name__)


class CloudwatchLogRead:
    S3_BUCKET = os.environ.get("INTEGRATION_TESTS_LOGS_BUCKET", "pscr-integration-test-logs")
    UTC_ZERO = datetime.fromtimestamp(0, tz=timezone.utc)

    def __init__(
        self,
        log_groups: Sequence[AnyStr],
        lower_bound: int = 0,
        write_to_s3: bool = False,
    ):
        self.log_groups = log_groups
        self.logs_client = boto3.client("logs")
        self.lower_bound = lower_bound
        self.write_to_s3 = write_to_s3
        self._build_log_summary()

    @staticmethod
    def _add_one_to_level(level: AnyStr, summary_dict: Dict[AnyStr, int]):
        if level not in summary_dict:
            summary_dict[level] = 1
        else:
            summary_dict[level] += 1

    def _pull_all_logs(self, log_group: AnyStr) -> Sequence[Dict[AnyStr, AnyStr]]:
        rsp = self.logs_client.filter_log_events(logGroupName=log_group, startTime=self.lower_bound)
        events = rsp["events"]
        while "nextToken" in rsp:
            rsp = self.logs_client.filter_log_events(
                logGroupName=log_group,
                startTime=self.lower_bound,
                nextToken=rsp["nextToken"],
            )
            events += rsp["events"]
        return events

    def _get_log_messages(self, log_group: AnyStr) -> Iterable[Dict[AnyStr, Any]]:
        if self.write_to_s3:
            f_s3 = smart_open.smart_open(f"s3://{self.S3_BUCKET}{log_group}", "w")
        try:
            logs = self._pull_all_logs(log_group)
        except self.logs_client.exceptions.ResourceNotFoundException as e:
            raise self.logs_client.exceptions.ResourceNotFoundException from e
        for log in logs:
            if self.write_to_s3:
                f_s3.write(f"{log['message']}\n")
            try:
                pscr_message = json.loads(log["message"])
                pscr_message["raw"] = False
            except json.decoder.JSONDecodeError:
                pscr_message = {
                    "raw": True,
                    "message": log["message"],
                    "timestamp": str(self.UTC_ZERO + timedelta(milliseconds=int(log["timestamp"]))),
                }
            yield pscr_message
        f_s3.close()

    def _build_log_summary(self):
        self.summary_levels = {}
        _logger.info("Checking logs")
        for log_group in self.log_groups:
            self.summary_levels[log_group] = {}
            for pscr_message in self._get_log_messages(log_group):
                if pscr_message["raw"]:
                    self._add_one_to_level("not-decoded", self.summary_levels[log_group])
                else:
                    pscr_message["skip"] = False
                    # Exclude know benign errors. Some happen only during env startup/shutdown and are recoverable
                    if "message" in pscr_message and CloudwatchLogRead.is_benign_error(pscr_message["message"]):
                        pscr_message["skip"] = True
                    elif "exception" in pscr_message:
                        # python emits exception as an array
                        if isinstance(pscr_message["exception"], list):
                            for line in pscr_message["exception"]:
                                if CloudwatchLogRead.is_benign_error(line):
                                    pscr_message["skip"] = True
                                    break
                        elif CloudwatchLogRead.is_benign_error(pscr_message["exception"]):
                            pscr_message["skip"] = True
                    if not pscr_message["skip"]:
                        self._add_one_to_level(pscr_message["level"], self.summary_levels[log_group])
                        if pscr_message["level"] == "WARN" or pscr_message["level"] == "ERROR":
                            _logger.error("%s: %s", log_group, pscr_message)

        _logger.info("Finish Checking logs")

    @staticmethod
    def is_benign_error(text):
        return (
            'relation "indexes" does not exist' in text
            or "psycopg2.OperationalError: could not translate host name" in text
            or "psycopg2.OperationalError: could not connect to server" in text
            or "the database system is starting up" in text
            or "Unable to load AWS credentials from environment variables" in text
            or ('database "' in text and '" does not exist' in text)
            or "password authentication failed for user" in text
            or "No search cores available" in text
            or "NoSearcherCoresAvailableException" in text
            or "Unable to load AWS credentials" in text
            or "Determined that we are not in AWS. Program will run assuming localstack setup." in text
            or "leases and maxLeasesForWorker" in text
            or "Starting optimize... Reading and rewriting" in text
            or "shutdown: NOT invoking checkpoint" in text
            or "I/O error on GET request" in text
            or "Store Lookup build failed" in text
            or "Loading stored lookup data failed, possibly not cached yet" in text
            or "No search cores available" in text
            or "Unknown watchlist" in text
            or "Received configuration for region" in text
            or "Error accessing S3 GeoIP file" in text
            or "uniqueKey is not stored - distributed search and MoreLikeThis will not work" in text
            or "Unable to load AWS credentials from environment variables" in text
            or "PERFORMANCE WARNING: Overlapping onDeckSearchers" in text
            or "No override for report" in text
            or "Received configuration for endpoint" in text
            or "Invalid lease response" in text
            or "Service: AmazonKinesis; Status Code: 500; Error Code: InternalFailure" in text
            or "java.lang.RuntimeException: Error running child process" in text
            or "Failed to add record to producer dispatch_kinesis_producer" in text
            or "Will stop dispatching to this client for 1 hour" in text
            or "The function is currently in the following state: Pending" in text
            or "this usually means the sharp map has changed" in text
            or "sqlalchemy.exc.OperationalError: (psycopg2.OperationalError) timeout expired" in text
            or "KMS Exception: UnrecognizedClientExceptionKMS Message: The security token included" in text
            or ("Skipping record because it is missing a Records key" in text and "s3:TestEvent" in text)
            or "Table not found results_" in text
            or "Errors batch deleting partitions results_" in text
            or (
                "Unknown error during manifest_notifications" in text
                and "The specified queue does not exist or you do not have access to it" in text
            )
            or ("timeout expired" in text and "Error during lambda handler" in text)
            or ("No partition keys are utilized (possibly due to CloudWatch metrics issue)" in text)
            or ("New shards map is empty. Returning without updating shards." in text)
            or ("Forcing summary persistence for agg Summary Aggregator" in text)
            or (
                "Timestream Query Exception: An error occurred (ValidationException) when calling the Query operation"
                in text
            )
            or ("Received GOAWAY with ENHANCE_YOUR_CALM" in text)
            or ("Increased keepalive time nanos to" in text)
            or ("No entitlements found for the org" in text)
            or ("Error sending points to Wavefront (HTTP -1). Data will be requeued and resent." in text)
            or ("Unable to obtain status code from the Wavefront service at" in text)
        )  # Index registry startup error

    def get_log_summary(self) -> Dict[AnyStr, Dict[AnyStr, int]]:
        return self.summary_levels
