## Common functions across tests in search_at_scale, process_details.
import json
import logging
import time
from datetime import datetime

import boto3
import pytest
from botocore.exceptions import ClientError

from pscr_integration.api.search_api_v2 import SearchAPIV2
from pscr_integration.test_lifecycle.active_session import is_k8s, is_kc
from pscr_integration.util import wait_until, get_sqs_job_id

_logger = logging.getLogger(__name__)

SEPARATOR = "_"
SQS_SERVICE = "sqsquery_optimized"


class SearchCommon:
    def get_latest_schema(self, config, org_key, query_id):
        schema = {}
        s3 = boto3.client("s3")
        if not is_k8s() or is_kc():
            bucket_name = f"{config['env_vars']['ENV_NAME']}-solr-results-bucket"
        else:
            bucket_name = f"{config['env_vars']['ENV_NAME']}-psc-lc-shared-infra-s3results"
        file_name = f"search/{org_key}/{query_id}/0/schema.json"

        # Adding wait_until to locate schema.json
        def object_exists():
            response = s3.list_objects_v2(Bucket=bucket_name, Prefix=file_name)
            return "Contents" in response

        if not wait_until(object_exists, timeout=120):
            pytest.fail(f"{file_name} not found")
        try:
            schema = json.loads(s3.get_object(Bucket=bucket_name, Key=file_name)["Body"].read())
        except Exception as error:
            pytest.fail(f"Error while downloading schema: {error}")
        return schema

    def parse_schema(self, schema_data):
        schema_dict = {}
        for data in schema_data["fields"]:
            schema_dict[data["name"]] = [
                data["type"],
                data["aggregation"],
                data["array"],
            ]
            _logger.debug(f"schema_field - {data['name']}:{schema_dict[data['name']]}")
        return schema_dict

    def compare_timestamp(self, time1, time2):
        try:
            time1_value = datetime.strptime(time1, "%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            # To cater timestamp without milliseconds.
            time1_value = datetime.strptime(time1, "%Y-%m-%dT%H:%M:%SZ")
        try:
            time2_value = datetime.strptime(time2, "%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            time2_value = datetime.strptime(time2, "%Y-%m-%dT%H:%M:%SZ")
        op = time.mktime(time1_value.timetuple()) - time.mktime(time2_value.timetuple())
        return op

    def compare_combine_array_values(self, array1, array2):
        diff = set(array2) - set(array1)
        if len(diff) > 0:
            array1.extend(diff)
        return array1

    def get_schema(self, config, load_org_1, query):
        env = config["env_vars"]["ENV_NAME"]

        # Check if the queues are created by indexer
        _wait_for_org_queues(_get_sqs_prefix(env), load_org_1["loader"].org_id)

        _logger.info("Download latest Schema json ")
        v2 = SearchAPIV2(
            privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
        )
        v2_job_id = v2.create_process_result_v2(query=query).response["job_id"]
        schema_op = self.get_latest_schema(config, load_org_1["loader"].org_id, v2_job_id)

        ## Parse Schema
        schema_data = self.parse_schema(schema_op)
        return schema_data

    def construct_v2_data(self, test_param, v2_data):
        v2_dict = {}
        _logger.info(f"v2_data - {v2_data}")
        for data in v2_data["results"]:
            if data[test_param] not in v2_dict:
                # Assign the initial values of process_guid/event_id
                v2_dict[data[test_param]] = data
        return v2_dict

    def get_device_timestamp_list(self, v2_data):
        device_timestamp_list = []
        _logger.info(f"v2_data - {v2_data}")
        for data in v2_data["results"]:
            timestamp = data["device_timestamp"]
            try:
                timestamp = time.mktime(datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S.%fZ").timetuple())
            except ValueError:
                timestamp = time.mktime(datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%SZ").timetuple())
            if timestamp not in device_timestamp_list:
                device_timestamp_list.append(int(timestamp))
        return sorted(device_timestamp_list)

    def get_v2_data_with_time_range(
        self,
        config,
        load_org_1,
        query,
        schema_fields,
        device_timestamp_list,
        test_param,
    ):
        max_len = len(device_timestamp_list)
        for i in range(0, int(max_len)):
            min_device_ts = time.strftime("%Y-%m-%dT%H:%M:%S.000Z", time.gmtime(device_timestamp_list[i]))
            max_device_ts = time.strftime(
                "%Y-%m-%dT%H:%M:%S.000Z",
                time.gmtime(device_timestamp_list[(max_len - 1) - i]),
            )
            device_ts_range = {"start": min_device_ts, "end": max_device_ts}
            _logger.info(f"Device timestamp range: {device_ts_range}")

            # POST V2 query and get data
            _logger.info(f"V2 POST to get query_id for {query}".center(100, "="))
            v2 = SearchAPIV2(
                privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
                org_id=load_org_1["loader"].org_id,
            )
            if test_param == "process_guid":
                v2_job_id = v2.create_process_result_v2(
                    query=query, fields=schema_fields, time_range=device_ts_range
                ).response["job_id"]
            else:
                v2_job_id = v2.create_enriched_events_result(
                    query=query, fields=schema_fields, time_range=device_ts_range
                ).response["job_id"]

            # Supporting SNS/ SQS Mode - If Job ID does not -sqs, let's append it to Job ID.
            v2_job_id = get_sqs_job_id(v2_job_id)

            _logger.info(f"V2 job_id : {v2_job_id}")

            # wait_until v2 query gets data
            # pylint: disable=cell-var-from-loop
            def get_v2_data():
                v2_data_inner = v2.get_process_result_v2(v2_job_id, row_count="10000000").response
                if v2_data_inner["contacted"] == v2_data_inner["completed"]:
                    return v2_data_inner

                return False

            v2_data = wait_until(get_v2_data, timeout=90)

            assert v2_data, f"V2 process search did not complete in 90 seconds. Job ID: {v2_job_id}"

            if v2_data["num_found"] == 0:
                pytest.fail(f"ERROR: V2 query's num_found is 0 | Response: {v2_data}")

            if v2_data["num_found"] == v2_data["num_available"]:
                return [v2_data, min_device_ts, max_device_ts]

            _logger.warning(
                f"num_found: {v2_data['num_found']} != num_available: {v2_data['num_available']} \
                                                    | Retrying search with narrow timestamp..."
            )
        return False

    def reset_all_favored(self, schema_data, existing_data, new_data):
        # Ex: 'process_hash': ['string', 'allfavored']
        for key, value in schema_data.items():
            if value[1] == "allfavored":
                if key in new_data:
                    existing_data[key] = new_data[key]
        return existing_data

    def delete_first_search_results(self, config, org_id, v2_job_id):
        """
        Deletes all results from the first search of a search job.
        If called within the retry period, it will prompt SearchAPI to start a retry of that search.
        """

        # Supporting SNS/ SQS Mode - If Job ID does not -sqs, let's append it to Job ID.
        if "sqs" not in v2_job_id:
            v2_job_id += "-sqs"

        # client to interact with S3
        s3_client = boto3.client("s3")

        # should we be getting these from env variables?
        # S3_SOLRSEARCHRESULTSBUCKETNAME, solr_results_bucket, aws_s3_results_bucket
        if not is_k8s() or is_kc():
            s3_results_bucket = f"{config['env_vars']['ENV_NAME']}-solr-results-bucket"
        else:
            s3_results_bucket = f"{config['env_vars']['ENV_NAME']}-psc-lc-shared-infra-s3results"
        # S3_SEARCHRESULTSKEYPREFIX, aws_s3_results_key_prefix
        s3_results_prefix = "search"

        # SearchAPI computes the query ID (and S3 key for search data) using this logic
        query_id = f"{org_id}/{v2_job_id}"
        s3_results_key = f"{s3_results_prefix}/{query_id}/0"

        _logger.info(f"Deleting objects in bucket {s3_results_bucket} with key prefix {s3_results_key}")

        # delete all objects in the directory
        res = s3_client.list_objects(Bucket=s3_results_bucket, Prefix=s3_results_key)

        objects_to_delete = [{"Key": obj["Key"]} for obj in res["Contents"]]

        s3_client.delete_objects(
            Bucket=s3_results_bucket,
            Delete={"Objects": objects_to_delete},
        )


def _wait_for_org_queues(prefix, org_id):
    sqs = boto3.client("sqs")
    kwargs = {"MaxResults": 1000, "QueueNamePrefix": prefix}
    _logger.info(f"Getting queues with prefix: {prefix}")
    queue_wait_retries = 0
    while True:
        err_retries = 0
        while True:
            try:
                queues = sqs.list_queues(**kwargs)
            except ClientError as e:
                _log_and_sleep(str(e), err_retries, max_sleep=120)
            except sqs.exceptions.InternalErrorException as e:
                _log_and_sleep(str(e), err_retries, max_sleep=120)
            else:
                break
            err_retries += 1
        if "QueueUrls" in queues:
            for queue_url in queues["QueueUrls"]:
                if org_id in queue_url:
                    return
            try:
                kwargs["NextToken"] = queues["NextToken"]
            except KeyError:
                break
        else:
            if queue_wait_retries > 20:
                raise Exception(f"Queues with prefix {prefix} not created after {queue_wait_retries} retries")

            _log_and_sleep(
                f"No queues found with prefix {prefix}",
                queue_wait_retries,
                max_sleep=300,
            )
            queue_wait_retries += 1


def _get_sqs_prefix(env):
    return env + SEPARATOR + SQS_SERVICE


def _log_and_sleep(message, retries=0, base_sleep=0.1, max_sleep=120):
    """
    Utility function to use for exponential backoff when retrying
    """
    sleep_time = min((2**retries) * base_sleep, max_sleep)
    _logger.info(f"{message}. Sleeping for {sleep_time} seconds")
    time.sleep(sleep_time)
    return message
