import logging
from typing import Optional

import time

from pscr_integration.test_lifecycle.active_session import is_smoke

_logger = logging.getLogger(__name__)


class Searcher:
    # Retry refetches much longer on smoke tests because searcher upgrades that may have happened right before may
    # take a bit to accomplish
    if is_smoke():
        REFETCH_RETRIES = 180
    else:
        REFETCH_RETRIES = 10

    def __init__(self, org_id: str):
        self.org_id = org_id

    def submit_query(self, query_string, comprehensive_search=False, facet=None):
        raise NotImplementedError

    def get_results(self, query_id: str, rows: Optional[int] = None):
        raise NotImplementedError

    def historically_tag_feed(self, watchlist_id, report_id, ioc_id: Optional[str] = None):
        raise NotImplementedError

    def get_events(self, query_string, process_guid, facet=None):
        raise NotImplementedError

    def is_error(self, query_data):
        raise NotImplementedError

    def is_no_searchers_available_error(self, query_data):
        raise NotImplementedError

    def is_success(self, query_data):
        raise NotImplementedError

    def get_status(self, query_id: str):
        raise NotImplementedError

    def get_query_ids(self):
        raise NotImplementedError

    def get_process_tree(self, process_guid: str, parent_guid: Optional[str] = None):
        raise NotImplementedError

    def get_process_limits(self):
        raise NotImplementedError

    def validate_query(self, query):
        raise NotImplementedError

    def count_query(
        self,
        timeout,
        query,
        result_count,
        stage=0,
        greater_than_ok=True,
        comprehensive_search=False,
        refech_retries=0,
    ):
        _logger.debug("Query: %s", query)
        query_data = self.submit_query(query, comprehensive_search=comprehensive_search)
        i = 0
        if refech_retries < self.REFETCH_RETRIES:
            refech_retries = self.REFETCH_RETRIES

        if is_smoke() and stage > 0:
            raise Exception("Smoke tests can't detect data in stages, always expect data to be returned in stage 0")

        while True:
            if self.is_error(query_data):
                if self.is_no_searchers_available_error(query_data):
                    _logger.info("org_id: %s -- No searchers available", self.org_id)
                    i += 1
                    if i > timeout:
                        raise Exception("Timed out waiting for searchers to be available")
                    time.sleep(1)
                else:
                    _logger.info(
                        "org_id: %s -- Error from api %s, error code: %s",
                        self.org_id,
                        query_data.response,
                        query_data.status_code,
                    )
                    raise Exception(f"Error from api {query_data.response}, error code {query_data.status_code}")
            elif self.is_success(query_data):
                if not query_data.response.get("query_id"):
                    raise Exception(f"Did not get query_id in response: {query_data.response}")
                print(f"org_id={self.org_id} status_code={query_data.status_code} response={query_data.response}")
                query_results_data = self.get_results(query_data.response["query_id"])
                j = 0
                while True:
                    _logger.info(
                        "org_id: %s -- %d Results Returned",
                        self.org_id,
                        query_results_data.response["response_header"]["num_found"],
                    )
                    offline = 0
                    if not is_smoke():
                        for searcher in query_data.response["debug"]["searchers"]:
                            _logger.info(
                                "org_id: %s, Query ID: %s,  index stage: %d, "
                                "start-time: %d, end-time: %d, endpoint: %s",
                                self.org_id,
                                searcher["searcher_id"],
                                searcher["index_stage"],
                                searcher["starttime"],
                                searcher["endtime"],
                                searcher["endpoint"],
                            )
                            if searcher["endpoint"] == "":
                                offline += 1
                    if (
                        query_results_data.response["response_header"]["searchers_meta"]["contacted"] - offline
                        != query_results_data.response["response_header"]["searchers_meta"]["completed"]
                    ):
                        _logger.info(
                            "Not all searchers returned: %d contacted, %d completed, %d offline",
                            query_results_data.response["response_header"]["searchers_meta"]["contacted"],
                            query_results_data.response["response_header"]["searchers_meta"]["completed"],
                            offline,
                        )
                        if j > refech_retries:
                            _logger.error("Too many refetch tries")
                            raise Exception("Too many refetch tries")
                        time.sleep(1)
                        _logger.info("Refetching results")
                        query_results_data = self.get_results(query_data.response["query_id"])
                        j += 1
                    else:
                        _logger.debug("org_id: %s -- successful query", self.org_id)
                        break
                if not is_smoke():
                    _logger.debug(
                        "query num_found: %d, query stage %d",
                        query_results_data.response["response_header"]["num_found"],
                        query_data.response["debug"]["searchers"][0]["index_stage"],
                    )
                _logger.debug("passed count: %d, passed stage %d", result_count, stage)
                if greater_than_ok:
                    if is_smoke():
                        if query_results_data.response["response_header"]["num_found"] >= result_count:
                            break
                    else:
                        if (
                            query_results_data.response["response_header"]["num_found"] >= result_count
                            and query_data.response["debug"]["searchers"][0]["index_stage"] >= stage
                        ):
                            break

                else:
                    if is_smoke():
                        if query_results_data.response["response_header"]["num_found"] == result_count:
                            break
                    else:
                        if (
                            query_results_data.response["response_header"]["num_found"] == result_count
                            and query_data.response["debug"]["searchers"][0]["index_stage"] >= stage
                        ):
                            break
                _logger.debug(f"retry: {i}, timeout {timeout}")
                i += 1
                if i > timeout:
                    raise Exception(
                        f"Time out waiting for results to complete, found "
                        f"{query_results_data.response['response_header']['num_found']} results"
                    )
                time.sleep(1)
            else:
                _logger.error("Did not receive an error, but also did not get successful query message")
                raise Exception("Did not receive an error, but also did not get successful query message")
            query_data = self.submit_query(query)
        return result_count
