import logging
from typing import Optional

from pscr_integration.api.api_request import HTTPApiResponse
from pscr_integration.api.search_api import SearchAPI
from pscr_integration.search.searcher import Searcher
from pscr_integration.search.searcher_type import FacetParams

_logger = logging.getLogger(__name__)


class SearcherService(Searcher):
    def __init__(
        self,
        private_link_hostname: str,
        org_id: str,
        device_id_start: int,
        device_id_end: int,
    ):
        super().__init__(org_id)
        self.search_service = SearchAPI(private_link_hostname, org_id, device_id_start, device_id_end)

    def submit_query(
        self,
        query_string: str,
        comprehensive_search: bool = False,
        facet: FacetParams = None,
    ):
        return self.search_service.create_process_result(
            q=query_string, comprehensive_search=comprehensive_search, facet=facet
        )

    def get_results(self, query_id: str, rows: Optional[int] = None) -> HTTPApiResponse:
        return self.search_service.get_process_result(query_id=query_id, row_count=rows)

    def historically_tag_feed(self, watchlist_id: str, report_id: str, ioc_id: Optional[str] = None):
        return self.search_service.evaluate_processes(watchlist_id=watchlist_id, report_id=report_id, ioc_id=ioc_id)

    def get_events(self, query_string: str, process_guid: str, facet: FacetParams = None):
        return self.search_service.get_events(q=query_string, cb_process_guid=process_guid, facet=facet)

    def is_error(self, query_data: HTTPApiResponse):
        return query_data.status_code >= 400 or self.is_no_searchers_available_error(query_data)

    def is_no_searchers_available_error(self, query_data: HTTPApiResponse):
        return "translation_key" in query_data.response and query_data.response["translation_key"] in [
            "threathunter_search_no_data_available",
            "threathunter_search_no_searchers_error",
        ]

    def is_success(self, query_data: HTTPApiResponse):
        return query_data.status_code == 200 and not self.is_no_searchers_available_error(query_data)

    def get_status(self, query_id: str):
        return self.search_service.get_process_result_status(query_id)

    def get_query_ids(self):
        return self.search_service.get_process_query_ids()

    def get_process_tree(self, process_guid: str, parent_guid: Optional[str] = None):
        return self.search_service.get_process_tree(process_guid, parent_guid)

    def get_process_limits(self):
        return self.search_service.get_process_limits()

    def validate_query(self, query):
        return self.search_service.get_process_validation(query=query)
