###################################################################################
# This script is intended to test Federated Search functionality
# Process & Event searches with Customer Group tests
# Validate API users can search "data" about a process or an event using Federated Search API routes.
# Validate All returned data fields are aggregated in a way that preserves all relevant data
# Test includes:
# 1] Process search for a group of  process_guids
# 2] Event search for a group of event_ids
# #################################################################################

import logging
import string
import random
import time

import pytest

from pscr_integration.api.federated_search_api import FederatedSearchAPI
from pscr_integration.api.search_api_v2 import SearchAPIV2
from pscr_integration.search.search_common import SearchCommon
from pscr_integration.util import wait_until
from pscr_integration.lc_services import LcServices

_logger = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def federated_search_api(config):
    yield FederatedSearchAPI(
        privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
    )


def _get_random_alphanumeric_string(length: int = 7):
    letters_and_digits = string.ascii_letters + string.digits
    result_str = "".join((random.choice(letters_and_digits) for i in range(length)))
    return result_str


def _get_timestamp_lists(config, load_org_1, query, schema_fields):
    _logger.info("Get timestamps for federated searching".center(100, "="))
    # POST V2 query and get data
    _logger.info(f"V2 POST to get query_id for {query}".center(100, "="))
    v2 = SearchAPIV2(
        privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
        org_id=load_org_1["loader"].org_id,
    )
    v2_job_id = v2.create_process_result_v2(query=query, fields=schema_fields).response["job_id"]
    _logger.info(f"V2 job_id : {v2_job_id}")

    # wait_until v2 query gets data
    def get_v2_data():
        v2_data = v2.get_process_result_v2(v2_job_id, row_count="10000000").response
        if v2_data["contacted"] == v2_data["completed"]:
            return v2_data
        return False

    v2_data = wait_until(get_v2_data, timeout=180)
    if not v2_data:
        pytest.fail(f"ERROR: V2 process search did not complete in 90 seconds. Job ID: {v2_job_id}\n")
    sc = SearchCommon()
    return sc.get_device_timestamp_list(v2_data)


@pytest.mark.skipif(
    LcServices.target_service() not in LcServices.federated_search_services(),
    reason=f"Test is irrelevant for f{LcServices.target_service()}.",
)
@pytest.mark.smoke_test
class TestFederatedSearch:
    def test_customer_groups(self, federated_search_api):
        group_name = "group1" + _get_random_alphanumeric_string()
        org_keys = ["org1", "org2"]
        created_group = federated_search_api.upsert_customer_group(group_name, org_keys=org_keys).response

        assert created_group["name"] == group_name
        assert created_group["org_keys"]
        assert created_group["org_keys"].sort() == org_keys.sort()

        get_group = federated_search_api.get_customer_group(created_group["name"]).response

        assert get_group["name"] == group_name
        assert get_group["org_keys"].sort() == org_keys.sort()

        get_groups = federated_search_api.get_customer_groups().response

        assert len(get_groups) > 0
        assert any(group["name"] == group_name for group in get_groups)

        org_keys.append("org3")

        update_group = federated_search_api.upsert_customer_group(group_name, org_keys=org_keys).response

        assert created_group["name"] == update_group["name"]
        assert all(result in update_group["org_keys"] for result in created_group["org_keys"])
        assert "org3" in update_group["org_keys"]

        get_group_final = federated_search_api.get_customer_group(created_group["name"]).response

        assert get_group_final["org_keys"]
        assert get_group_final["org_keys"].sort() == org_keys.sort()

    @pytest.mark.parametrize(
        "search_type",
        [FederatedSearchAPI.TYPE_PROCESS, FederatedSearchAPI.TYPE_ENRICHED_EVENTS],
    )
    def test_federated_search(self, federated_search_api, search_type, config, load_org_1):
        query = "*:*"
        sort = [{"field": "device_timestamp", "order": "DESC"}]

        # Download and parse latest schema
        _logger.info("Download and parse latest schema")
        sc = SearchCommon()
        schema_data = sc.get_schema(config, load_org_1, query)
        schema_fields = list(schema_data.keys())

        timestamp_list = _get_timestamp_lists(config, load_org_1, query, schema_fields)

        assert len(timestamp_list) > 1

        start_time = time.strftime("%Y-%m-%dT%H:%M:%S.000Z", time.gmtime(timestamp_list[0]))
        end_time = time.strftime(
            "%Y-%m-%dT%H:%M:%S.000Z",
            time.gmtime(timestamp_list[(len(timestamp_list) - 1)]),
        )

        # POST search
        _logger.info(f"POST to get federated job_id for {query}".center(100, "="))
        create_search = federated_search_api.create_search(
            search_type=search_type,
            query=query,
            fields=schema_fields,
            org_keys=[load_org_1["loader"].org_id],
            rows=100,
            sort=sort,
            time_range={"start": start_time, "end": end_time},
        ).response

        assert create_search["job_id"]

        job_id = create_search["job_id"]

        _logger.info(f"Federated job_id : {job_id}")

        # wait_until search is completed
        def get_status():
            status_response = federated_search_api.get_search_status(search_type, job_id).response
            if status_response["status"] == "COMPLETED":
                return status_response
            return False

        status = wait_until(get_status, timeout=180)
        if not status:
            pytest.fail(f"ERROR: Federated {search_type} search did not complete in 90 seconds. Job ID: {job_id}\n")

        if status["orgs_completed"] == 0:
            pytest.fail(f"ERROR: V1 query's orgs_completed is 0 | Response: {status}\n")

        if status["orgs_total"] != status["orgs_completed"]:
            pytest.fail(f"ERROR: V1 search orgs completed is different than total | Response: {status}\n")

        stats = federated_search_api.get_search_stats(search_type, job_id).response

        assert stats["status"]
        assert stats["orgs_total"]
        assert stats["orgs_completed"]
        assert stats["orgs_completed"] == status["orgs_completed"]
        assert stats["orgs_total"] == status["orgs_total"]
        assert stats["status"] == status["status"]
        assert len(stats["orgs"]) > 0
        for org_data in stats["orgs"]:
            assert org_data["org_key"]
            assert org_data["status"] == "COMPLETED"
            assert org_data["event_hit_count"]
            assert org_data["results_count"]
            assert org_data["min_solr_processing_time"]
            assert org_data["max_solr_processing_time"]
            assert org_data["avg_solr_processing_time"]

        result_response = federated_search_api.get_search_results(search_type, job_id).response

        assert result_response["status"]
        assert result_response["orgs_total"]
        assert result_response["orgs_completed"]
        assert len(result_response["orgs"]) == len(stats["orgs"])
        assert result_response["orgs_completed"] == status["orgs_completed"]
        assert result_response["orgs_total"] == status["orgs_total"]
        assert result_response["status"] == status["status"]
        assert len(result_response["orgs"]) > 0
        for org_data in result_response["orgs"]:
            assert org_data["org_key"]
            assert org_data["status"] == "COMPLETED"
            assert org_data["event_hit_count"]
            assert org_data["results_count"]
            assert org_data["min_solr_processing_time"]
            assert org_data["max_solr_processing_time"]
            assert org_data["avg_solr_processing_time"]
            assert len(org_data["results"]) > 0, f"Search has no results found. Data: {org_data}"
            for result in org_data["results"]:
                assert "device_policy_id" in result

        csr_response = federated_search_api.get_csr(search_type, job_id).response

        _logger.info(f"Response for CSR {csr_response}")

        # assert len(csr_response) > 0
        # assert csr_response[load_org_1["loader"].org_id]
