###################################################################################
# This script is intended to test data integrity of process summary by comparing
# v2 results of process_guids with v1 by validating primary fields of process,
# sibling, children and parent.
# #################################################################################

import logging
import time
from collections import Counter
from datetime import datetime
import pytest
from pscr_integration.api.search_api import SearchAPI
from pscr_integration.api.search_api_v2 import SearchAPIV2
from pscr_integration.util import wait_until
from pscr_integration.lc_services import LcServices

_logger = logging.getLogger(__name__)


@pytest.mark.skipif(
    LcServices.target_service() not in LcServices.search_services(),
    reason=f"Test is irrelevant for f{LcServices.target_service()}.",
)
@pytest.mark.smoke_test
class TestProcessSummary:
    def get_process_guids(self, v2_data):
        process_guid_dict = {}
        for data in v2_data["results"]:
            if data["process_guid"] not in process_guid_dict:
                if "parent_guid" in data:
                    process_guid_dict[data["process_guid"]] = data["parent_guid"]
                else:
                    process_guid_dict[data["process_guid"]] = None
        return process_guid_dict

    def construct_v1_data(self, results):
        exception = False
        data_dict = {}
        type_list = ["process", "siblings", "parent", "children"]
        # Some results don't have this field at all, so check existence first
        if "exception" in results:
            if results["exception"]:
                _logger.warning(f"Exception v1: {results['exception']}\n")
                exception = True
        for data in results:
            if data in type_list:
                data_dict[data] = results[data]
        return [exception, data_dict]

    def construct_v2_data(self, results):
        exception = False
        data_dict = {}
        type_list = ["process", "siblings", "parent", "children"]
        if "exception" in results:
            if results["exception"]:
                _logger.warning(f"Exception v2: {results['exception']}\n")
                exception = True
        # Formulate data struct
        for data in results:
            if data in type_list:
                data_dict[data] = results[data]
        return [exception, data_dict]

    def compare_timestamp(self, time1, time2):
        try:
            time1_value = datetime.strptime(time1, "%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            # To cater timestamp without milliseconds.
            time1_value = datetime.strptime(time1, "%Y-%m-%dT%H:%M:%SZ")
        try:
            time2_value = datetime.strptime(time2, "%Y-%m-%dT%H:%M:%S.%fZ")
        except ValueError:
            time2_value = datetime.strptime(time2, "%Y-%m-%dT%H:%M:%SZ")
        op = time.mktime(time1_value.timetuple()) - time.mktime(time2_value.timetuple())
        return op

    def compare_data(self, v1_dict, v2_dict, primary_fields, msg="ERROR"):
        total_fields = 0
        field_missing = 0
        match = 0
        mismatch = 0
        passed = True
        result_store = []
        timestamp_fields = [
            "backend_timestamp",
            "device_timestamp",
            "process_start_time",
        ]
        boolean_fields = ["has_children", "hits"]
        # As per earlier discussions, making this change to not flag expected failure
        if self.compare_timestamp(v1_dict["device_timestamp"], v2_dict["device_timestamp"]) < 0:
            msg = "WARNING"
        elif "enriched" in v1_dict and "enriched" not in v2_dict:
            msg = "WARNING"
        for field in v1_dict:
            if field not in primary_fields:
                continue
            total_fields += 1
            if field not in v2_dict:
                # As per earlier discussions, making this change:
                # If v1's has_children = False (or) hits = False, v2 doesn't have that field itself.
                if field in boolean_fields and not v1_dict[field]:
                    total_fields -= 1
                    continue
                passed = False
                field_missing += 1
                result_store.append(f"{field} not found in v2\n")
            else:
                if field in timestamp_fields:
                    if self.compare_timestamp(v1_dict[field], v2_dict[field]) != 0:
                        passed = False
                        mismatch += 1
                    # Any mismatch from device_timestamp is decided to be considered WARNING.
                    if field == "device_timestamp":
                        msg = "WARNING"
                        result_store.append(
                            f"{msg}: Mismatch found in {field} => v1: {v1_dict[field]} | v2: {v2_dict[field]}\n"
                        )
                    else:
                        match += 1
                elif v1_dict[field] != v2_dict[field]:
                    # Ex.process_hash
                    if isinstance(v2_dict[field], list) and len(v2_dict[field]) > len(v1_dict[field]):
                        match += 1
                        continue
                    # Two identical list
                    if isinstance(v1_dict[field], list) and isinstance(v2_dict[field], list):
                        if set(v1_dict[field]) == set(v2_dict[field]):
                            match += 1
                            continue
                    # If v1 is false and v2 is true it's decided to mark as WARNING.
                    if field in boolean_fields and not v1_dict[field]:
                        msg = "WARNING"
                    # Any mismatch from process_reputation is decided to be considered WARNING.
                    if field == "process_reputation":
                        msg = "WARNING"
                    passed = False
                    mismatch += 1
                    result_store.append(f"Mismatch found in {field} => v1: {v1_dict[field]} | v2: {v2_dict[field]}\n")
                else:
                    match += 1
        # Some expected failures are suppressed to WARNING (to debug if required)
        if msg == "WARNING":
            passed = True
            for line in result_store:
                _logger.warning(line)
        else:
            for line in result_store:
                _logger.error(line)
        if not passed:
            # To debug mismatch from logs
            _logger.info(f"\nv1 => {v1_dict}\n")
            _logger.info(f"\nv2 => {v2_dict}\n")
        return [total_fields, match, mismatch, field_missing, passed]

    def verify_data_integrity(self, exception, v1_data, v2_data, primary_fields):
        process_dict = {"total": 0, "passed": 0, "failed": 0}
        parent_dict = {"total": 0, "passed": 0, "failed": 0}
        sibling_dict = {"total": 0, "passed": 0, "failed": 0}
        children_dict = {"total": 0, "passed": 0, "failed": 0}
        type_count = {
            "process": process_dict,
            "parent": parent_dict,
            "siblings": sibling_dict,
            "children": children_dict,
        }
        for v1_key, v1_value in v1_data.items():
            if v1_key not in v2_data:
                _logger.error(f"{v1_key} type not in entire v2_data")
                continue
            v2_value = v2_data[v1_key]
            if not v1_value or not v2_value:
                _logger.warning(f"Empty value => v1 {v1_key}: {v1_value} | v2 {v1_key}: {v2_value}")
                continue
            if not isinstance(v1_value, list):
                v1_value = [v1_value]
            if not isinstance(v2_value, list):
                v2_value = [v2_value]
            # Find duplicate process_guids in v2 result if any
            dup_finder_dict = Counter(v2_dict["process_guid"] for v2_dict in v2_value)
            for dkey, dvalue in dup_finder_dict.items():
                if dvalue > 1:
                    _logger.error(f"{dkey} found {dvalue} times in {v1_key}\n")
            for v1_dict in v1_value:
                total_fields = 0
                field_missing = 0
                match = 0
                mismatch = 0
                type_count[v1_key]["total"] += 1
                passed = True
                if "process_guid" not in v1_dict:
                    passed = False
                    _logger.error(f"'process_guid' not present in v1_value {v1_dict}")
                else:
                    _logger.info(f"\n*** Validating process_guid: {v1_dict['process_guid']} | Type: {v1_key} ***\n")
                    found_in_v2 = False
                    for v2_dict in v2_value:
                        if v1_dict["process_guid"] == v2_dict["process_guid"]:
                            if found_in_v2:
                                continue
                            found_in_v2 = True
                            (
                                total_fields,
                                match,
                                mismatch,
                                field_missing,
                                passed,
                            ) = self.compare_data(v1_dict, v2_dict, primary_fields)
                    if not found_in_v2:
                        if exception:
                            _logger.warning(f"process_guid {v1_dict['process_guid']} not found in v2\n")
                        else:
                            passed = False
                            _logger.error(f"process_guid {v1_dict['process_guid']} not found in v2\n")
                    else:
                        _logger.info(
                            f"\n**** Status: Total_fields: {total_fields}, Match: {match}, \
                                            Mismatch: {mismatch}, Missing: {field_missing} ****\n"
                        )
                if passed:
                    type_count[v1_key]["passed"] += 1
                else:
                    type_count[v1_key]["failed"] += 1
        return type_count

    def test_process_summary(self, config, load_org_1):
        error_count = 0
        p_guid_count = 0
        exception = False
        query = "process_guid: *"
        process_dict = {"total": 0, "passed": 0, "failed": 0}
        parent_dict = {"total": 0, "passed": 0, "failed": 0}
        sibling_dict = {"total": 0, "passed": 0, "failed": 0}
        children_dict = {"total": 0, "passed": 0, "failed": 0}
        total_count = {
            "process": process_dict,
            "parent": parent_dict,
            "siblings": sibling_dict,
            "children": children_dict,
        }
        primary_fields = [
            "device_id",
            "device_group_id",
            "device_policy",
            "device_os",
            "device_group",
            "device_timestamp",
            "process_guid",
            "process_cmdline",
            "process_pid",
            "process_hash",
            "process_reputation",
            "process_name",
            "hits",
            "has_children",
            "sensor_action",
        ]
        v1 = SearchAPI(
            privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
        )
        v2 = SearchAPIV2(
            privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
        )

        ## POST V2 search query and get data
        _logger.info(f"V2 POST to get query_id for {query}".center(100, "="))
        v2_job_id = v2.create_process_result_v2(query=query).response["job_id"]
        _logger.info(f"V2 search job_id : {v2_job_id}")

        # wait_until v2 query gets data
        def get_v2_process_data():
            v2_data = v2.get_process_result_v2(v2_job_id, row_count="10000000").response
            if v2_data["contacted"] == v2_data["completed"]:
                return v2_data
            return False

        v2_data = wait_until(get_v2_process_data, timeout=90, interval=30)
        if not v2_data:
            pytest.fail(f"V2 process search did not complete in 90 seconds. Job ID: {v2_job_id}")

        if v2_data["num_found"] == 0:
            pytest.fail(f"V2 query's num_found is 0 | Response: {v2_data}")

        # From the search result, construct the process_guids, parent_guids to test with
        process_guid_dict = self.get_process_guids(v2_data)
        _logger.info(f"\n Total process_guids: {len(process_guid_dict)}\n")
        _logger.info(f"\nprocess_guid_dict: {process_guid_dict}\n")
        for process_guid, parent_guid in process_guid_dict.items():
            # To reduce run time, test only 10 process_guids.
            if p_guid_count > 10:
                break
            p_guid_count += 1
            _logger.info(f"\n process_guid_num: {p_guid_count}\n")

            # If V1 end up with non-optimal result move to next process_guid
            # Ex. 404, BadStatusLine('No status line received'), empty result.
            v1_data = v1.get_process_summary(process_guid, parent_guid).response
            if "incomplete_results" not in v1_data:
                _logger.warning(f"Unexpected v1 data => {v1_data}")
                continue

            # Get V1 data, wait_until v1 query gets data
            def get_v1_data(process_guid=process_guid, parent_guid=parent_guid):
                v1_data = v1.get_process_summary(process_guid, parent_guid).response
                if not v1_data["incomplete_results"]:
                    return v1_data
                return False

            v1_data = wait_until(get_v1_data, timeout=90, interval=30)
            if not v1_data:
                _logger.warning(f"Unexpected v1 data => {v1_data}. Moving to next one..")
                continue

            # POST V2 summary get data
            _logger.info(f"V2 POST to get job_id for {process_guid}".center(100, "="))
            v2_job_id = v2.create_process_summary_v2(process_guid=process_guid, parent_guid=parent_guid).response[
                "job_id"
            ]
            _logger.info(f"V2 process summary job_id : {v2_job_id}")

            def get_v2_data():
                v2_data = v2.get_process_summary_v2(v2_job_id).response
                if v2_data["contacted"] == v2_data["completed"] and v2_data["completed"] != 0:
                    return v2_data["summary"]
                _logger.warning(f"v2 data => {v2_data}")
                return False

            v2_data = wait_until(get_v2_data, timeout=90, interval=30)
            if not v2_data:
                _logger.warning(f"Unexpected v2 data => {v2_data}. Moving to next one..")
                continue

            # Construct v1/v2 data compatible with process summary result format
            v1_exception, v1_data = self.construct_v1_data(v1_data)
            v2_exception, v2_data = self.construct_v2_data(v2_data)
            if v1_exception or v2_exception:
                exception = True
            _logger.debug(f"\nV1 data => {v1_data}\n")
            _logger.debug(f"\nV2 data => {v2_data}\n")

            # Metric for v1/v2 data
            _logger.info(f"\n ************** v1 and v2 metric for process_guid: {process_guid} ***************\n")
            for type in total_count:
                v1_value = v1_data[type]
                v2_value = v2_data[type]
                if not v1_value:
                    v1_value = []
                if not v2_value:
                    v2_value = []
                if not isinstance(v1_value, list):
                    v1_value = [v1_value]
                if not isinstance(v2_value, list):
                    v2_value = [v2_value]
                _logger.info(f"{type} => v1: {len(v1_value)} | v2: {len(v2_value)}\n")
                if len(v1_value) > len(v2_value):
                    _logger.error(f"ERROR: v2 seem to be missing data for type {type}")
                    error_count += 1

            # Verify data integrity
            op = self.verify_data_integrity(exception, v1_data, v2_data, primary_fields)
            _logger.info(f"\n ************** Test summary for process_guid: {process_guid} ***************\n")
            for key, value in op.items():
                total_count[key]["total"] += value["total"]
                total_count[key]["passed"] += value["passed"]
                total_count[key]["failed"] += value["failed"]
                _logger.info(
                    f"\n{key} => Total: {value['total']}, Passed: {value['passed']}, Failed: {value['failed']}\n"
                )

        _logger.info("\n ******************* Overall summary of categories tested *********************\n")
        for key, value in total_count.items():
            error_count += value["failed"]
            _logger.info(f"{key} => Total: {value['total']}, Passed: {value['passed']}, Failed: {value['failed']}\n")
        if error_count:
            pytest.fail(f"Test encountered total of {error_count} errors/failures")
