import logging
import time
import pytest
from pscr_integration.api.search_api_v2 import SearchAPIV2
from pscr_integration.search.search_common import SearchCommon
from pscr_integration.util import wait_until

_logger = logging.getLogger(__name__)


@pytest.mark.skip("Retries are irrelevant in SQS Only mode")
class TestRetry:
    # how many seconds SearchAPI waits until retrying a search
    retry_time = 90

    def test_retry_process_search(self, config, load_org_1):
        test_param = "process_guid"
        query = test_param + ":*"

        # Download and parse latest schema
        _logger.info("Download and parse latest schema")
        sc = SearchCommon()
        schema_data = sc.get_schema(config, load_org_1, query)
        schema_fields = list(schema_data.keys())

        # POST V2 query
        _logger.info(f"S@S V2 POST to get query_id for {query}".center(100, "="))
        v2 = SearchAPIV2(
            privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
        )

        search_start_time = time.time()

        v2_job_id = v2.create_process_result_v2(query=query, fields=schema_fields).response["job_id"]
        _logger.info(f"V2 job_id : {v2_job_id}")

        # wait_until v2 query gets data
        def get_v2_data():
            v2_data = v2.get_process_result_v2(v2_job_id, row_count="10000000").response
            if "contacted" in v2_data and v2_data["contacted"] == v2_data["completed"]:
                return v2_data
            return False

        v2_data_first_try = wait_until(get_v2_data, timeout=self.retry_time)
        if not v2_data_first_try:
            pytest.fail(f"ERROR: V2 process search retry did not complete in 90 seconds. Job ID: {v2_job_id}\n")

        # delete all files from the first search to prompt SearchAPI to start a retry
        sc.delete_first_search_results(config, v2.org_id, v2_job_id)

        # now wait for the retry search to start and complete
        # (it is expected to see 404 errors until the retry search starts)
        poll_period = 5
        v2_data_second_try = wait_until(
            get_v2_data,
            timeout=(self.retry_time * 2) // poll_period,
            interval=poll_period,
        )

        search_end_time = time.time()
        search_time = search_end_time - search_start_time

        # if we got our results earlier, then we failed to break the first search
        assert search_time > self.retry_time, f"Search finished in {search_time}s, before a retry was started"

        if not v2_data_second_try:
            pytest.fail(f"ERROR: V2 process search retry did not complete in 90 seconds. Job ID: {v2_job_id}\n")
        if v2_data_second_try["num_found"] == 0:
            pytest.fail(f"ERROR: V2 query's num_found is 0 | Response: {v2_data_second_try}\n")

        _logger.info("V2 process search retry completed successfully")
