###################################################################################
# This script is intended to test data integrity of process and event search
# by comparing the v2 results with v1 and by validating the aggregation strategy
# as defined in the schema.
# Test includes:
# 1] Process search for all process_guid's
# 2] Event search for all event_id's
# #################################################################################
import logging
import re
import time
from datetime import datetime
import pytest
from pscr_integration.api.search_api import SearchAPI
from pscr_integration.api.search_api_v2 import SearchAPIV2
from pscr_integration.search.search_common import SearchCommon
from pscr_integration.util import wait_until
from pscr_integration.lc_services import LcServices

_logger = logging.getLogger(__name__)


@pytest.mark.skipif(
    LcServices.target_service() not in LcServices.search_services(),
    reason=f"Test is irrelevant for f{LcServices.target_service()}.",
)
class TestSearchAtScale:
    def construct_v1_data(self, test_param, v1_data, schema_data):
        v1_dict = {}
        sc = SearchCommon()
        for data in v1_data["data"]:
            if data[test_param] not in v1_dict:
                # Assign the initial values of process_guid/event_id
                v1_dict[data[test_param]] = data
                if "enriched" in v1_dict[data[test_param]]:
                    v1_dict[data[test_param]]["non-enriched"] = False
                else:
                    v1_dict[data[test_param]]["non-enriched"] = True
            else:
                # See if any fields need to be combined to the dict value
                existing_op = v1_dict[data[test_param]]
                keys = set(data.keys()) - set(existing_op.keys())
                for key in keys:
                    v1_dict[data[test_param]][key] = data[key]
                # Check values as per the schema such as find max/combine array.
                for key in data.keys():
                    if key in existing_op and key in schema_data:
                        existing_value = existing_op[key]
                        new_value = data[key]
                        schema_field = schema_data[key]
                        data_type = schema_field[0]
                        aggregation_type = schema_field[1]
                        if aggregation_type == "max" and data_type == "timestamp":
                            return_value = sc.compare_timestamp(existing_value, new_value)
                            if return_value < 0:
                                existing_op[key] = new_value
                        elif aggregation_type == "max" and data_type == "boolean":
                            if existing_value == "false" and new_value == "true":
                                existing_op[key] = new_value
                        elif aggregation_type == "max" and "int" in data_type:
                            if new_value > existing_value:
                                existing_op[key] = new_value
                        elif aggregation_type == "allfavored":
                            # Found non-enriched data, reset values to favor non-enriched
                            if not existing_op["non-enriched"] and "enriched" not in data:
                                existing_op["non-enriched"] = True
                                existing_op = sc.reset_all_favored(schema_data, existing_op, data)
                            # Aggregate further non-enriched unique values as encountered
                            elif existing_op["non-enriched"] and "enriched" not in data:
                                if existing_value != new_value:
                                    existing_op[key] = sc.compare_combine_array_values(existing_value, new_value)
                            # Treat as regular "all" category as long it's enriched only
                            elif not existing_op["non-enriched"] and "enriched" in data:
                                if existing_value != new_value:
                                    existing_op[key] = sc.compare_combine_array_values(existing_value, new_value)
                        # Limited - has a limit on no. of values it can hold.
                        # May not affect IT but revisit in case of failure
                        elif aggregation_type in ("all", "limited"):
                            if existing_value != new_value:
                                existing_op[key] = sc.compare_combine_array_values(existing_value, new_value)
                        elif aggregation_type == "any":
                            if not isinstance(existing_value, list):
                                existing_value = [str(existing_value)]
                            if str(new_value) not in existing_value:
                                existing_value.extend([str(new_value)])
                                existing_op[key] = existing_value
        return v1_dict

    def check_data_integrity(self, test_param, v1_dict, v2_dict, schema_data, ignored_fields):
        total_error_count = 0
        sc = SearchCommon()
        for v1_key, v1_value in v1_dict.items():
            _logger.info(f"\n\t\t*********  Checking {test_param}: {v1_key}  ***********\n")
            local_error_count = 0
            if v1_key in v2_dict:
                v2_value = v2_dict[v1_key]
                _logger.debug(f"V1 VALUE ->{v1_value}")
                _logger.debug(f"V2 VALUE ->{v2_value}")
            else:
                _logger.error(f"V2 is missing '{test_param}: {v1_key}' from V1\n")
                local_error_count += 1
                total_error_count += local_error_count
                continue
            # Validation criteria for each field in V1 and V2 data sets as per aggregration strategy in schema
            for key in v1_value:
                # Some fields no longer applicable for process_search/event_search.
                reg_list = map(re.compile, ignored_fields)
                if any(regex.match(key) for regex in reg_list):
                    continue
                if key in v2_value:
                    schema_field = schema_data[key]
                    data_type = schema_field[0]
                    aggregation_type = schema_field[1]
                    if data_type == "timestamp":
                        if sc.compare_timestamp(v1_value[key], v2_value[key]) > 0:
                            _logger.error(
                                f"*** V2's timestamp is not latest ***\n\n V1 => '{key}: {v1_value[key]}' \
                                                                            | V2 => '{key}: {v2_value[key]}'\n"
                            )
                            local_error_count += 1
                    elif aggregation_type in ("all", "limited", "allfavored"):
                        diff = set(v1_value[key]) - set(v2_value[key])
                        if len(diff) > 0:
                            _logger.error(f"*** Missing Values ***\n\n V2 '{key}' is missing {diff}\n")
                            local_error_count += 1
                    elif aggregation_type == "any":
                        if isinstance(v1_value[key], list):
                            if str(v2_value[key]) not in v1_value[key]:
                                _logger.error(
                                    f"*** Mismatch found ***\n\n V2 '{key}:{v2_value[key]}' didn't match \
                                                            any of the values in V1 for '{key}:{v1_value[key]}'\n"
                                )
                                local_error_count += 1
                        if isinstance(v1_value[key], str):
                            if str(v2_value[key]).strip() != v1_value[key].strip():
                                _logger.error(
                                    f"*** Mismatch found ***\n\n V1 => '{key}: {v1_value[key]}' \
                                                                    | V2 => '{key}: {v2_value[key]}'\n"
                                )
                                local_error_count += 1
                    elif aggregation_type == "enriched":
                        if str(v1_value[key]) not in v2_value[key]:
                            _logger.error(
                                f"*** Mismatch found ***\n\n V2 '{key}:{v2_value[key]}' doesn't \
                                                            contain V1's value '{key}:{v1_value[key]}'\n"
                            )
                            local_error_count += 1
                    # generic comparison for everything else
                    elif v1_value[key] != v2_value[key]:
                        _logger.error(
                            f"*** Mismatch found ***\n\n V1 => '{key}: {v1_value[key]}' \
                                                    | V2 => '{key}: {v2_value[key]}'\n"
                        )
                        local_error_count += 1
                else:
                    _logger.error(f"*** Missing Field ***\n\n V2 is missing '{key}: {v1_value[key]}'\n")
                    local_error_count += 1
            # Keep count of errors for each process_guid/event_search
            if not local_error_count:
                _logger.info("All the schema fields and values from V1 are verified in V2 successfully")
            else:
                _logger.error(f"{local_error_count} error found\n")
                total_error_count += local_error_count
        return total_error_count

    def test_process_search_at_scale(self, config, load_org_1):
        test_param = "process_guid"
        query = test_param + ":*"
        time_range = False
        collapse = "{!collapse field=process_collapse_id sort='max(0,legacy) asc,device_timestamp desc'}"
        ignored_fields = [
            "event_id",
            "event_type",
            "event_description",
            "childproc_guid",
            "event_network_",
            "document_guid",
            "non-enriched",
            # Remove cb.observationsSearchOnlyFields from comparison
            # in v2 search cb.observationsSearchOnlyFields  are available for INSIGHT category only
            # in v1 they are returned and need to be skipped during result comparison
            "observation_id",
            "observation_type",
            "attack_tactic",
            "attack_technique",
        ]

        # Download and parse latest schema
        _logger.info("Download and parse latest schema")
        sc = SearchCommon()
        schema_data = sc.get_schema(config, load_org_1, query)
        schema_fields = list(schema_data.keys())

        # POST V2 query and get data
        _logger.info(f"S@S V2 POST to get query_id for {query}".center(100, "="))
        v2 = SearchAPIV2(
            privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
        )
        v2_job_id = v2.create_process_result_v2(query=query, fields=schema_fields).response["job_id"]
        _logger.info(f"V2 job_id : {v2_job_id}")

        # wait_until v2 query gets data
        def get_v2_data():
            v2_data = v2.get_process_result_v2(v2_job_id, row_count="10000000").response
            if v2_data["contacted"] == v2_data["completed"]:
                return v2_data
            return False

        v2_data = wait_until(get_v2_data, timeout=90)
        if not v2_data:
            pytest.fail(f"ERROR: V2 process search did not complete in 90 seconds. Job ID: {v2_job_id}\n")

        if v2_data["num_found"] == 0:
            pytest.fail(f"ERROR: V2 query's num_found is 0 | Response: {v2_data}\n")

        # Check num_found == num_available before proceeding with the test.
        # PS: If num_found != num_available, chances we will get inconsistent results hence this check.
        if v2_data["num_found"] != v2_data["num_available"]:
            # Try narrowing the time window and check whether the condition is met
            timestamp_list = sc.get_device_timestamp_list(v2_data)
            op = sc.get_v2_data_with_time_range(config, load_org_1, query, schema_fields, timestamp_list, test_param)
            assert op, "V2's process search resulted in 'num_found' != 'num_available' even with adjusted time range"
            v2_data = op[0]
            min_ts = time.mktime(datetime.strptime(op[1], "%Y-%m-%dT%H:%M:%S.%fZ").timetuple())
            max_ts = time.mktime(datetime.strptime(op[2], "%Y-%m-%dT%H:%M:%S.%fZ").timetuple())
            time_range = True

        # POST V1 query and get data
        _logger.info(f"S@S V1 POST to get query_id for {query}".center(100, "="))
        v1 = SearchAPI(
            privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
        )
        if time_range:
            v1_query_id = v1.create_process_result(
                q=query,
                fl=",".join(schema_fields),
                fq=collapse,
                min_device_ts=min_ts,
                max_device_ts=max_ts,
            ).response["query_id"]
        else:
            v1_query_id = v1.create_process_result(q=query, fl=",".join(schema_fields), fq=collapse).response[
                "query_id"
            ]
        _logger.info(f"V1 query_id : {v1_query_id}")

        # wait_until v1 query gets data
        def get_v1_data():
            v1_data = v1.get_process_result(v1_query_id, row_count="10000000").response
            if (
                v1_data["response_header"]["searchers_meta"]["contacted"]
                == v1_data["response_header"]["searchers_meta"]["completed"]
            ):
                return v1_data
            return False

        v1_data = wait_until(get_v1_data, timeout=90)
        if not v1_data:
            pytest.fail(f"ERROR: V1 process search did not complete in 30 seconds. Query ID: {v1_query_id}")

        # Compute V1 data struct for comparison, end result can be compared 1/1 with V2.
        v1_dict = self.construct_v1_data(test_param, v1_data, schema_data)

        # Compute V2 data struct for comparison
        v2_dict = sc.construct_v2_data(test_param, v2_data)

        ## Data Comparison that goes over each field in V1/V2 for each process_guid/event_id
        error_count = self.check_data_integrity(test_param, v1_dict, v2_dict, schema_data, ignored_fields)
        ## PASS/FAIL criteria: Assert if error encountered during this test
        if not error_count:
            _logger.info("Search Aggregation validation for V2 completed successfully")
        else:
            pytest.fail(f"Encountered total of {error_count} error(s)")

    @pytest.mark.smoke_test
    def test_event_search_at_scale(self, config, load_org_1):
        test_param = "event_id"
        query = test_param + ":*"
        time_range = False
        collapse = '{!collapse field=event_id sort="device_timestamp desc"}'
        ignored_fields = [
            "childproc_guid",
            "document_guid",
            "non-enriched",
            # Remove cb.insightSearchOnlyFields from v1_data
            # in v2 search cb.observationsSearchOnlyFields  are available for INSIGHT category only
            # in v1 they are returned and need to be skipped during result comparison
            "observation_id",
            "observation_type",
            "attack_tactic",
            "attack_technique",
        ]

        # Download and parse latest schema
        _logger.info("Download and parse latest schema")
        sc = SearchCommon()
        schema_data = sc.get_schema(config, load_org_1, query)
        schema_fields = list(schema_data.keys())

        ## POST V2 query and get data
        _logger.info(f"S@S V2 POST to get query_id for {query}".center(100, "="))
        v2 = SearchAPIV2(
            privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
        )
        v2_job_id = v2.create_enriched_events_result(query=query, fields=schema_fields).response["job_id"]
        _logger.info(f"V2 job_id : {v2_job_id}")

        # wait_until v2 query gets data
        def get_v2_data():
            v2_data = v2.get_enriched_events_result(v2_job_id, row_count="10000000").response
            if v2_data["contacted"] == v2_data["completed"]:
                return v2_data
            return False

        v2_data = wait_until(get_v2_data, timeout=90)
        if not v2_data:
            pytest.fail(f"ERROR: V2 process search did not complete in 90 seconds. Job ID: {v2_job_id}\n")

        if v2_data["num_found"] == 0:
            pytest.fail(f"ERROR: V2 query's num_found is 0 | Response: {v2_data}\n")

        # Check num_found == num_available before proceeding with the test.
        # PS: If num_found != num_available, chances we will get inconsistent results hence this check.
        if v2_data["num_found"] != v2_data["num_available"]:
            # Try narrowing the time window and check whether the condition is met
            timestamp_list = sc.get_device_timestamp_list(v2_data)
            op = sc.get_v2_data_with_time_range(config, load_org_1, query, schema_fields, timestamp_list, test_param)
            assert op, "V2's process search resulted in 'num_found' != 'num_available' even with adjusted time range"
            v2_data = op[0]
            min_ts = time.mktime(datetime.strptime(op[1], "%Y-%m-%dT%H:%M:%S.%fZ").timetuple())
            max_ts = time.mktime(datetime.strptime(op[2], "%Y-%m-%dT%H:%M:%S.%fZ").timetuple())
            time_range = True

        # POST V1 query and get data
        _logger.info(f"S@S V1 POST to get query_id for {query}".center(100, "="))
        v1 = SearchAPI(
            privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
        )
        if time_range:
            v1_query_id = v1.create_enriched_events_result(
                q=query,
                fl=",".join(schema_fields),
                fq=collapse,
                min_device_ts=min_ts,
                max_device_ts=max_ts,
            ).response["query_id"]
        else:
            v1_query_id = v1.create_enriched_events_result(q=query, fl=",".join(schema_fields), fq=collapse).response[
                "query_id"
            ]
        _logger.info(f"V1 query_id : {v1_query_id}")

        # wait_until v1 query gets data
        def get_v1_data():
            v1_data = v1.get_enriched_events_result(v1_query_id, row_count="10000000").response
            if (
                v1_data["response_header"]["searchers_meta"]["contacted"]
                == v1_data["response_header"]["searchers_meta"]["completed"]
            ):
                return v1_data
            return False

        v1_data = wait_until(get_v1_data, timeout=90)
        if not v1_data:
            pytest.fail(f"ERROR: V1 event search did not complete in 30 seconds. Query ID: {v1_query_id}")

        # Compute V1 data struct for comparison, end result can be compared 1/1 with V2.
        v1_dict = self.construct_v1_data(test_param, v1_data, schema_data)

        # Compute V2 data struct for comparison
        v2_dict = sc.construct_v2_data(test_param, v2_data)

        ## Data Comparsion that goes over each field in V1/V2 for each process_guid/event_id
        error_count = self.check_data_integrity(test_param, v1_dict, v2_dict, schema_data, ignored_fields)
        ## PASS/FAIL criteria: Assert if error encountered during this test
        if not error_count:
            _logger.info("Search Aggregation validation for V2 completed successfully")
        else:
            pytest.fail(f"Encountered total of {error_count} error(s)")
