###################################################################################
# This script is intended to test https://jira.carbonblack.local/browse/CBC-324:
# Process & Event Details in v2 Search Experience
# Validate API users can retrieve "all available data" about a process or an event using v2 Search API routes.
# Validate All returned data fields are aggregated in a way that preserves all relevant data
# Test includes:
# 1] Process search for a group of  process_guids
# 2] Event search for a group of event_ids
# #################################################################################

import logging
import random
import time
from datetime import datetime
import pytest
from pscr_integration.api.search_api import SearchAPI
from pscr_integration.api.search_api_v2 import SearchAPIV2
from pscr_integration.search.search_common import SearchCommon
from pscr_integration.util import wait_until
from pscr_integration.lc_services import LcServices

_logger = logging.getLogger(__name__)


@pytest.mark.skipif(
    LcServices.target_service() not in LcServices.search_services(),
    reason=f"Test is irrelevant for f{LcServices.target_service()}.",
)
class TestSearchDetail:
    detail_fields = [
        "alert_category",
        "alert_id",
        "backend_timestamp",
        "blocked_hash",
        "childproc_cmdline",
        "childproc_cmdline_length",
        "childproc_count",
        "childproc_effective_reputation",
        "childproc_guid",
        "childproc_hash",
        "childproc_name",
        "childproc_pid",
        "childproc_publisher",
        "childproc_publisher_state",
        "childproc_reputation",
        "crossproc_action",
        "crossproc_api",
        "crossproc_cmdline",
        "crossproc_cmdline_length",
        "crossproc_count",
        "crossproc_effective_reputation",
        "crossproc_guid",
        "crossproc_hash",
        "crossproc_name",
        "crossproc_pid",
        "crossproc_target",
        "device_external_ip",
        "device_group",
        "device_group_id",
        "device_id",
        "device_installed_by",
        "device_internal_ip",
        "device_location",
        "device_name",
        "device_os",
        "device_os_version",
        "device_policy",
        "device_policy_id",
        "device_target_priority",
        "device_timestamp",
        "document_guid",
        "enriched",
        "enriched_event_type",
        "event_attack_stage",
        "event_description",
        "event_network_remote_port",
        "event_id",
        "event_network_local_ipv4",
        "event_network_location",
        "event_network_protocol",
        "event_network_remote_ipv4",
        "event_threat_score",
        "event_type",
        "file_scan_result",
        "fileless_scriptload_cmdline",
        "fileless_scriptload_cmdline_length",
        "fileless_scriptload_hash",
        "filemod_action",
        "filemod_count",
        "filemod_hash",
        "filemod_name",
        "filemod_publisher",
        "filemod_publisher_state",
        "filemod_type",
        "hash",
        "ingress_time",
        "legacy",
        "modload_count",
        "modload_hash",
        "modload_name",
        "modload_publisher_state",
        "netconn_count",
        "netconn_domain",
        "netconn_failed",
        "netconn_inbound",
        "netconn_ipv4",
        "netconn_ipv6",
        "netconn_listen",
        "netconn_local_ipv4",
        "netconn_local_ipv6",
        "netconn_local_location",
        "netconn_local_port",
        "netconn_location",
        "netconn_port",
        "netconn_protocol",
        "netconn_proxy_domain",
        "netconn_proxy_ipv4",
        "netconn_proxy_ipv6",
        "netconn_proxy_port",
        "org_id",
        "parent_cmdline",
        "parent_cmdline_length",
        "parent_effective_reputation",
        "parent_guid",
        "parent_hash",
        "parent_name",
        "parent_pid",
        "parent_publisher",
        "parent_publisher_state",
        "parent_reputation",
        "process_cmdline",
        "process_cmdline_length",
        "process_company_name",
        "process_duration",
        "process_effective_reputation",
        "process_elevated",
        "process_end_time",
        "process_file_description",
        "process_guid",
        "process_hash",
        "process_integrity_level",
        "process_internal_name",
        "process_loaded_script_hash",
        "process_loaded_script_name",
        "process_name",
        "process_original_filename",
        "process_pid",
        "process_privileges",
        "process_product_name",
        "process_product_version",
        "process_publisher",
        "process_publisher_state",
        "process_reputation",
        "process_service_name",
        "process_sha256",
        "process_start_time",
        "process_terminated",
        "process_username",
        "regmod_action",
        "regmod_count",
        "regmod_name",
        "report_id",
        "report_severity",
        "report_watchlist_id",
        "scriptload_content",
        "scriptload_content_length",
        "scriptload_count",
        "scriptload_hash",
        "scriptload_name",
        "scriptload_publisher_state",
        "sensor_action",
        "sensor_action_reason",
        "ttp",
        "watchlist_hit",
        "watchlist_id",
        "watchlist_name",
    ]

    def create_process_guid_list(self, v2_data):
        process_guid_list = []
        for data in v2_data["results"]:
            process_guid_list.append(data["process_guid"])
        process_guid_list = list(set(process_guid_list))
        _logger.info(f"process_guid_list - {process_guid_list}")
        return process_guid_list

    def create_event_id_list(self, v2_data):
        event_id_list = []
        for data in v2_data["results"]:
            event_id_list.append(data["event_id"])
        event_id_list = list(set(event_id_list))
        _logger.info(f"event_id_list - {event_id_list}")
        return event_id_list

    def construct_v1_data(
        self,
        test_param,
        v1_data,
        schema_data,
        ignored_fields,
        detail_fields,
        v1_query_id,
    ):
        _logger.info(f"v1_data: {v1_data}\n")
        v1_dict = {}
        sc = SearchCommon()
        for data in v1_data["data"]:
            not_check_list = []
            for d in data.keys():
                if d not in detail_fields:
                    not_check_list.append(d)
            for n in not_check_list:
                if n in data:
                    data.pop(n)
            for ignored_field in ignored_fields:
                if ignored_field in data:
                    data.pop(ignored_field)
            if data[test_param] not in v1_dict:
                v1_dict[data[test_param]] = data
                if "enriched" in v1_dict[data[test_param]]:
                    v1_dict[data[test_param]]["non-enriched"] = False
                else:
                    v1_dict[data[test_param]]["non-enriched"] = True
            else:
                # See if any fields need to be combined to the dict value
                existing_op = v1_dict[data[test_param]]
                keys = set(data.keys()) - set(existing_op.keys())
                for key in keys:
                    v1_dict[data[test_param]][key] = data[key]
                # Check values as per the schema such as find max/combine array.
                for key in data.keys():
                    if key in existing_op and key in schema_data:
                        existing_value = existing_op[key]
                        new_value = data[key]
                        schema_field = schema_data[key]
                        data_type = schema_field[0]
                        aggregation_type = schema_field[1]
                        if aggregation_type == "max" and data_type == "timestamp":
                            return_value = sc.compare_timestamp(existing_value, new_value)
                            if return_value < 0:
                                existing_op[key] = new_value
                        elif aggregation_type == "max" and data_type == "boolean":
                            if existing_value == "false" and new_value == "true":
                                existing_op[key] = new_value
                        elif aggregation_type == "max" and "int" in data_type:
                            if new_value > existing_value:
                                existing_op[key] = new_value
                        elif aggregation_type == "allfavored":
                            # Found non-enriched data, reset values to favor non-enriched
                            if not existing_op["non-enriched"] and "enriched" not in data:
                                existing_op["non-enriched"] = True
                                existing_op = sc.reset_all_favored(schema_data, existing_op, data)
                            # Aggregate further non-enriched unique values as encountered
                            elif existing_op["non-enriched"] and "enriched" not in data:
                                if existing_value != new_value:
                                    existing_op[key] = sc.compare_combine_array_values(existing_value, new_value)
                            # Treat as regular "all" category as long it's enriched only
                            elif not existing_op["non-enriched"] and "enriched" in data:
                                if existing_value != new_value:
                                    existing_op[key] = sc.compare_combine_array_values(existing_value, new_value)
                        elif aggregation_type in ("all", "limited"):
                            if key == "process_pid":
                                if isinstance(existing_value, int):
                                    existing_value = [existing_value]
                                if isinstance(new_value, int):
                                    new_value = [new_value]
                            if existing_value != new_value:
                                existing_op[key] = sc.compare_combine_array_values(existing_value, new_value)
                        elif aggregation_type == "any":
                            if not isinstance(existing_value, list):
                                existing_value = [str(existing_value)]
                            if str(new_value) not in existing_value:
                                existing_value.extend([str(new_value)])
                                existing_op[key] = existing_value
        return v1_dict

    def check_data_integrity(self, test_param, v1_dict, v2_dict, schema_data, ignored_fields, event_search):
        total_error_count = 0
        sc = SearchCommon()
        for v1_key, v1_value in v1_dict.items():
            local_error_count = 0
            if v1_key in v2_dict:
                v2_value = v2_dict[v1_key]
                _logger.debug(f"V1 VALUE ->{v1_value}")
                _logger.debug(f"V2 VALUE ->{v2_value}")
            else:
                _logger.error(f"V2 is missing '{test_param}: {v1_key}' from V1\n")
                local_error_count += 1
                total_error_count += local_error_count
                continue
            for key in v1_value:
                if key == "non-enriched":
                    continue
                if key in v2_value:
                    schema_field = schema_data[key]
                    data_type = schema_field[0]
                    aggregation_type = schema_field[1]
                    array_type = schema_field[2]
                    if data_type == "timestamp" and key != "process_start_time" and key != "process_end_time":
                        if sc.compare_timestamp(v1_value[key], v2_value[key]) > 0:
                            _logger.error(
                                f"*** V2's timestamp is not latest ***\n\n V1 => '{key}: {v1_value[key]}' |\
                                        V2 => '{key}: {v2_value[key]}'\n"
                            )
                            local_error_count += 1
                    elif aggregation_type in ("all", "allfavored"):
                        if isinstance(v1_value[key], int) and isinstance(v2_value[key], list):
                            v1_value[key] = [v1_value[key]]
                        diff = set(v1_value[key]) - set(v2_value[key])
                        if len(diff) > 0:
                            _logger.error(f"*** Missing Values ***\n\n V2 '{key}' is missing {diff}\n")
                            local_error_count += 1
                    elif aggregation_type == "limited":
                        if len(v1_value[key]) <= 10 and set(v1_value[key]) != set(v2_value[key]):
                            _logger.error(f"*** Missing alert_id ***\n\n V1 {key} \n")
                            diff = set(v1_value[key]) - set(v2_value[key])
                            _logger.error(f"*** Missing Values ***\n\n V2 '{key}' is missing {diff}\n")
                            local_error_count += 1
                        elif len(v1_value[key]) > 10 and not set(v2_value[key]).issubset(set(v1_value[key])):
                            diff = set(v2_value[key]) - set(v1_value[key])
                            _logger.error(
                                f"*** Mismatch found ***\n\n V1 '{key}' doesn't contain V2's value\
                                                                            '{diff}'\n"
                            )
                            local_error_count += 1
                    elif aggregation_type == "any":
                        if isinstance(v1_value[key], list):
                            if v2_value[key] not in v1_value[key]:
                                # if str(v2_value[key]) not in v1_value[key]:
                                _logger.error(
                                    f"*** Mismatch found ***\n\n V2 '{key}:{v2_value[key]}' didn't\
                                            match any of the values in V1 for '{key}:{v1_value[key]}'\n"
                                )
                                local_error_count += 1

                        if isinstance(v1_value[key], str):
                            if str(v2_value[key]).strip() != v1_value[key].strip():
                                _logger.error(
                                    f"*** Mismatch found ***\n\n V1 => '{key}: {v1_value}' |\
                                    V2 => '{key}: {v2_value[key]}'\n"
                                )
                                local_error_count += 1
                    elif aggregation_type == "enriched":
                        # if this is an array type, ignore
                        if not array_type and isinstance(v2_value[key], list) and event_search:
                            _logger.error(f"*** V2 Details field '{key}' has incorrect type '{data_type}'\n")
                            local_error_count += 1
                        if isinstance(v1_value[key], list) and isinstance(v2_value[key], list):
                            if not set(v1_value[key]).issubset(set(v2_value[key])):
                                _logger.error(
                                    f"*** Mismatch found ***\n\n V2 '{key}:{v2_value[key]}' doesn't contain \
                                    V1's value '{key}:{v1_value[key]}'\n"
                                )
                                local_error_count += 1
                        elif isinstance(v1_value[key], list) and not isinstance(v2_value[key], list):
                            if v2_value[key] not in v1_value[key]:
                                _logger.error(
                                    f"*** Mismatch found ***\n\n V2 '{key}:{v2_value[key]}' doesn't contain \
                                    V1's value '{key}:{v1_value[key]}'\n"
                                )
                            local_error_count += 1
                        elif not isinstance(v1_value[key], list) and isinstance(v2_value[key], list):
                            if v1_value[key] not in v2_value[key]:
                                _logger.error(
                                    f"*** Mismatch found ***\n\n V2 '{key}:{v2_value[key]}' doesn't \
                                    contain V1's value '{key}:{v1_value[key]}'\n"
                                )
                        elif v1_value[key] != v2_value[key]:
                            _logger.error(
                                f"*** Mismatch found ***\n\n V1 => '{key}: {v1_value[key]}' |\
                                V2 => '{key}: {v2_value[key]}'\n"
                            )
                            local_error_count += 1
                else:
                    _logger.error(f"*** Missing Field ***\n\n V2 is missing '{key}: {v1_value[key]}'\n")
                    local_error_count += 1

            if not local_error_count:
                _logger.info(f"{v1_key} is verified!\n")
            else:
                _logger.error(f"{local_error_count} error found in {v1_key}\n")
                total_error_count += local_error_count
        return total_error_count

    def test_process_search_detail(self, config, load_org_1):
        test_param = "process_guid"
        query = test_param + ":*"
        time_range = False
        ignored_fields = [
            "childproc_cmdline",
            "childproc_cmdline_length",
            "childproc_effective_reputation",
            "childproc_guid",
            "childproc_hash",
            "childproc_name",
            "childproc_pid",
            "childproc_publisher",
            "childproc_publisher_state",
            "childproc_reputation",
            "crossproc_action",
            "crossproc_api",
            "crossproc_cmdline",
            "crossproc_cmdline_length",
            "crossproc_effective_reputation",
            "crossproc_guid",
            "crossproc_hash",
            "crossproc_name",
            "crossproc_pid",
            "crossproc_target",
            "document_guid",
            "event_id",
            "event_description",
            "event_type",
            "event_network_local_ipv4",
            "event_network_location",
            "event_network_protocol",
            "event_network_remote_ipv4",
            "event_network_remote_port",
            "fileless_scriptload_cmdline",
            "fileless_scriptload_cmdline_length",
            "fileless_scriptload_hash",
            "filemod_action",
            "filemod_hash",
            "filemod_name",
            "filemod_publisher",
            "filemod_publisher_state",
            "filemod_type",
            "hash",
            "modload_hash",
            "modload_name",
            "modload_publisher_state",
            "netconn_domain",
            "netconn_failed",
            "netconn_inbound",
            "netconn_ipv4",
            "netconn_ipv6",
            "netconn_listen",
            "netconn_local_ipv4",
            "netconn_local_ipv6",
            "netconn_local_location",
            "netconn_local_port",
            "netconn_location",
            "netconn_port",
            "netconn_protocol",
            "netconn_proxy_domain",
            "netconn_proxy_ipv4",
            "netconn_proxy_ipv6",
            "netconn_proxy_port",
            "process_loaded_script_hash",
            "process_loaded_script_name",
            "regmod_name",
            "report_id",
            "report_severity",
            "report_watchlist_id",
            "scriptload_content",
            "scriptload_content_length",
            "scriptload_hash",
            "scriptload_name",
            "scriptload_publisher_state",
            "watchlist_id",
            "watchlist_name",
            "modload_publisher",
            "regmod_action",
            "process_publisher_state",
            "file_scan_result",
            "process_cmdline_length",
            "parent_cmdline_length",
            "parent_publisher_state",
            "device_target_priority",
            "device_location",
            "process_publisher",
            "parent_publisher",
            "parent_reputation",
            "parent_cmdline",
            "process_cmdline_length",
        ]

        # Download and parse latest schema
        _logger.info("Download and parse latest schema")
        sc = SearchCommon()
        schema_data = sc.get_schema(config, load_org_1, query)
        _logger.info(f"schema_data: {schema_data}\n")
        schema_fields = list(schema_data.keys())

        ## POST V2 query process_guid:* to get all process_guid
        _logger.info(f"Search Detail V2 POST to get query_id for {query}".center(100, "="))
        v2 = SearchAPIV2(
            privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
        )
        v2_job_id = v2.create_process_result_v2(query=query, fields=schema_fields).response["job_id"]
        _logger.info(f"V2 job_id : {v2_job_id}")

        # wait_until v2 query gets data
        def get_v2_data():
            v2_data = v2.get_process_result_v2(v2_job_id, row_count="100000000").response
            if v2_data["contacted"] == v2_data["completed"]:
                return v2_data
            return False

        v2_data = wait_until(get_v2_data, timeout=90)
        if not v2_data:
            pytest.fail(f"ERROR: V2 process search did not complete in 90 seconds. Job ID: {v2_job_id}")

        if v2_data["num_found"] == 0:
            pytest.fail(f"ERROR: V2 query's num_found is 0 | Response: {v2_data}")

        # Check num_found == num_available before proceeding with the test.

        if v2_data["num_found"] != v2_data["num_available"]:
            # Try narrowing the time window and check whether the condition is met
            timestamp_list = sc.get_device_timestamp_list(v2_data)
            op = sc.get_v2_data_with_time_range(config, load_org_1, query, schema_fields, timestamp_list, test_param)
            if not op:
                pytest.fail(
                    "ERROR: V2's process search resulted in 'num_found' != 'num_available' even with adjusted range"
                )
            v2_data = op[0]
            min_ts = time.mktime(datetime.strptime(op[1], "%Y-%m-%dT%H:%M:%S.%fZ").timetuple())
            max_ts = time.mktime(datetime.strptime(op[2], "%Y-%m-%dT%H:%M:%S.%fZ").timetuple())
            _logger.info(f"Time Range is changed to {min_ts} to {max_ts}")
            time_range = True

        # Create process_guid list
        process_guid_list = self.create_process_guid_list(v2_data)

        # Take a group of process_guid from process_guid_list to test
        _logger.info("Starting to compare V1 and V2 detail query")
        g_c = len(process_guid_list)
        if g_c > 32:
            sample_list = [1, 2, 4, 8, 16]
        else:
            sample_list = [1, g_c // 8, g_c // 4, g_c // 2, g_c]

        process_detail_error_count = 0
        for sample in sample_list:
            multiple_guid_list = random.sample(process_guid_list, sample)
            _logger.info(f"V2 process_guids : {multiple_guid_list}")
            v2_job_id = v2.create_process_detail_v2(process_guids=multiple_guid_list).response["job_id"]
            _logger.info(f"V2 process_guids job_id : {v2_job_id}")

            # wait_until v2 query gets data
            def get_v2_detail_data():
                v2_data = v2.get_process_detail_v2(v2_job_id, row_count="100000000").response
                if v2_data["contacted"] == v2_data["completed"]:
                    return v2_data
                return False

            v2_data = wait_until(get_v2_detail_data, timeout=90)
            if not v2_data:
                pytest.fail(f"ERROR: V2 process search did not complete in 90 seconds. Job ID: {v2_job_id}")

            if v2_data["num_found"] == 0:
                _logger.error(f"ERROR: V2 {multiple_guid_list} num_found = 0")
                continue

            # Compute V2 data struct for comparison
            v2_dict = sc.construct_v2_data(test_param, v2_data)

            # V1 query every process_guid with cb.all_fields = True, construct all_v1_dict for later comparison.
            v1_dict = {}
            all_v1_dict = {}
            for p_guid in multiple_guid_list:
                process_detail_query = test_param + ":" + p_guid
                _logger.info(f"V1 Query {process_detail_query}".center(100, "="))
                v1 = SearchAPI(
                    privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
                    org_id=load_org_1["loader"].org_id,
                )
                if time_range:
                    v1_query_id = v1.create_process_result(
                        q=process_detail_query,
                        all_fields=True,
                        min_device_ts=min_ts,
                        max_device_ts=max_ts,
                    ).response["query_id"]
                else:
                    v1_query_id = v1.create_process_result(q=process_detail_query, all_fields=True).response["query_id"]

                # wait_until v1 query gets data
                # pylint: disable=cell-var-from-loop
                def get_v1_data(v1_query_id=v1_query_id):
                    v1_data_inner = v1.get_process_result(v1_query_id, row_count="100000000").response
                    if (
                        v1_data_inner["response_header"]["searchers_meta"]["contacted"]
                        == v1_data_inner["response_header"]["searchers_meta"]["completed"]
                    ):
                        return v1_data_inner
                    return False

                v1_data = wait_until(get_v1_data, timeout=90)
                if not v1_data or v1_data["response_header"]["num_found"] == 0:
                    pytest.fail(f"ERROR: V1 process query not complete. Query ID: {v1_query_id}")

                # Construct all_v1_dict for later comparison with v2 query
                v1_dict = self.construct_v1_data(
                    test_param,
                    v1_data,
                    schema_data,
                    ignored_fields,
                    self.detail_fields,
                    v1_query_id,
                )
                all_v1_dict.update(v1_dict)

                ## Data Comparsion that goes over each field in V1/V2 for each process_guid
                _logger.info(f"v1_dict: {all_v1_dict}\n")
                _logger.info(f"v2_dict: {v2_dict}\n")
                error_count = self.check_data_integrity(
                    test_param, all_v1_dict, v2_dict, schema_data, ignored_fields, False
                )
                process_detail_error_count += error_count
                if not error_count:
                    _logger.info(f"{sample} process_guids Validation Passed!")
                else:
                    _logger.warning(f"Encountered total of {error_count} error(s)")

        if process_detail_error_count > 0:
            pytest.fail(f"Process details search: {process_detail_error_count} error(s)")
        else:
            _logger.info("Process Detail Validation Succcessful!")

    @pytest.mark.smoke_test
    def test_event_search_detail(self, config, load_org_1):
        test_param = "event_id"
        query = test_param + ":*"
        time_range = False
        ignored_fields = [
            "document_guid",
            "fileless_scriptload_cmdline",
            "fileless_scriptload_cmdline_length",
            "fileless_scriptload_hash",
            "hash",
            "modload_hash",
            "modload_name",
            "modload_publisher_state",
            "modload",
            "netconn",
            "netconn_proxy_domain",
            "netconn_proxy_ipv4",
            "netconn_proxy_ipv6",
            "netconn_proxy_port",
            "report_id",
            "report_severity",
            "report_watchlist_id",
            "scriptload_content",
            "scriptload_content_length",
            "watchlist_id",
            "watchlist_name",
            "crossproc",
            "segment_id",
            "protobuf_metadata",
            "netconn",
            "childproc",
            "crossproc",
            "regmod",
            "filemod",
            "scriptload",
            "fileless",
            "device_time_correction",
        ]

        # Download and parse latest schema
        _logger.info("Download and parse latest schema")
        sc = SearchCommon()
        schema_data = sc.get_schema(config, load_org_1, query)
        schema_fields = list(schema_data.keys())

        ## POST V2 query and get data
        _logger.info(f"Search Detail V2 POST to get query_id for {query}".center(100, "="))
        v2 = SearchAPIV2(
            privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
        )
        v2_job_id = v2.create_enriched_events_result(query=query, fields=schema_fields).response["job_id"]
        _logger.info(f"V2 job_id : {v2_job_id}")

        # wait_until v2 query gets data
        def get_v2_data():
            v2_data_inner = v2.get_enriched_events_result(v2_job_id, row_count="10000000").response
            if v2_data_inner["contacted"] == v2_data_inner["completed"]:
                return v2_data_inner
            return False

        v2_data = wait_until(get_v2_data, timeout=90)
        if not v2_data:
            pytest.fail(f"ERROR: V2 event search did not complete in 90 seconds. Job ID: {v2_job_id}")

        if v2_data["num_found"] == 0:
            _logger.error(f"ERROR: V2 query's num_found is 0 | Response: {v2_data}")

        # Check num_found == num_available before proceeding with the test.
        if v2_data["num_found"] != v2_data["num_available"]:
            # Try narrowing the time window and check whether the condition is met
            timestamp_list = sc.get_device_timestamp_list(v2_data)
            op = sc.get_v2_data_with_time_range(config, load_org_1, query, schema_fields, timestamp_list, test_param)
            if not op:
                pytest.fail(
                    "ERROR: V2's event search resulted in 'num_found' != 'num_available' even with adjusted time range"
                )
            v2_data = op[0]
            min_ts = time.mktime(datetime.strptime(op[1], "%Y-%m-%dT%H:%M:%S.%fZ").timetuple())
            max_ts = time.mktime(datetime.strptime(op[2], "%Y-%m-%dT%H:%M:%S.%fZ").timetuple())
            time_range = True

        # Create event_id list
        event_id_list = self.create_event_id_list(v2_data)

        _logger.info("Starting to compare V1 and V2 event detail query!")
        e_c = len(event_id_list)

        if e_c > 32:
            sample_list = [1, 2, 4, 8, 16]
        else:
            sample_list = [1, e_c // 8, e_c // 4, e_c // 2, e_c]
        event_detail_error_count = 0
        for sample in sample_list:
            multiple_id_list = event_id_list[:sample]
            _logger.info(f"V2 event_ids : {multiple_id_list}")
            v2_job_id = v2.create_events_detail_result(event_ids=multiple_id_list).response["job_id"]
            _logger.info(f"V2 event_ids job_id : {v2_job_id}")

            # wait_until v2 query gets data
            def get_v2_detail_data():
                v2_data = v2.get_events_detail_result(v2_job_id, row_count="100000000").response
                if v2_data["contacted"] == v2_data["completed"]:
                    return v2_data
                return False

            v2_data = wait_until(get_v2_detail_data, timeout=90)
            if not v2_data:
                pytest.fail(f"ERROR: V2 event search did not complete in 90 seconds. Job ID: {v2_job_id}")
            if v2_data["num_found"] == 0:
                pytest.fail(f"ERROR: V2 {multiple_id_list} num_found = 0")
                continue

            # Construct v1_dict.
            all_v1_dict = {}
            v1_dict = {}
            for e_id in multiple_id_list:
                event_detail_query = test_param + ":" + '"' + e_id + '"'
                _logger.info(f"V1 Query {event_detail_query}".center(100, "="))
                v1 = SearchAPI(
                    privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
                    org_id=load_org_1["loader"].org_id,
                )
                if time_range:
                    v1_query_id = v1.create_enriched_events_result(
                        q=event_detail_query,
                        all_fields=True,
                        min_device_ts=min_ts,
                        max_device_ts=max_ts,
                    ).response["query_id"]
                else:
                    v1_query_id = v1.create_enriched_events_result(q=event_detail_query, all_fields=True).response[
                        "query_id"
                    ]

                # wait_until v1 query gets data
                # pylint: disable=cell-var-from-loop
                def get_v1_data(v1_query_id=v1_query_id):
                    v1_data_inner = v1.get_enriched_events_result(v1_query_id, row_count="10000000").response
                    if (
                        v1_data_inner["response_header"]["searchers_meta"]["contacted"]
                        == v1_data_inner["response_header"]["searchers_meta"]["completed"]
                    ):
                        return v1_data_inner
                    return False

                v1_data = wait_until(get_v1_data, timeout=90)
                if not v1_data or v1_data["response_header"]["num_found"] == 0:
                    pytest.fail(f"ERROR: V1 event query not complete in 90 seconds. Query ID: {v1_query_id}")

                v1_dict = self.construct_v1_data(
                    test_param,
                    v1_data,
                    schema_data,
                    ignored_fields,
                    self.detail_fields,
                    v1_query_id,
                )
                all_v1_dict.update(v1_dict)

                # Compute V2 data struct for comparison
                v2_dict = sc.construct_v2_data(test_param, v2_data)

                ## Data Comparsion that goes over each field in V1/V2 for each process_guid/event_id
                _logger.info(f"v1_dict: {all_v1_dict}\n")
                _logger.info(f"v2_dict: {v2_dict}\n")
                error_count = self.check_data_integrity(
                    test_param, all_v1_dict, v2_dict, schema_data, ignored_fields, True
                )
                event_detail_error_count += error_count

                ## PASS/FAIL criteria: Assert if error encountered during this test
                if not error_count:
                    _logger.info(f"{sample} event_ids Validation Passed!")
                else:
                    _logger.error(f"Encountered total of {error_count} error(s)")

        if event_detail_error_count > 0:
            pytest.fail(f"Event details search: {event_detail_error_count} error(s)")
        else:
            _logger.info("Event Detail Validation Succcessful!")
