import logging
import time

import pytest

from pscr_integration.api.search_api_v2 import SearchAPIV2
from pscr_integration.data_loader import Loader
from pscr_integration.lc_services import LcServices
from pscr_integration.search.searcher_factory import SearcherFactory
from pscr_integration.search.searcher_type import SearcherType, FacetParams
from pscr_integration.test_lifecycle.deprovision import deprovision
from pscr_integration.util import wait_until

_logger = logging.getLogger(__name__)


@pytest.mark.parametrize("query_api", [SearcherType.SERVICE])
@pytest.mark.skip("Need to fix those tests: DSER-19142")
class TestEventSearches:
    @staticmethod
    def _get_process_guids(query):
        start_response = query.submit_query("*:*")
        start_query_id = start_response.response.get("query_id", "")
        assert len(start_query_id) > 0

        def get_process_guid():
            results = query.get_results(start_query_id)
            header = results.response.get("response_header", {})
            if header.get("num_found", 0) > 10:
                d = results.response.get("data", [])
                if d and len(d) > 0:
                    return tuple(set(el.get("process_guid") for el in d))[:5]
            else:
                return tuple()

        return wait_until(condition=get_process_guid, timeout=20)

    def test_get_events(self, config, load_org_1, query_api):
        query = SearcherFactory.get_searcher(
            query_api,
            service_endpoint=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
            device_id_start=load_org_1["loader"].device_id_start,
            device_id_end=load_org_1["loader"].device_id_end,
        )
        process_guids = self._get_process_guids(query)

        assert process_guids
        assert all(process_guids)
        assert any(
            (
                len(query.get_events(f"process_guid:{process_guid}", process_guid).response.get("docs", [])) > 0
                for process_guid in process_guids
            )
        )

    def test_facet_query(self, config, load_org_1, query_api):
        facet_fields = ["process_guid", "event_type", "event_guid"]

        query = SearcherFactory.get_searcher(
            query_api,
            service_endpoint=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
            device_id_start=load_org_1["loader"].device_id_start,
            device_id_end=load_org_1["loader"].device_id_end,
        )
        process_guid = self._get_process_guids(query)[0]

        facet_results = query.get_events(
            f"process_guid:{process_guid}",
            process_guid,
            facet=FacetParams(field=facet_fields),
        ).response["facet_counts"]["facet_fields"]
        for facet_field in facet_fields:
            assert facet_field in facet_results


@pytest.mark.skipif(
    LcServices.target_service() not in LcServices.search_services(),
    reason=f"Test is irrelevant for f{LcServices.target_service()}.",
)
@pytest.mark.parametrize("query_api", [SearcherType.SERVICE])
class TestSearches:
    # @pytest.mark.smoke_test
    def test_count_data(self, query_api, config, load_org_1):
        query = SearcherFactory.get_searcher(
            query_api,
            service_endpoint=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
            device_id_start=load_org_1["loader"].device_id_start,
            device_id_end=load_org_1["loader"].device_id_end,
        )
        result_count = query.count_query(
            20,
            "-report_id:* -enriched:*",
            config["config_file"]["test_constants"]["org1_process_count"],
            0,
        )
        assert result_count == config["config_file"]["test_constants"]["org1_process_count"]

    def test_count_data_optimized(self, query_api, config, load_org_1):
        query = SearcherFactory.get_searcher(
            query_api,
            service_endpoint=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
            device_id_start=load_org_1["loader"].device_id_start,
            device_id_end=load_org_1["loader"].device_id_end,
        )
        result_count = query.count_query(
            600,
            "-report_id:* -enriched:*",
            config["config_file"]["test_constants"]["org1_process_count"],
            2,
        )
        assert result_count == config["config_file"]["test_constants"]["org1_process_count"]

    @pytest.mark.smoke_test
    def test_count_legacy_data(self, query_api, config, load_org_1):
        query = SearcherFactory.get_searcher(
            query_api,
            service_endpoint=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
            device_id_start=load_org_1["loader"].device_id_start,
            device_id_end=load_org_1["loader"].device_id_end,
        )
        result_count = query.count_query(
            20,
            "enriched_event_type:*",
            config["config_file"]["test_constants"]["org1_legacy_count"],
            0,
        )
        assert result_count == config["config_file"]["test_constants"]["org1_legacy_count"]

    @pytest.mark.smoke_test
    def test_count_legacy_alert1_data(self, query_api, config, load_org_1):
        query = SearcherFactory.get_searcher(
            query_api,
            service_endpoint=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
            device_id_start=load_org_1["loader"].device_id_start,
            device_id_end=load_org_1["loader"].device_id_end,
        )
        result_count = query.count_query(
            20,
            "alert_id:" + config["config_file"]["test_constants"]["org1_alertid1"],
            config["config_file"]["test_constants"]["org1_alertid1_count"],
            0,
        )
        assert result_count == config["config_file"]["test_constants"]["org1_alertid1_count"]

    @pytest.mark.smoke_test
    def test_count_legacy_alert2_data(self, query_api, config, load_org_1):
        query = SearcherFactory.get_searcher(
            query_api,
            service_endpoint=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
            device_id_start=load_org_1["loader"].device_id_start,
            device_id_end=load_org_1["loader"].device_id_end,
        )
        result_count = query.count_query(
            20,
            "alert_id:" + config["config_file"]["test_constants"]["org1_alertid2"],
            config["config_file"]["test_constants"]["org1_alertid2_count"],
            0,
        )
        assert result_count == config["config_file"]["test_constants"]["org1_alertid2_count"]

    @pytest.mark.smoke_test
    def test_count_legacy_staticscan(self, query_api, config, load_org_1):
        query = SearcherFactory.get_searcher(
            query_api,
            service_endpoint=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
            device_id_start=load_org_1["loader"].device_id_start,
            device_id_end=load_org_1["loader"].device_id_end,
        )
        result_count = query.count_query(
            20,
            "file_scan_result:" + config["config_file"]["test_constants"]["org1_staticscan_value"],
            config["config_file"]["test_constants"]["org1_staticscan_count"],
            0,
        )
        assert result_count == config["config_file"]["test_constants"]["org1_staticscan_count"]

    @pytest.mark.smoke_test
    def test_count_legacy_event_process(self, query_api, config, load_org_1):
        query = SearcherFactory.get_searcher(
            query_api,
            service_endpoint=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
            device_id_start=load_org_1["loader"].device_id_start,
            device_id_end=load_org_1["loader"].device_id_end,
        )
        result_count = query.count_query(
            20,
            "event_type:crossproc",
            config["config_file"]["test_constants"]["org1_legacy_event_process_count"],
            0,
        )
        assert result_count == config["config_file"]["test_constants"]["org1_legacy_event_process_count"]

    @pytest.mark.smoke_test
    def test_count_legacy_event_file(self, query_api, config, load_org_1):
        query = SearcherFactory.get_searcher(
            query_api,
            service_endpoint=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
            device_id_start=load_org_1["loader"].device_id_start,
            device_id_end=load_org_1["loader"].device_id_end,
        )
        result_count = query.count_query(
            20,
            "event_type:filemod",
            config["config_file"]["test_constants"]["org1_legacy_event_file_count"],
            0,
        )
        assert result_count == config["config_file"]["test_constants"]["org1_legacy_event_file_count"]

    @pytest.mark.smoke_test
    def test_count_legacy_event_network(self, query_api, config, load_org_1):
        query = SearcherFactory.get_searcher(
            query_api,
            service_endpoint=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
            device_id_start=load_org_1["loader"].device_id_start,
            device_id_end=load_org_1["loader"].device_id_end,
        )
        result_count = query.count_query(
            20,
            "event_type:netconn",
            config["config_file"]["test_constants"]["org1_legacy_event_network_count"],
            0,
        )
        assert result_count == config["config_file"]["test_constants"]["org1_legacy_event_network_count"]

    @pytest.mark.smoke_test
    def test_count_legacy_event_registry(self, query_api, config, load_org_1):
        query = SearcherFactory.get_searcher(
            query_api,
            service_endpoint=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
            device_id_start=load_org_1["loader"].device_id_start,
            device_id_end=load_org_1["loader"].device_id_end,
        )
        result_count = query.count_query(
            20,
            "event_type:regmod",
            config["config_file"]["test_constants"]["org1_legacy_event_registry_count"],
            0,
        )
        assert result_count == config["config_file"]["test_constants"]["org1_legacy_event_registry_count"]

    @pytest.mark.smoke_test
    def test_facet_search(self, query_api, config, load_org_1):
        facet_fields = ["process_name", "process_hash", "device_name"]

        query = SearcherFactory.get_searcher(
            query_api,
            service_endpoint=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
            device_id_start=load_org_1["loader"].device_id_start,
            device_id_end=load_org_1["loader"].device_id_end,
        )
        submit_response = query.submit_query(query_string="-report_id:*", facet=FacetParams(field=facet_fields))

        query_id = submit_response.response["query_id"]

        def facet_search_complete():
            query_results_data = query.get_results(query_id)
            return (
                query_results_data.response["response_header"]["searchers_meta"]["contacted"]
                == query_results_data.response["response_header"]["searchers_meta"]["completed"]
            )

        assert wait_until(
            facet_search_complete
        ), f"Facets search did not complete, response: {query.get_results(query_id=query_id)}"

        facet_results = query.get_results(query_id=query_id).response["facets"]["facet_fields"]
        for facet_field in facet_fields:
            assert facet_field in facet_results

    @pytest.mark.smoke_test
    def test_xdr_events(self, query_api, config, load_org_1):
        query = SearcherFactory.get_searcher(
            query_api,
            service_endpoint=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
            device_id_start=load_org_1["loader"].device_id_start,
            device_id_end=load_org_1["loader"].device_id_end,
        )
        result_count = query.count_query(
            20,
            "netconn_application_protocol:HTTP OR netconn_application_protocol:TLS",
            config["config_file"]["test_constants"]["org1_xdr_event_count"],
            0,
        )
        assert result_count == config["config_file"]["test_constants"]["org1_xdr_event_count"]

    # @pytest.mark.smoke_test
    def test_container_events(self, query_api, config, load_org_1):
        schema_fields = list(["process_guid"])
        v2 = SearchAPIV2(
            privatelink_hostname=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=load_org_1["loader"].org_id,
            device_id_start=None,
            device_id_end=None,
        )
        v2_job_id = v2.create_process_result_v2(query="container_id:*", fields=schema_fields).response["job_id"]
        _logger.info(f"V2 job_id : {v2_job_id}")

        # wait_until v2 query gets data
        def get_v2_data():
            v2_data = v2.get_process_result_v2(v2_job_id, row_count="10000000").response
            if v2_data["contacted"] == v2_data["completed"]:
                return v2_data
            return False

        v2_data = wait_until(get_v2_data, timeout=90)
        if not v2_data:
            pytest.fail(f"ERROR: V2 process search did not complete in 90 seconds. Job ID: {v2_job_id}\n")

        if v2_data["num_found"] == 0:
            pytest.fail(f"ERROR: V2 query's num_found is 0 | Response: {v2_data}\n")


@pytest.mark.skipif(
    LcServices.target_service() not in LcServices.search_services(),
    reason=f"Test is irrelevant for f{LcServices.target_service()}.",
)
@pytest.mark.parametrize("query_api", [SearcherType.SERVICE])
class TestTimingSearches:
    @pytest.mark.smoke_test
    def test_data_showup_timer(self, config, query_api):
        loader = Loader(
            task_definition=config["env_vars"]["TEST_ORG1_TASK_DEFINITION"],
            ecs_cluster_name=config["env_vars"]["ECS_CLUSTER_NAME"],
            security_group=config["env_vars"]["ECS_CLUSTER_SECURITY_GROUP"],
            subnet=config["env_vars"]["PRIVATE_SUBNET"],
            documents_bucket=config["env_vars"]["DOCUMENTS_BUCKET"],
            events_stream=config["env_vars"]["EVENTS_STREAM"],
            device_count=config["config_file"]["test_constants"]["org1_device_count"],
            org_id_suffix="3",
        )
        loader.load_data()
        finished_loading_time = time.time()
        query = SearcherFactory.get_searcher(
            query_api,
            service_endpoint=config["env_vars"]["API_PRIVATELINK_HOSTNAME"],
            org_id=loader.org_id,
            device_id_start=loader.device_id_start,
            device_id_end=loader.device_id_end,
        )
        data_available_for_query_threshold = config["config_file"]["test_constants"][
            "data_availiable_for_query_threshold"
        ]
        query.count_query(data_available_for_query_threshold * 1.5, "*:*", 0, 0)
        data_found_time = time.time()
        deprovision(config["env_vars"]["DEPROVISIONING_SQS_QUEUE"], loader.org_id, loader.org_id)
        assert data_found_time - finished_loading_time < data_available_for_query_threshold
