import os
from enum import Enum
from typing import AnyStr, Optional


class TestMode(Enum):
    integration = 1
    smoke = 2


class _ActiveSession:
    """
    This disgusting private singleton is a workaround for test run "settings" also working when running single
    functions from intellij or when running pytest below the root directory where the root conftest.py is not
    interpreted.
    The pytest.smoke solution was convenient but after looking around for a better solution I discovered
    that it worked by monkeypatching the pytest object. :(

    This pytest independent "Active Session" state should probably only be accessed through the helper functions below.
    """

    __instance = None

    def __new__(cls):
        if _ActiveSession.__instance is None:
            _ActiveSession.__instance = object.__new__(cls)
            # defaults settings for session
            _ActiveSession.mode = TestMode.integration
        return _ActiveSession.__instance


def set_execution_mode(mode: TestMode = TestMode.integration):
    """
    Set the current test execution mode.
    Right now can be one of TestMode.integration or TestMode.smoke
    :param mode: TestMode enum type
    """
    session = _ActiveSession()
    session.mode = mode


def is_smoke() -> bool:
    """
    Quick check to see if the current execution mode is smoke.
    :return: boolean
    """
    session = _ActiveSession()
    return session.mode == TestMode.smoke


def is_k8s() -> bool:
    return os.environ.get("RUNTIME_ENVIRONMENT_TYPE") == "k8s"


def is_kc() -> bool:
    """
    Hardcoded check to see if we're in one of the paired KC envs.
    """
    return is_k8s() and (os.environ.get("ENV_NAME") in {"dev", "staging", "usprod", "euprod", "ausprod", "apjprod"})


def is_ecs() -> bool:
    """
    Check if we are running from the it env pool or in commercial ecs
    :return: boolean
    """
    return not is_k8s()


def is_gov() -> bool:
    """
    Hardcoded check to see if we're in one of the gov envs.
    """
    return os.environ.get("ENV_NAME") in {"gint1usw2", "gstg1usgw1", "gprd1usgw1"}


def has_clm() -> bool:
    return is_k8s() and (
        os.environ.get("ENV_NAME")
        in {
            "dev",
            "staging",
            "apjprod",
            "ausprod",
            "euprod",
            "usprod",
            "int1usw2",
            "int2use2",
            "prd1ew2",
            "scl4use2",
        }
    )


class Env:
    _kc_map = {
        "dev": "kcdev01",
        "staging": "kctest03",
        "apjprod": "prd1apne1",
        "ausprod": "prd1apse2",
        "euprod": "prd1ec1",
        "usprod": "prd5use1",
    }
    _commercial_map = {v: k for k, v in _kc_map.items()}

    def __init__(self, commercial_name: Optional[AnyStr] = None, kc_name: Optional[AnyStr] = None):
        if not any((commercial_name, kc_name)):
            raise AttributeError("Must set commercial_name or kc_name")
        if not kc_name:
            self.commercial_name = commercial_name
            self.kc_name = self._kc_map[commercial_name]
        else:
            self.kc_name = kc_name
            self.commercial_name = self._commercial_map[kc_name]
