from time import sleep
from typing import Callable, TypeVar

import hashlib
from copy import deepcopy
import struct
from io import BytesIO
import zipfile
import yaml

T = TypeVar("T")


def get_sqs_job_id(job_id):
    # Supporting SNS/ SQS Mode - If Job ID does not -sqs, let's append it to Job ID.
    if "sqs" not in job_id:
        job_id += "-sqs"

    return job_id


def wait_until(condition: Callable[[], T], timeout: int = 15, interval: int = 1) -> T:
    """
    Wait until given condition is "truish" (either True, or anything that makes "if return value" true)
    :param condition: a lambda, function or otherwise callable class that returns "truish" and "falsish" values.
    :param timeout: timeout counted in ticks
    :param interval: how long each "tick" is
    :return: The result of the callable condition when it succeeded or False in case of timeout.
    """
    result = None
    for _ in range(timeout):
        result = condition()
        if result:
            return result
        sleep(interval)
    return result


def pretty_event(*events: dict) -> str:
    """
    Return a nice string of indented yaml, better than the compact dicts one would get from logging raw events.
    Useful for debugging.
    Example: _logger.info(pretty_event(*all_results))
    :param events: one or more event objects
    :return: A nicely formatted and indented string
    """
    return yaml.dump(events, indent=2, default_flow_style=False)


def _report_iocsv2(report):
    """
    Translates iocs -> iocs_v2
    """
    fields = {
        "md5": "process_hash",
        "domain": "netconn_domain",
        "ipv4": "netconn_ipv4",
        "ipv6": "netconn_ipv6",
    }

    iocs = report.get("iocs", None)

    if iocs:
        report = deepcopy(report)  # we don't want to change the original

        iocs_v2 = []

        for k, v in iocs.items():
            # Create id from as md5 of ioc type and values.
            md5 = hashlib.md5(usedforsecurity=False)
            md5.update(k.encode())
            for s in sorted(v):
                md5.update(s.encode())

            iocs_v2.append(
                {
                    "id": md5.hexdigest(),
                    "match_type": "equality",
                    "field": fields[k],
                    "values": iocs[k],
                    "link": None,
                }
            )
        report["iocs_v2"] = iocs_v2
        report["iocs_total_count"] = len(iocs_v2)
        report["iocs"] = None

    return report


def generate_zip(f):
    mem_zip = BytesIO()

    with zipfile.ZipFile(mem_zip, mode="w", compression=zipfile.ZIP_DEFLATED) as zf:
        zf.writestr(f[0], f[1])

    return mem_zip.getvalue()


def little_end(port: int) -> int:
    "Transform int to bytes in little-endian order"
    port_bytes = port.to_bytes(4, byteorder="big")
    return struct.unpack("<I", port_bytes)[0]
